#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>

#include "config.h"

#include "xdata.h"
#include "ban.h"
#include "all.h"
#include "pixmaps.h"

void CreateCtrl(int x, int y, Ctrl *ctrl)
{
    ConstructWindata(&xdata, win.id,
                     x, y, CTRL_WIDTH, CTRL_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &ctrl->win);
    XSetWindowBackgroundPixmap(xdata.display, ctrl->win.id, CTRL_XPM.pixmap);
    CreateButton(&xdata, &ctrl->win, TOP_XPM.pixmap, TOP_F_XPM.pixmap,
                 0, 0, CTRL_WIDTH / 2, CTRL_HEIGHT / 2, 0,
                 &ctrl->top);
    CreateButton(&xdata, &ctrl->win, PREV_XPM.pixmap, PREV_F_XPM.pixmap,
                 0, CTRL_HEIGHT / 2, CTRL_WIDTH / 2, CTRL_HEIGHT / 2, 0,
                 &ctrl->prev);
    CreateButton(&xdata, &ctrl->win, BOTTOM_XPM.pixmap, BOTTOM_F_XPM.pixmap,
                 CTRL_WIDTH / 2, 0, CTRL_WIDTH / 2, CTRL_HEIGHT / 2, 0,
                 &ctrl->bottom);
    CreateButton(&xdata, &ctrl->win, NEXT_XPM.pixmap, NEXT_F_XPM.pixmap,
                 CTRL_WIDTH / 2, CTRL_HEIGHT / 2, CTRL_WIDTH / 2, CTRL_HEIGHT / 2, 0,
                 &ctrl->next);
}

void MapCtrl(Ctrl *ctrl, int ctrlmode)
{
    switch(ctrlmode)
    {
      case CTRL_MODE_TOP:
        XUnmapWindow(xdata.display, ButtonId(ctrl->top));
        XUnmapWindow(xdata.display, ButtonId(ctrl->prev));
        XMapRaised(xdata.display, ButtonId(ctrl->next));
        XMapRaised(xdata.display, ButtonId(ctrl->bottom));
        break;
      case CTRL_MODE_MID:
        XMapRaised(xdata.display, ButtonId(ctrl->top));
        XMapRaised(xdata.display, ButtonId(ctrl->prev));
        XMapRaised(xdata.display, ButtonId(ctrl->next));
        XMapRaised(xdata.display, ButtonId(ctrl->bottom));
        break;
      case CTRL_MODE_BOTTOM:
        XMapRaised(xdata.display, ButtonId(ctrl->top));
        XMapRaised(xdata.display, ButtonId(ctrl->prev));
        XUnmapWindow(xdata.display, ButtonId(ctrl->next));
        XUnmapWindow(xdata.display, ButtonId(ctrl->bottom));
        break;
      case CTRL_MODE_NONE:
        XUnmapWindow(xdata.display, ButtonId(ctrl->top));
        XUnmapWindow(xdata.display, ButtonId(ctrl->prev));
        XUnmapWindow(xdata.display, ButtonId(ctrl->next));
        XUnmapWindow(xdata.display, ButtonId(ctrl->bottom));
        break;
    }
    XMapRaised(xdata.display, ctrl->win.id);
    return;
}

void ExposeLife(Life *life)
{
    int i, j;

    XClearWindow(xdata.display, life->win.id);
    for(i = 0; i < 2; i++)
    {
	XDrawString(xdata.display, life->win.id, gcset.type[7],
		    20, 32 + i * 64, name[i], strlen(name[i]));
	for( j = 0; j < life->life[i]; j++)
	    XFillRectangle(xdata.display, life->win.id, gcset.type[4 - i * 2],
			   20 + j * 20, 44 + i * 64, 18, 24);
    }

    return;
}

void MapLife(Life *life)
{
    XMapRaised(xdata.display, life->win.id);
    return;
}

void CreateLife(int x, int y, Life *life)
{
    ConstructWindata(&xdata, win.id,
                     x, y, LIFE_WIDTH, LIFE_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &life->win);
    XSetWindowBackgroundPixmap(xdata.display, life->win.id, LIFE_XPM.pixmap);
    XSelectInput(xdata.display, life->win.id, ExposureMask);

    life->life[0] = life->life[1] = 8;
}

void DrawSingleKoma(Drawable d, int x, int y, int type)
{
    if( type ) {
        XSetClipOrigin(xdata.display,
                       gcset.type[type & TuSideMask ? KOMAGC1 : KOMAGC0],
		       x, y);
	XCopyArea(xdata.display, komapm[type & TuKomaMask], d,
		  gcset.type[type & TuSideMask ? KOMAGC1 : KOMAGC0],
		  0, 0, KOMA_WIDTH, KOMA_HEIGHT, x, y);
    }
	
    return;
}

void DrawBan(Ban *ban, Pos pos, int mode)
{
    int x, y;

    XCopyArea(xdata.display, ban->pm, ban->disppm, gcset.type[0],
              0, 0, BAN_WIDTH, BAN_HEIGHT, 0, 0);

    for(y = 0; y < 9; y++) {
	for(x = 0; x < 9; x++) {
	    if(ban->data[y][x] == EMPTY)
		continue;
	    DrawSingleKoma(ban->disppm,
			   (side ? x : 8 - x)
			   * MASU_WIDTH + KOMA_OFFSET + BAN_OFFSET,
			   (side ? 8 - y : y)
			   * MASU_HEIGHT + KOMA_OFFSET + BAN_OFFSET,
			   ban->data[y][x]);
	}
    }

    if(mode && pos.x != FROM_DAI)
    {
        XDrawRectangle(xdata.display, ban->disppm, gcset.type[4],
                       ((side == SENTE) ? 8 - pos.x : pos.x)
                       * MASU_WIDTH + BAN_OFFSET,
                       ((side == SENTE) ? pos.y : 8 - pos.y)
                       * MASU_HEIGHT + BAN_OFFSET,
                       MASU_WIDTH - 1, MASU_HEIGHT - 1);
    }

    return;
}

void DrawDai(Dai *dai, Pos pos, int mode)
{
    int i, y;
    char buff[BUFFLEN];

    XCopyArea(xdata.display, dai->pm, dai->disppm, gcset.type[0],
              0, 0, DAI_WIDTH, DAI_HEIGHT, 0, 0);
    for(i = 0; i < 2; i++)
    {
        for(y = 0; y < 7; y++)
        {
            if(dai->data[i][y] == 0)
                continue;
            DrawSingleKoma(dai->disppm, KOMA_OFFSET + DAI_OFFSET,
                           y * MASU_HEIGHT + KOMA_OFFSET + DAI_OFFSET,
                           0x20 | y | (i ? TuSideMask : 0));
            if(dai->data[i][y] == 1)
                continue;
            sprintf(buff, " \xD7 %d", dai->data[i][y]);
            XDrawString(xdata.display, dai->disppm, gcset.type[7],
                        64, 52 + y * MASU_HEIGHT,
                        buff, strlen(buff));
        }
    }
    if(mode && pos.x == FROM_DAI)
    {
        XDrawRectangle(xdata.display, dai->disppm, gcset.type[4],
                       DAI_OFFSET, pos.y * MASU_HEIGHT + DAI_OFFSET,
                       MASU_WIDTH - 1, MASU_HEIGHT - 1);
    }

    return;
}

void MapBan(Ban *ban)
{
    XMapRaised(xdata.display, ban->win.id);
    return;
}

void MapDai(Dai *dai)
{
    XMapRaised(xdata.display, dai->win.id);
    return;
}

void ExposeBan(Ban *ban)
{
    XSetWindowBackgroundPixmap(xdata.display, ban->win.id, ban->disppm);
    XClearWindow(xdata.display, ban->win.id);
    return;
}

void ExposeDai(Dai *dai)
{
    XSetWindowBackgroundPixmap(xdata.display, dai->win.id, dai->disppm);
    XClearWindow(xdata.display, dai->win.id);
    return;
}

void CreateBan(int x, int y, int init_mode, Ban *ban)
{
    int suji, dan;
    static Pixmap ban_disppm = None;

    ConstructWindata(&xdata, win.id,
                     x, y, BAN_WIDTH, BAN_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &ban->win);
    ban->pm = BAN_XPM.pixmap;
    if( ban_disppm == None ) {
	ban_disppm = XCreatePixmap(xdata.display, win.id,
				   BAN_XPM.width, BAN_XPM.height, xdata.depth);
	if( ban_disppm == None ) {
	    fprintf(stderr, "%s: cannot create pixmap!\n", myname);
	    exit(1);
	}
    }
    ban->disppm = ban_disppm;
    XSetWindowBackgroundPixmap(xdata.display, ban->win.id, ban->pm);
    XSelectInput(xdata.display, ban->win.id,
		 ButtonPressMask | ExposureMask | KeyPressMask);

    for( dan = 0 ; dan < 9 ; dan++ )
	for( suji = 0 ; suji < 9; suji++ )
	    ban->data[dan][suji] = EMPTY;

    if(side == SENTE) {
	if(init_mode == MY_SIDE || init_mode == BOTH_SIDE) {
	    ban->data[8][8] = ban->data[8][0] = KYOSHA;
	    ban->data[8][7] = ban->data[8][1] = KEIMA;
    	    ban->data[8][6] = ban->data[8][2] = GIN;
	    ban->data[8][5] = ban->data[8][3] = KIN;
	    ban->data[8][4] = OU;
    	    ban->data[7][7] = KAKU;
	    ban->data[7][1] = HISHA;
	    for( suji = 0 ; suji < 9 ; suji++ )
    	        ban->data[6][suji] = FU;
	}
	if(init_mode == OPPONENT_SIDE || init_mode == BOTH_SIDE)
	{
	    ban->data[0][8] = ban->data[0][0] = KYOSHA	| TuSideMask;
    	    ban->data[0][7] = ban->data[0][1] = KEIMA	| TuSideMask;
	    ban->data[0][6] = ban->data[0][2] = GIN	| TuSideMask;
	    ban->data[0][5] = ban->data[0][3] = KIN	| TuSideMask;
    	    ban->data[0][4] = OU	| TuSideMask;
	    ban->data[1][1] = KAKU	| TuSideMask;
	    ban->data[1][7] = HISHA	| TuSideMask;
    	    for( suji = 0 ; suji < 9 ; suji++ )
		ban->data[2][suji] = FU	| TuSideMask;
	}
    }
    else
    {
	if(init_mode == MY_SIDE || init_mode == BOTH_SIDE)
	{
	    ban->data[0][8] = ban->data[0][0] = KYOSHA;
	    ban->data[0][7] = ban->data[0][1] = KEIMA;
	    ban->data[0][6] = ban->data[0][2] = GIN;
	    ban->data[0][5] = ban->data[0][3] = KIN;
	    ban->data[0][4] = OU;
	    ban->data[1][1] = KAKU;
	    ban->data[1][7] = HISHA;
    	    for( suji = 0 ; suji < 9 ; suji++ )
		ban->data[2][suji] = FU;
	}
	if(init_mode == OPPONENT_SIDE || init_mode == BOTH_SIDE)
	{
	    ban->data[8][8] = ban->data[8][0] = KYOSHA	| TuSideMask;
	    ban->data[8][7] = ban->data[8][1] = KEIMA	| TuSideMask;
    	    ban->data[8][6] = ban->data[8][2] = GIN	| TuSideMask;
	    ban->data[8][5] = ban->data[8][3] = KIN	| TuSideMask;
	    ban->data[8][4] = OU	| TuSideMask;
    	    ban->data[7][7] = KAKU	| TuSideMask;
	    ban->data[7][1] = HISHA	| TuSideMask;
	    for( suji = 0; suji < 9; suji++)
    	        ban->data[6][suji] = FU	| TuSideMask;
	}
    }
}

void CreateDai(int x, int y, Dai *dai)
{
    int i, j;
    static Pixmap dai_disppm = None;

    ConstructWindata(&xdata, win.id,
                     x, y, DAI_WIDTH, DAI_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &dai->win);
    dai->pm = DAI_XPM.pixmap;
    if( dai_disppm == None ) {
        dai_disppm = XCreatePixmap( xdata.display, win.id,
                                DAI_XPM.width, DAI_XPM.height, xdata.depth );
        if( dai_disppm == None ) {
	    fprintf(stderr, "%s: cannot create pixmap!\n", myname);
	    exit(1);
        }
    }
    dai->disppm = dai_disppm;
    XSetWindowBackgroundPixmap(xdata.display, dai->win.id, dai->pm);
    XSelectInput(xdata.display, dai->win.id, ButtonPressMask | ExposureMask);

    for(i = 0; i < 2; i++)
        for(j = 0; j < 7; j++)
            dai->data[i][j] = 0;
}
