/*
 *	XTuitate
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xdata.h"
#include "commdata.h"
#include "all.h"

static int sock_desc = 0;
static unsigned char sock_io_buf[1024];
static unsigned char *sock_io_ptr;
static unsigned int sock_io_cnt;

int setclient(char *hostname, u_short port)
{
    struct hostent *servhost;
    struct sockaddr_in server;
    int s;
    char *private_host;

    private_host = (hostname != NULL) ? hostname : "localhost";

    while( (servhost = gethostbyname(private_host)) == NULL ) {
	switch( h_errno ) {
	case HOST_NOT_FOUND:
	    fprintf(stderr, "Error: `%s' is unknown.\n", hostname);
	    return -1;
	case NO_DATA:
	    fprintf(stderr,
		    "Error: `%s' is valid but does not have an IP address.\n",
		    hostname);
	    return -1;
	case NO_RECOVERY:
	    fprintf(stderr,
		    "Error: A non-recoverable name server error occurred.\n");
	    return -1;
	case TRY_AGAIN:
	    break;
	default:
	    fprintf(stderr,
		    "Fatal Error: Undefined error occured, h_errno = %d\n",
		    h_errno);
	    exit(1);
	}
	fprintf(stderr,
		"A temporary error occurred on an authoritative name "
		"server. Retrying...\n");
	sleep(1);
    }

    bzero((char *)&server, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_port = htons(port);
    bcopy(servhost->h_addr, (char *)&server.sin_addr, servhost->h_length);

    if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
	perror("Socket allocation");
	return CONNECT_SocketError;
    }

    if( DebuggingNetwork ) {
	int i;
	fprintf(stderr, "Trying %s(%d",
		servhost->h_name, (unsigned char)servhost->h_addr_list[0][0]);
	for( i = 1 ; i < servhost->h_length ; i++ )
	    fprintf(stderr, ".%d", (unsigned char)servhost->h_addr_list[0][i]);
	fprintf(stderr, "):%u ...\n", port);
    }

    if( connect(s, (struct sockaddr *)&server, sizeof(server)) ) {
	perror("CONNECT");
	return CONNECT_SocketError;
    }
    if( DebuggingNetwork )
	fprintf(stderr, "Connection established.\n");

    sock_io_ptr = sock_io_buf;
    sock_io_cnt = 0;
    sock_desc = s;

    return s;
}

int setup_connection(const char *host_name, const char *service_name)
{
    struct hostent *server_host;
    struct servent *service;
    struct sockaddr_in server;
    int s;
    int port_no;
    char *ptr;
#ifdef	__GNUC__
    char host_name_real[strlen(host_name) + 1];
#else
    char * host_name_real;

    ALLOCA(host_name_real, strlen(host_name) + 1);
#endif

    for( ptr = host_name_real ; ; ) {
	if( *host_name == ':' ) {
	    *ptr++ = '\0';
	    if( service_name )
		break;
	    service_name = ++host_name;
	}
	else if( !(*ptr++ = *host_name++) )
	    break;
    }

    if( DebuggingNetwork )
	fprintf(stderr, "Server: `%s', port: %s\n",
		host_name_real, service_name);

    while( (server_host = gethostbyname(host_name_real)) == NULL ) {
	switch( h_errno ) {
	case HOST_NOT_FOUND:
	    fprintf(stderr, "Error: `%s' is unknown.\n", host_name_real);
	    return CONNECT_HostNameLookupFailue;
	case NO_DATA:
	    fprintf(stderr,
		    "Error: `%s' is valid but does not have an IP address.\n",
		    host_name);
	    return CONNECT_HostNameLookupFailue;
	case NO_RECOVERY:
	    fprintf(stderr,
		    "Error: A non-recoverable name server error occurred.\n");
	    return CONNECT_HostNameLookupFailue;
	case TRY_AGAIN:
	    break;
	default:
	    fprintf(stderr,
		    "Fatal Error: Undefined error occured, h_errno = %d\n",
		    h_errno);
	    exit(1);
	}
	fprintf(stderr,
		"A temporary error occurred on an authoritative name "
		"server. Retrying...\n");
	sleep(1);
    }
    if( service_name ) {
	while( *service_name == ' ' ) service_name++;
	if( !*service_name )
	    return CONNECT_NoServiceSpecified;
	if( isdigit(*service_name) )
	    port_no = htons(atoi(service_name));
	else {
	    if( (service = getservbyname(service_name, "tcp")) == NULL )
		return CONNECT_ServiceLookupFailue;
	    port_no = service->s_port;
	}
    }
    else
	port_no = htons(9070);

    bzero((char *)&server, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_port = port_no;
    bcopy(server_host->h_addr, (char *)&server.sin_addr, server_host->h_length);

    if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
	perror("Socket allocation");
	return -1;
    }

    if( DebuggingNetwork ) {
	int i;
	fprintf(stderr, "Trying %s(%d",
		server_host->h_name,
		(unsigned char)server_host->h_addr_list[0][0]);
	for( i = 1 ; i < server_host->h_length ; i++ )
	    fprintf(stderr, ".%d",
		    (unsigned char)server_host->h_addr_list[0][i]);
	fprintf(stderr, "):%u ...\n", port_no);
    }

    if( connect(s, (struct sockaddr *)&server, sizeof(server)) ) {
	perror("CONNECT");
	return -1;
    }
    if( DebuggingNetwork )
	fprintf(stderr, "Connection established.\n");

    sock_io_ptr = sock_io_buf;
    sock_io_cnt = 0;
    sock_desc = s;

    return s;
}


int sock_getc(void)
{
    if( !sock_io_cnt ) {
	int ret;

	ret = read(sock_desc, sock_io_buf, sizeof(sock_io_buf));
	if( ret == 0 ) {
	    fprintf(stderr, "Connection closed by TUITATE server.\n");
	    exit(1);
	}
	else if( ret == -1 ) {
	    perror("Socket read");
	    exit(1);
	}

	sock_io_cnt = ret;
	sock_io_ptr = sock_io_buf;
    }
    sock_io_cnt--;
    return *sock_io_ptr++;
}

int SendData(char *buff, int len)
{
    int wrote_size;
    if( DebuggingProtocol )
	fprintf(stderr, "Sending:   %s", buff);
    if( (wrote_size = write(sock_desc, buff, len)) < 0 ) {
	perror("Socket write");
	exit(1);
    }
    return wrote_size;
}

int RecvData(char *buff, int len)
{
    char *ptr = buff;
    char *endp = buff + len;

    if( DebuggingProtocol )
	fprintf(stderr, "Receiving: ");
    while( 1 ) {
	char c;

	if( ptr >= endp ) {			/* 1ԤĹ᤮		*/
	    fprintf(stderr, "Line too long.\n");
	    exit(1);
	}

	c = sock_getc();
	if( DebuggingProtocol )
	    fputc(c, stderr);
	if( c == '\n' ) break;
	*ptr++ = c;
    }
    
    *ptr = '\0';

    return ptr - buff;
}

int WaitComm(void)
{
    int nfds;
    fd_set canread;

    /*  X 饤٥ȤƤ */
    if( XEventsQueued(xdata.display, QueuedAfterReading) )
	return FROM_DISPLAY;
    XFlush(xdata.display);

    FD_ZERO(&canread);
    FD_SET(sock_desc, &canread);
    FD_SET(ConnectionNumber(xdata.display), &canread);

    nfds = ConnectionNumber(xdata.display);
    if( nfds < sock_desc )
	nfds = sock_desc;
    nfds++;

    if( select(nfds, &canread, NULL, NULL, NULL) < 0 ) {
	perror("SELECT");
	exit(1);
    }
    if( FD_ISSET(sock_desc, &canread) )
	return FROM_SERVER;	/* ĤƥФå褿 	*/
    else /*  if( FD_ISSET(ConnectionNumber(xdata.display), &canread) )*/
	return FROM_DISPLAY;	/* X 饤٥Ȥ褿		*/
}

int IsInclude(char **wordbuff, const char *str, int start, int end)
{
	int ret = 0;
	int i;
	for(i = start; i <= end; i++)
		if(strcmp(wordbuff[i], str) == 0)
			ret++;
	return ret;
}

int splitstr(char *buff, char **wordbuff, int maxword, const char delimiter)
{
    int count = 0;
    char *src_ptr = buff;
	
    while( *src_ptr ) {
	while( *src_ptr == delimiter )
	    src_ptr++;
	if( !*src_ptr )
	    break;
	wordbuff[count++] = src_ptr;
	while( *src_ptr && *src_ptr != delimiter )
	    src_ptr++;
	if( !*src_ptr )
	    break;
	*src_ptr++ = '\0';
	if( count == maxword )
	    break;
    }
    return count;
}

int CutStr(char *buff, char **minibuff, char token)
{
        int ct, minict, city;

        ct = 0;
        minict = 0;
        city = 0;
        
        while(1)
        {
                if(buff[ct] == token)
                {
                        minibuff[city][minict] = '\0';
                        ct++;
                        minict = 0;
                        city++;
                }
                if(buff[ct] == '\0')
                {
                        minibuff[city][minict] = '\0';
                        break;
                }

                minibuff[city][minict] = buff[ct];
                ct++;
                minict++;
        }

        return city + 1;
}

int token_equal(char **str, const char *token)
{
    char *ptr = *str;
    char *tokenp = (char *)token;

    while( *ptr == ' ' || *ptr == '\t' ) ptr++;

    while( *tokenp ) {
	if( *ptr != *tokenp ) return 0;
	ptr++;
	tokenp++;
    }
    if( *ptr && *ptr != ' ' && *ptr != '\t' ) return 0;

    while( *ptr == ' ' || *ptr == '\t' ) ptr++;
    *str = ptr;

    return 1;
}

int fetch_numerical_token(char **str)
{
    char *ptr = *str;
    int val = 0;

    while( *ptr == ' ' || *ptr == '\t' ) ptr++;

    do {
	if( *ptr < '0' || '9' < *ptr )
	    return -1;
	val= val * 10 + (*ptr++ - '0');
    } while( *ptr && *ptr != ' ' && *ptr != '\t' );

    while( *ptr == ' ' || *ptr == '\t' ) ptr++;

    *str = ptr;

    return val;
}

void send_quit(void)
{
    SendData("QUIT\n", 5);
}

void closeConnection()
{
    if( sock_desc )
	close(sock_desc);
}

static char * itoa_sub(char *str, unsigned int n)
{
    unsigned int r;

    if( (r = n / 10) )
	str = itoa_sub(str, r);
    *str++ = '0' + n % 10;
    return str;
}

static void sendPrintf(const char * fmt, ...)
{
    char buffer[TUP_LINE_MAX];
    char numbuf[12];
    char *bufptr = buffer;
    char *p;
    size_t left = sizeof(buffer);
    char c;
    size_t write_size;
    va_list ap;

    va_start(ap, fmt);
    while( (c = *fmt++) ) {
	if( c == '%' ) {
	    switch( (c = *fmt++) ) {
	    case 'd':
		*itoa_sub(numbuf, va_arg(ap, int)) = '\0';
		p = numbuf;
		goto CopyString;
	    case 's':
		p = va_arg(ap, char *);
		if( p == NULL )
		    p = "(null)";
	    CopyString:
		while( *p ) {
		    if( !--left )
			goto LineTooLong;
		    *bufptr++ = *p++;
		}
		break;
	    default:
		goto OneCharacter;
	    }
	}
	else {
	OneCharacter:
	    if( !--left ) {
	    LineTooLong:
		fprintf(stderr, "Error: Line too long: \"%.*s\"\n",
			(int)sizeof(buffer), buffer);
		exit(1);
	    }
	    *bufptr++ = c;
	}
    }
#ifdef	DEBUG
    if( DebuggingProtocol ) {
	*bufptr = '\0';
	fprintf(stderr, "Sending:   %s\n", buffer);
    }
#endif
    *bufptr = '\n';

    write_size = sizeof(buffer) - (left - 1);
    if( write(sock_desc, buffer, write_size) < write_size ) {
	perror("Socket write");
	exit(1);
    }
}

#if 0
/*
 * ΥꥹȤɤ߹
 */
TuNotationList * ReadNotationList(max, nentries_return)
    unsigned int max;			/* ȥκ */
    unsigned int *nentries_return;	/* ȥ֤ */
{
    char buff[BUFFLEN];
    int i;
    unsigned int nentries;
    unsigned int discards;
    TuNotationList *list, *ptr;

    SendData("NOTATIONLIST\n", 13);
    RecvData(buff, sizeof(buff));
    sscanf(buff, "NOTATIONLIST %d", &nentries);
    if( DebuggingProtocol )
	fprintf(stderr, "(%d notation entries)\n", nentries);

    discards = (nentries > max) ? nentries - max : 0;

    list = xmalloc(sizeof(TuNotationList) * (nentries > max ? max : nentries));
    ptr = list;
    for( i = 0 ; i < nentries ; i++ ) {
	int junk;

	RecvData(buff, sizeof(buff));
	if( buff[0] == '.' ) {
	    fprintf(stderr, "Notation-list abnormally terminated\n");
	    exit(1);
	}
        sscanf(buff, "%d %d %d", &junk, &ptr->begin, &ptr->end);
        RecvData(ptr->name_address, sizeof(ptr->name_address));
	{
	    char *p = ptr->name_address;
	    while( *p ) {
		if( *p == ' ' ) *p = '\0';
		p++;
	    }
	}
	discards ? discards-- : ptr++;
    }
    if( nentries_return != NULL ) *nentries_return = nentries;
    return list;
}
#endif
