
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>

#include "drawstring.h"
#include "all.h"

/* ե */
#ifndef	ASCFONT1
#define	ASCFONT1 "-adobe-times-bold-r-normal-*-*-180-*-*-p-*-iso8859-*"
#endif
#ifndef	ASCFONT2
#define	ASCFONT2 "-adobe-times-bold-r-normal-*-*-240-*-*-p-*-iso8859-*"
#endif
#ifndef	KNJFONT1
#define	KNJFONT1 "kanji16"
#endif
#ifndef	KNJFONT2
#define	KNJFONT2 "kanji24"
#endif

#define	XCHAR2B_BUFSIZE	256

#ifdef	__CHAR_UNSIGNED__
#define	CHAR_MSB(x) ((x) & 0x80)
#else
#define	CHAR_MSB(x) ((x) < 0)
#endif

#define	FirstCodeOf(x) ((x)[0])

static XFontStruct *ascfont[2] = { NULL, NULL };
static XFontStruct *knjfont[2] = { NULL, NULL };
static Pixmap sente_bitmap;
static Pixmap gote_bitmap;

#include "bitmaps/sente.xbm"
#include "bitmaps/gote.xbm"

/* եȤν򤹤 */

static XFontStruct *XtuitateLoadQueryFont(Display *disp, const char *fontname)
{
    XFontStruct *font_struct;

    ifDebug('X')
	fprintf(stderr, "Loading font \"%s\" ... ", fontname);
    if( (font_struct = XLoadQueryFont(disp, fontname)) == NULL ) {
	ifDebug('X')
	    fprintf(stderr, "Failed!\n");
	fprintf(stderr, "Cannot load font \"%s\"\n", fontname);
	exit(1);
    }
    ifDebug('X')
	fprintf(stderr, "Done.\n");

    return font_struct;
}

void XtuitateLoadFonts(Display *disp)
{
    if( ascfont[0] == NULL )
	ascfont[0] = XtuitateLoadQueryFont(disp, ASCFONT1);
    if( ascfont[1] == NULL )
	ascfont[1] = XtuitateLoadQueryFont(disp, ASCFONT1);
    if( knjfont[0] == NULL )
	knjfont[0] = XtuitateLoadQueryFont(disp, KNJFONT1);
    if( knjfont[1] == NULL )
	knjfont[1] = XtuitateLoadQueryFont(disp, KNJFONT2);
}

static Pixmap MyXCreateBitmapFromData(disp, drawable, data, width, height)
    Display *disp;
    Drawable drawable;
    unsigned char data[ ];
    unsigned int width, height;
{
    Pixmap bitmap;

    bitmap = XCreateBitmapFromData(disp, drawable, data, width, height);
    if( bitmap == None ) {
	fprintf(stderr, "%s: No enough memory to create bitmap.\n", myname);
	exit(1);
    }
    return bitmap;
}

void XtuitateInitializeFonts(disp)
    Display *disp;
{
    XtuitateLoadFonts(disp);
    sente_bitmap = MyXCreateBitmapFromData(disp, DefaultRootWindow(disp),
					   sente_bits,
					   sente_width, sente_height);
    gote_bitmap  = MyXCreateBitmapFromData(disp, DefaultRootWindow(disp),
					   gote_bits,
					   gote_width, gote_height);
}

int XtuitateDrawEUCString(Display *disp, Drawable win,
			  GC gc, int x, int y, char *str)
{
    int cur_afont = 0;
    int cur_kfont = 0;
    XChar2b buf[XCHAR2B_BUFSIZE];	/* 2ХʸΥХåե	*/

    /* եȤɤƤʤХɤ */
    if( ascfont[0] == NULL || knjfont[0] == NULL )
	XtuitateLoadFonts(disp);

    while( *str ) {
	if( CHAR_MSB(*str) ) {
	    XSetFont(disp, gc, knjfont[cur_kfont]->fid);
	    do {
		int index = 0;	/* buf[ ] Υǥå */
		do {
		    buf[index].byte1 = *str++ & 0x7F;
		    buf[index].byte2 = *str++ & 0x7F;
		    index++;
		} while( index < XCHAR2B_BUFSIZE && *str && CHAR_MSB(*str) );
		XDrawString16(disp, win, gc, x, y, buf, index);
		x += XTextWidth16(knjfont[cur_kfont], buf, index);
	    } while( *str && CHAR_MSB(*str) );
	}
	else if( *str >= ' ' ) {
	    char *head = str;	/* ASCIIʸƬ¸	*/
	    int count;		/* ʸ򥫥Ȥ		*/
	    for( count = 1 ; *++str && *str >= ' ' && !CHAR_MSB(*str) ;
		 count++ ) ;
	    XSetFont(disp, gc, ascfont[cur_afont]->fid);
	    XDrawString(disp, win, gc, x, y, head, count);
	    x += XTextWidth(ascfont[cur_afont], head, count);
	}
	else switch( *str++ ) {
	    Pixmap mask;
	case CodeA1:
	    cur_afont = 0;
	    break;
	case CodeA2:
	    cur_afont = 1;
	    break;
	case CodeK1:
	    cur_kfont = 0;
	    break;
	case CodeK2:
	    cur_kfont = 1;
	    break;
	case CodeSente:
	    mask = sente_bitmap;
	    goto put_bitmap;
	case CodeGote:
	    mask = gote_bitmap;
	put_bitmap:
	    XSetClipMask(disp, gc, mask);
	    XSetClipOrigin(disp, gc, x, y - knjfont[0]->ascent);
	    XFillRectangle(disp, win, gc, x, y - knjfont[0]->ascent, 16, 16);
	    x += 16;
	    XSetClipMask(disp, gc, None);
	    break;
	default:
	    fprintf(stderr, "Internal error: Undefined control sequence\n");
	    exit(1);
	}	    
    }
    return x;
}

int XtuitateEUCTextWidth(char *str)
{
    int cur_afont = 0;
    int cur_kfont = 0;
    int width = 0;
    XChar2b buf[XCHAR2B_BUFSIZE];	/* 2ХʸΥХåե	*/

    while( *str ) {
	if( CHAR_MSB(*str) ) {
	    do {
		int index = 0;	/* buf[ ] Υǥå */
		do {
		    buf[index].byte1 = *str++ & 0x7F;
		    buf[index].byte2 = *str++ & 0x7F;
		    index++;
		} while( index < XCHAR2B_BUFSIZE && *str && CHAR_MSB(*str) );
		width += XTextWidth16(knjfont[cur_kfont], buf, index);
	    } while( *str && CHAR_MSB(*str) );
	}
	else if( *str >= ' ' ) {
	    char *head = str;	/* ASCIIʸƬ¸	*/
	    int count;		/* ʸ򥫥Ȥ		*/
	    for( count = 1 ; *++str && !CHAR_MSB(*str) ; count++ ) ;
	    width += XTextWidth(ascfont[cur_afont], head, count);
	}
	else switch( *str++ ) {
	case CodeA1:
	    cur_afont = 0;
	    break;
	case CodeA2:
	    cur_afont = 1;
	    break;
	case CodeK1:
	    cur_kfont = 0;
	    break;
	case CodeK2:
	    cur_kfont = 1;
	    break;
	default:
	    fprintf(stderr, "Internal error: Undefined control sequence\n");
	    exit(1);
	}	    
    }
    return width;
}

void XtuitateDrawEUCStringFlushRight(Display *disp, Drawable win,
				     GC gc, int x, int y, char *str)
{
    x -= XtuitateEUCTextWidth(str);
    XtuitateDrawEUCString(disp, win, gc, x, y, str);
}

#if 0
int XtuitateDrawEUCStringFolding(Display *disp, Drawable win,
				 GC gc, int x, int y,
				 int x_max, int line_step, char *str)
{
    int cur_afont = 0;
    int cur_kfont = 0;
    char *head;
    int temp_x;
    XChar2b buf;	/* 2ХʸΥХåե	*/

    /* եȤɤƤʤХɤ */
    if( ascfont[0] == NULL || knjfont[0] == NULL )
	XtuitateLoadFonts(disp);

    head = str;
    temp_x = x;
    while( *str ) {
	if( CHAR_MSB(*str) ) {
	    buf.byte1 = *str++ & 0x7F;
	    buf.byte2 = *str++ & 0x7F;
	    temp_x += XTextWidth16(knjfont[cur_kfont], &buf, 1);
	}
	else if( *str >= ' ' ) {
	    char *head = str;	/* ASCIIʸƬ¸	*/
	    int count = 1;	/* ʸ򥫥Ȥ		*/

	    if( *str == ' ' ) {
		if( )
	 
	    while( *str != ' ' && *str != )
		
	    for( count = 1 ; *++str && !CHAR_MSB(*str) ; count++ ) ;
	    XSetFont(disp, gc, ascfont[cur_afont]->fid);
	    XDrawString(disp, win, gc, x, y, head, count);
	    x += XTextWidth(ascfont[cur_afont], head, count);
	    }
	}
	else switch( *str++ ) {
	case CodeA1:
	    cur_afont = 0;
	    break;
	case CodeA2:
	    cur_afont = 1;
	    break;
	case CodeK1:
	    cur_kfont = 0;
	    break;
	case CodeK2:
	    cur_kfont = 1;
	    break;
	LINEBREAK:
	case '\n':
	    XtuitateDrawEUCString(disp, win, gc, x, y, head);
	    head = str;
	    temp_x = x;
	    break;
	default:
	    fprintf(stderr, "Internal error: Undefined control sequence\n");
	    exit(1);
	}	    
    }
    return x;
}
#endif
