#include <X11/Xlib.h>
#include "eventproc.h"

ProcList proclistarea[ NPROCLIST ];

static unsigned char init_flag = 0;

static ProcList *free_link;

ProcList *add_proc(ProcList **list, const ProcFunc func)
{
    ProcList *allocated_cell;

    if( !init_flag ) {
	int i;

	for( i = 0 ; i < NPROCLIST - 1 ; i++ )
	    proclistarea[i].next = &proclistarea[i + 1];
	proclistarea[NPROCLIST - 1].next = NULL;
	free_link = proclistarea;
	init_flag = 1;
    }

    if( (allocated_cell = free_link) == NULL ) {
	fprintf(stderr, "FATAL: No ProcList to allocate\n");
	exit(1);
    }
    free_link = free_link->next;

    allocated_cell->next = *list;
    allocated_cell->func = func;
    *list = allocated_cell;

    return allocated_cell;
}

void delete_proc(ProcList **listp, ProcList *delete)
{
    ProcList *ptr;

    for( ptr = (ProcList *)listp ; ptr->next != delete ; ptr = ptr->next ) {
	if( ptr->next == NULL ) {
	    fprintf(stderr, "FATAL: delete_proc: Specified Proc not found\n");
	    exit(1);
	}
    }
    ptr->next = delete->next;
}

int process_proc(const ProcList *list, XEvent *event)
{
    ProcList *ptr = list;
    int ret;

    while( ptr != NULL ) {
	if( (ret = (ptr->func)(event)) )
	    return ret;
	ptr = ptr->next;
    }
    return 0;
}
