#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/xpm.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "commdata.h"
#include "all.h"
#include "game.h"
#include "ban.h"
#include "canmove.h"
#include "drawstring.h"
#include "eventproc.h"
#include "cursor.h"
#include "tuitate.h"
#include "strings.h"
#include "misc.h"
#include "pixmaps.h"

#define BAN_X			16
#define BAN_Y			16

#define DAI_X			(BAN_X + BAN_WIDTH + 16)
#define DAI_Y			(BAN_HEIGHT - DAI_HEIGHT + 16)

#define LIFE_X			BAN_X
#define LIFE_Y			(BAN_Y + BAN_HEIGHT + 16)

#define MESG_X			(LIFE_X + LIFE_WIDTH + 16)
#define MESG_Y			LIFE_Y

#define GIVEUP_WIDTH		128
#define GIVEUP_HEIGHT		64
#define GIVEUP_X		(MESG_X + MESG_WIDTH + 16)
#define GIVEUP_Y		(MESG_Y + MESG_HEIGHT - GIVEUP_HEIGHT)

#define QUIT_WIDTH		128
#define QUIT_HEIGHT		64
#define QUIT_X			GIVEUP_X
#define QUIT_Y			GIVEUP_Y

#define VIEWNOTE_WIDTH		128
#define VIEWNOTE_HEIGHT		64
#define VIEWNOTE_X		GIVEUP_X
#define VIEWNOTE_Y		(GIVEUP_Y - VIEWNOTE_HEIGHT - 16)

#define ASKWIN_X		((WIN_WIDTH - ASKWIN_WIDTH) / 2)
#define ASKWIN_Y		((WIN_HEIGHT - ASKWIN_HEIGHT) / 2)

#define NORMAL			0			/* for phase */
#define GRIP			1
#define ASKDROP			2
#define ASKPROMOTE		3
#define ENEMY			4
#define ASKGIVEUP		5

static void DisplayProc(void);
static void ServerProc(void);
static void ButtonPressProc(XEvent *ev);
static void RightButtonPressProc(XEvent ev);
static void MotionNotifyProc(XEvent ev);
static void CrossNotifyProc(XEvent ev, int mode);
static void ExposeProc(XEvent ev);
static void SendMove(int promotion);
static int DropKoma(int px, int py);
static int GripBanKoma(int px, int py);
static int GripDaiKoma(int px, int py);
static void KeyPressProc(XKeyEvent *);

void PromoteKoma(unsigned char *type);
char *KomaName(int type);

char **MallocWordbuff(void);		/* in selectenemy.c */
void FreeWordbuff(char **wordbuff);

static Ban ban;
static Dai dai;
static Life life;
static Mesg mesg;
static Button giveup, viewnote, quit;
static Askwin askgiveup, askdrop, askpromote;
static Pos pastpos, nowpos;
static int phase, prev_phase;
static int done;
static int promotemode;

int Game(void)
{
    XEvent ev;
    char buff[BUFFLEN];

    /* create windows */
    CreateBan(BAN_X, BAN_Y, MY_SIDE, &ban);
    CreateDai(DAI_X, DAI_Y, &dai);
    CreateLife(LIFE_X, LIFE_Y, &life);
    CreateMesg(&xdata, &win, MESG_X, MESG_Y, &mesg);
    CreateButton(&xdata, &win, GIVEUP_XPM.pixmap, GIVEUP_F_XPM.pixmap,
                 GIVEUP_X, GIVEUP_Y, GIVEUP_WIDTH, GIVEUP_HEIGHT, 1, &giveup);
    CreateButton(&xdata, &win, VIEWNOTE_XPM.pixmap, VIEWNOTE_F_XPM.pixmap,
                 VIEWNOTE_X, VIEWNOTE_Y, VIEWNOTE_WIDTH, VIEWNOTE_HEIGHT, 1,
                 &viewnote);
    CreateButton(&xdata, &win, QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, QUIT_Y, QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);
    CreateAskwin(&xdata, &win, "Give up?", ASKWIN_X, ASKWIN_Y, &askgiveup);
    CreateAskwin(&xdata, &win, "Ok?", ASKWIN_X, ASKWIN_Y, &askdrop);
    CreateAskwin(&xdata, &win, "Promote?", ASKWIN_X, ASKWIN_Y, &askpromote);

    /* map window */
    MapBan(&ban);
    MapDai(&dai);
    MapLife(&life);
    MapMesg(&xdata, &mesg);
    MapButton(&xdata, &giveup);

    {
	char * winname = format("xtuitate: %s versus %s", name[0], name[1]);
	XStoreName(xdata.display, win.id, winname);
    }

    /* init data */
    InitMesgText(&mesg);
    if(side == SENTE)
    {
        phase = NORMAL;
#if 0
        SetMesgText(&mesg, 0, "Your Turn.", 7);
#else
	SetMesgText(&mesg, 0, FntK2 "ʤ֤Ǥ", 7);
#endif
	SetCursor(ban.win.id, CSR_NORMAL);
	SetCursor(dai.win.id, CSR_NORMAL);
        if(bell)
            XBell(xdata.display, 0);
	if( auto_raise )
	    XRaiseWindow(xdata.display, win.id);
    }
    else if(side == GOTE)
    {
        phase = ENEMY;
#if 0
	sprintf(buff, "Please wait %s.", name[1]);
#else
	sprintf(buff, "%s αԤäƤޤ", name[1]);
#endif
	SetMesgText(&mesg, 0, buff, 7);
	SetCursor(ban.win.id, CSR_WAITING);
	SetCursor(dai.win.id, CSR_WAITING);
    }
    else
    {
        fprintf(stderr, "Game() : side = %d is invalid.\n", side);
        exit(1);
    }

    /* main loop */
    done = GAME_DONE_NORMAL;
    while(done == GAME_DONE_NORMAL)
    {
        XFlush(xdata.display);
        switch(WaitComm())
        {
          case FROM_DISPLAY:
            DisplayProc();
            break;
          case FROM_SERVER:
            ServerProc();
            break;
          default:
            fprintf(stderr, "FROM_OTHER?\n");
            exit(0);
            break;
        }
    }

    UnmapButton(&xdata, &giveup);
    MapButton(&xdata, &viewnote);
    MapButton(&xdata, &quit);
    phase = ENEMY;
    done = GAME_DONE_NORMAL;
    while(done == GAME_DONE_NORMAL)
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            ButtonPressProc(&ev);
            break;
          case EnterNotify:
            CrossNotifyProc(ev, BUTTON_ENTER);
            break;
          case LeaveNotify:
            CrossNotifyProc(ev, BUTTON_LEAVE);
            break;
          case Expose:
            ExposeProc(ev);
            break;
        }
    }

    XDestroyWindow(xdata.display, ban.win.id);
    XDestroyWindow(xdata.display, dai.win.id);
    XDestroyWindow(xdata.display, life.win.id);
    XDestroyWindow(xdata.display, mesg.win.id);
    XDestroyWindow(xdata.display, ButtonId(giveup));
    XDestroyWindow(xdata.display, ButtonId(viewnote));
    XDestroyWindow(xdata.display, ButtonId(quit));
    DestroyAskwin(&xdata, &askgiveup);
    DestroyAskwin(&xdata, &askdrop);
    DestroyAskwin(&xdata, &askpromote);

    return done;
}

/* from display */
static void DisplayProc(void)
{
    XEvent ev;
	
    while(XPending(xdata.display))
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
	case ButtonPress:
            switch(ev.xbutton.button)
            {
	    case 1:
                ButtonPressProc(&ev);
                break;
	    case 3:
                RightButtonPressProc(ev);
                break;
            }
            break;
	case MotionNotify:
            MotionNotifyProc(ev);
            break;
	case EnterNotify:
            CrossNotifyProc(ev, BUTTON_ENTER);
            break;
	case LeaveNotify:
            CrossNotifyProc(ev, BUTTON_LEAVE);
            break;
	case Expose:
            ExposeProc(ev);
            break;
	case MappingNotify:
	    XRefreshKeyboardMapping(&ev.xmapping);
	    break;
	case KeyPress:
	    KeyPressProc(&ev.xkey);
	    break;
        }
    }
    return;
}

/* from server */
static void ServerProc(void)
{
    char buff[BUFFLEN];
    int cy;
    int x, y;
    int i;
    char *ptr = buff;
    static Bool ohte_flag = False;
    char *str;
    TuToken token;
    
    struct reason_struct { char *win, *lose, *draw; } *reason;

    static struct reason_struct reason_checkmate = {
	"ͤ", "ͤ", "???"
    };
    static struct reason_struct reason_loselife = {
	"ȿ§9ãޤ", "ʬȿ§9ãޤ", "???"
    };
    static struct reason_struct reason_giveup = {
	"꤬λޤ", "λޤ", "???"
    };

    RecvData(buff, sizeof(buff));

    InitMesgText(&mesg);
    cy = 0;
    switch( TuParseToken(&ptr) ) {
    case TEND:					/* END */
	token = TuParseToken(&ptr);
	switch( TuParseToken(&ptr) ) {
	case TEMERGENCY:
	    SetMesgText(&mesg, cy++, "ȥФȤ", 7);
	    SetMesgText(&mesg, cy++, "³ڤޤ", 7);
	    SetMesgText(&mesg, cy++, "(жɤ³ԤǤޤ)", 7);
	    break;
	case TCHECKMATE:
	    reason = &reason_checkmate;
	    goto setmessage;
	case TLOSELIFE:
	    reason = &reason_loselife;
	    goto setmessage;
	case TGIVEUP:
	    reason = &reason_giveup;
	    goto setmessage;
	setmessage:
	    switch( token ) {
	    case TWIN:
		SetMesgText(&mesg, cy++, reason->win, 4);
		SetMesgText(&mesg, cy++, "ʤξǤ", 4);
		break;
	    case TLOSE:
		SetMesgText(&mesg, cy++, reason->lose, 2);
		SetMesgText(&mesg, cy++, "ʤ餱Ǥ", 2);
		break;
	    case TDRAW:
		SetMesgText(&mesg, cy++, reason->draw, 7);
		SetMesgText(&mesg, cy++, "ʬޤ", 7);
		break;
	    default:
		fprintf(stderr, "Undefined message: %s\n", buff);
		exit(1);
	    }
	}
	TuInitializeNotation(&notation);
	TuGetNotation(fetch_numerical_token(&ptr), &notation);

        done = GAME_DONE_VIEWNOTE;
	break;

    case TOK:					/* OK */
        phase = ENEMY;
	token = TuParseToken(&ptr);
	if( (i = TuTokenToKoma(token)) ) {
	    SetMesgTextF(&mesg, cy++, 4, FntK2 "%s" FntK1 " ޤ",
			 koma_abbrev[i & TuRawKomaMask]);
	    dai.data[0][i & TuRawKomaMask]++;
	    DrawDai(&dai, pastpos, phase == GRIP);
	    ExposeDai(&dai);
	    token = TuParseToken(&ptr);
	}
	if( token == TCHECK )
	    SetMesgText(&mesg, cy++, FntK2 "" FntK1 "ؤޤ", 4);

#if 0
        sprintf(text, "Please wait %s.", name[1]);
        SetMesgText(&mesg, cy++, text, 7);
#else
        SetMesgTextF(&mesg, cy++, 7, "%s ͤƤޤ...", name[1]);
	SetCursor(ban.win.id, CSR_WAITING);
	SetCursor(dai.win.id, CSR_WAITING);
#endif
	break;

    case TILLEGAL:				/* ILLEGAL */
        fprintf(stderr, "ILLEGAL!\n");
        exit(1);

    case TVIOLATE:				/* VIOLATE */
#if 0
        sprintf(text, "Type : %s", wordbuff[1]);
        SetMesgText(&mesg, cy++, "You made Violate.", 2);
        SetMesgText(&mesg, cy++, text, 2);
#else
	switch( TuParseToken(&ptr) ) {
	case TJUMPOVER:
	    str = "ζӱۤƤϤޤ";
	    break;
	case TCHECK:
	    str = ohte_flag ?
		"Ǥޤ" :
		    "鲦ˤʤ褦ʼϻؤޤ";
	    break;
	case TDROPON:
	    str = "ζξǤäƤϤޤ";
	    break;
	case TCHECKMATEBYDROPPINGPAWN:
	    str = "Ǥͤ";
	    break;
	default:
	    str = "???????";
	}

	SetMesgText(&mesg, cy++, FntK2 "ȿ§Ȥޤ", 2);
	SetMesgTextF(&mesg, cy++, 2, "(%s)", str);

#endif
        life.life[0]--;
        ban.data[nowpos.y][nowpos.x] = EMPTY;
        if(pastpos.x != FROM_DAI)
            ban.data[pastpos.y][pastpos.x] = pastpos.type;
        else
            dai.data[0][pastpos.y]++;
        DrawBan(&ban, pastpos, phase == GRIP);
        ExposeBan(&ban);
        DrawDai(&dai, pastpos, phase == GRIP);
        ExposeDai(&dai);
        ExposeLife(&life);
	break;

    case TNEXT:					/* NEXT */
        phase = NORMAL;
	token = TuParseToken(&ptr);		/* NEXT @ [x y] [CHECKED] */
	if( token == -1 ) {
	    x = token_value;
	    if( TuParseToken(&ptr) != -1 ) {	/* NEXT x @ y [CHECKED]	  */
		fprintf(stderr, "Can't parse: %s\n", buff);
		exit(1);
	    }
	    y = token_value;
            if( x < 0 || y < 0 ) {
		fprintf(stderr, "Cannot parse: %s\n", buff);
		exit( 1 );
	    }
	    ifDebug('P')
		fprintf(stderr, "Parsed: NEXT %d %d\n", x, y);

            x--;
            y--;
#if 0
            sprintf(text, "You lost %s.", komaname[ban.data[y][x]]);
            SetMesgText(&mesg, cy++, text, 2);
#else
	    SetMesgTextF(&mesg, cy++, 2,
			 FntK2 "%s%s%s" FntK1 " 򼺤ޤ",
			 jissuuji[x + 1], kansuuji[y + 1],
			 koma_abbrev[ban.data[y][x] & TuRawKomaMask]);
#endif
	    ban.data[y][x] = EMPTY;
	    DrawBan(&ban, pastpos, phase == GRIP);
            ExposeBan(&ban);
	    DrawDai(&dai, pastpos, phase == GRIP);
   	    ExposeDai(&dai);
	    token = TuParseToken(&ptr);		/* NEXT x y @ [CHECKED]	 */
        }
        if( token == TCHECKED ) {
#if 0
            SetMesgText(&mesg, cy++, "You are CHECKED.", 2);
#else
	    SetMesgText(&mesg, cy++,
			FntK2 "" FntK1 " 򤫤ޤ!", 2);
#endif
	    ohte_flag = True;
	}
	else
	    ohte_flag = False;
#if 0
        SetMesgText(&mesg, cy++, "Your turn.", 7);
#else
        SetMesgText(&mesg, cy++, FntK2 "ʤ֤Ǥ", 7);
#endif
	SetCursor(ban.win.id, CSR_NORMAL);
	SetCursor(dai.win.id, CSR_NORMAL);
        if( bell )
	    XBell(xdata.display, 0);
	if( auto_raise )
	    XRaiseWindow(xdata.display, win.id);
	break;

    case TENEMYVIOLATE:				/* ENEMYVIOLATE */
#if 0
	sprintf(text, "%s made Violate.", name[1]);
        SetMesgText(&mesg, cy++, text, 4);
#else
        SetMesgTextF(&mesg, cy++, 4, "%s ȿ§Ȥޤ", name[1]);
#endif
        life.life[1]--;
        ExposeLife(&life);
	break;

    default:
	fprintf(stderr, "Can't parse message \"%s\"\n", buff);
	exit(1);
    }
    ExposeMesg(&xdata, &mesg);

    return;
}

static void SelectOnBan(int x, int y)
{
    switch( phase ) {

    case NORMAL:
	if( GripBanKoma(x, y) ) {
	    phase = GRIP;
	    DrawBan(&ban, pastpos, phase == GRIP);
	    ExposeBan(&ban);
	}
	break;

    case GRIP:
	promotemode = DropKoma(x, y);
	switch( promotemode ) {
	case 0:			/* cannot drop */
	    phase = NORMAL;
	    break;
	case 1:			/* can move */
	case 2:			/* can promote */
	    MapAskwin(&xdata, &askdrop);
	    phase = ASKDROP;
	    break;
	case 3:			/* can move and promote*/
	    MapAskwin(&xdata, &askpromote);
	    phase = ASKPROMOTE;
	    break;
	}
	DrawDai(&dai, pastpos, phase == GRIP);
	ExposeDai(&dai);
	DrawBan(&ban, pastpos, phase == GRIP);
	ExposeBan(&ban);
	break;
    }
}

/* ButtonPress */
static void ButtonPressProc(XEvent *ev)
{
    Window rootwin;                 /* for XGetGeomotry */
    int winx, winy, winh, winw, winborder, windepth;
    static int nowpromote;          /* is promote now ?  in promotemode == 3 */

    if( ev->xbutton.window == ban.win.id )			/* ban */
	SelectOnBan(ev->xbutton.x, ev->xbutton.y);
    else if( ev->xbutton.window == dai.win.id )			/* ban */
    {
        if( phase == NORMAL ) {
            if(GripDaiKoma(ev->xbutton.x, ev->xbutton.y)) {
                phase = GRIP;
                DrawDai(&dai, pastpos, phase == GRIP);
                ExposeDai(&dai);
            }
        }
    }
    else if( ev->xbutton.window == ButtonId(giveup) ) {		/* giveup */
        MapAskwin(&xdata, &askgiveup);
        prev_phase = phase;
        phase = ASKGIVEUP;
    }
    else if(ev->xbutton.window == ButtonId(viewnote))		/* viewnote */
        done = GAME_DONE_VIEWNOTE;
    else if(ev->xbutton.window == ButtonId(quit))		/* quit */
        done = GAME_DONE_QUIT;
    else if(ev->xbutton.window == AskwinId(askgiveup))	/* askgiveup */
        MoveAskwin(&xdata, &askgiveup, &win);
    else if(ev->xbutton.window == AskwinYesId(askgiveup))
    {
        UnmapAskwin(&xdata, &askgiveup);
        SendData("GIVEUP\n", 7);
    }
    else if(ev->xbutton.window == AskwinNoId(askgiveup))
    {
        UnmapAskwin(&xdata, &askgiveup);
        phase = prev_phase;
    }
    else if(ev->xbutton.window == AskwinId(askdrop))		/* askdrop */
        MoveAskwin(&xdata, &askdrop, &win);
    else if(ev->xbutton.window == AskwinYesId(askdrop))
    {
        UnmapAskwin(&xdata, &askdrop);
        switch(promotemode)
        {
          case 1:
            phase = NORMAL;
            SendMove(0);
            break;
          case 2:
            phase = NORMAL;
            PromoteKoma(&ban.data[nowpos.y][nowpos.x]);
            SendMove(1);
            DrawBan(&ban, pastpos, phase == GRIP);
            ExposeBan(&ban);
            break;
          case 3:
            phase = NORMAL;
            if(nowpromote == 1)
                SendMove(1);
            else
                SendMove(0);
            break;
        }
    }
    else if(ev->xbutton.window == AskwinNoId(askdrop))
    {
        UnmapAskwin(&xdata, &askdrop);
        phase = NORMAL;
        ban.data[nowpos.y][nowpos.x] = EMPTY;
        if(pastpos.x != FROM_DAI)
            ban.data[pastpos.y][pastpos.x] = pastpos.type;
        else
            dai.data[0][pastpos.y]++;
        DrawBan(&ban, pastpos, phase == GRIP);
        ExposeBan(&ban);
        DrawDai(&dai, pastpos, phase == GRIP);
        ExposeDai(&dai);
    }
    else if(ev->xbutton.window == AskwinId(askpromote))	/* askpromote */
        MoveAskwin(&xdata, &askpromote, &win);
    else if(ev->xbutton.window == AskwinYesId(askpromote))
    {
        UnmapAskwin(&xdata, &askpromote);
        phase = ASKDROP;
        nowpromote = 1;
        PromoteKoma(&ban.data[nowpos.y][nowpos.x]);
        DrawBan(&ban, pastpos, phase == GRIP);
        ExposeBan(&ban);
        XGetGeometry(xdata.display, AskwinId(askpromote),
                     &rootwin, &winx, &winy,
                     &winw, &winh, &winborder, &windepth);
        XMoveWindow(xdata.display, AskwinId(askdrop),
                    winx, winy);
        MapAskwin(&xdata, &askdrop);
    }
    else if(ev->xbutton.window == AskwinNoId(askpromote))
    {
        UnmapAskwin(&xdata, &askpromote);
        phase = ASKDROP;
        nowpromote = 0;
        XGetGeometry(xdata.display, AskwinId(askpromote),
                     &rootwin, &winx, &winy,
                     &winw, &winh, &winborder, &windepth);
        XMoveWindow(xdata.display, AskwinId(askdrop),
                    winx, winy);
        MapAskwin(&xdata, &askdrop);
    }
    return;
}

/* RightButtonPress */
static void RightButtonPressProc(XEvent ev)
{
    switch( phase ) {
    case GRIP:
	phase = NORMAL;
	DrawDai(&dai, pastpos, phase == GRIP);
	ExposeDai(&dai);
	DrawBan(&ban, pastpos, phase == GRIP);
	ExposeBan(&ban);
	break;

    case ASKDROP:
	UnmapAskwin(&xdata, &askdrop);
	phase = NORMAL;
	ban.data[nowpos.y][nowpos.x] = EMPTY;
	if(pastpos.x != FROM_DAI)
	    ban.data[pastpos.y][pastpos.x] = pastpos.type;
	else
	    dai.data[0][pastpos.y]++;
	DrawBan(&ban, pastpos, phase == GRIP);
	ExposeBan(&ban);
	DrawDai(&dai, pastpos, phase == GRIP);
	ExposeDai(&dai);
	break;

    case ASKPROMOTE:
	UnmapAskwin(&xdata, &askpromote);
	phase = NORMAL;
	SendMove(0);
	break;

    case ASKGIVEUP:
	UnmapAskwin(&xdata, &askgiveup);
	phase = prev_phase;
	break;

    }
}

/* MotionNotify */
static void MotionNotifyProc(XEvent ev)
{
    if(ev.xmotion.window == AskwinId(askgiveup))	/* askgiveup */
	MoveAskwin(&xdata, &askgiveup, &win);
    else if(ev.xmotion.window == AskwinId(askdrop))	/* askdrop */
	MoveAskwin(&xdata, &askdrop, &win);
    else if(ev.xmotion.window == AskwinId(askpromote))	/* askpromote */
	MoveAskwin(&xdata, &askpromote, &win);
}

/* Enter/LeaveNotify */
static void CrossNotifyProc(XEvent ev, int mode)
{
    if(ev.xcrossing.window == ButtonId(giveup))			/* giveup */
	FocusButton(&xdata, &giveup, mode);
    if(ev.xcrossing.window == ButtonId(viewnote))		/* viewnote */
	FocusButton(&xdata, &viewnote, mode);
    if(ev.xcrossing.window == ButtonId(quit))			/* quit */
	FocusButton(&xdata, &quit, mode);
    else if(ev.xcrossing.window == AskwinYesId(askgiveup))	/* askgiveup */
	FocusButton(&xdata, &askgiveup.yes, mode);
    else if(ev.xcrossing.window == AskwinNoId(askgiveup))
	FocusButton(&xdata, &askgiveup.no, mode);
    else if(ev.xcrossing.window == AskwinYesId(askdrop))	/* askdrop */
	FocusButton(&xdata, &askdrop.yes, mode);
    else if(ev.xcrossing.window == AskwinNoId(askdrop))
	FocusButton(&xdata, &askdrop.no, mode);
    else if(ev.xcrossing.window == AskwinYesId(askpromote))	/* askpromote */
	FocusButton(&xdata, &askpromote.yes, mode);
    else if(ev.xcrossing.window == AskwinNoId(askpromote))
	FocusButton(&xdata, &askpromote.no, mode);
    return;
}

/* Expose */
static void ExposeProc(XEvent ev)
{
    if(ev.xexpose.window == win.id)				/* win */
	XClearWindow(xdata.display, win.id);
    else if(ev.xexpose.window == ban.win.id)			/* ban */
    {
	DrawBan(&ban, pastpos, phase == GRIP);
	ExposeBan(&ban);
    }
    else if(ev.xexpose.window == dai.win.id)			/* dai */
    {
	DrawDai(&dai, pastpos, phase == GRIP);
	ExposeDai(&dai);
    }
    else if(ev.xexpose.window == life.win.id)			/* life */
	ExposeLife(&life);
    else if(ev.xexpose.window == mesg.win.id)			/* mesg */
	ExposeMesg(&xdata, &mesg);
    else if(ev.xexpose.window == ButtonId(giveup))		/* giveup */
	ExposeButton(&xdata, &giveup);
    else if(ev.xexpose.window == ButtonId(viewnote))		/* viewnote */
	ExposeButton(&xdata, &viewnote);
    else if(ev.xexpose.window == ButtonId(quit))		/* quit */
	ExposeButton(&xdata, &quit);
    else if(ev.xexpose.window == AskwinId(askgiveup))		/* askgiveup */
	XClearWindow(xdata.display, AskwinId(askgiveup));
    else if(ev.xexpose.window == AskwinYesId(askgiveup))
	ExposeButton(&xdata, &askgiveup.yes);
    else if(ev.xexpose.window == AskwinNoId(askgiveup))
	ExposeButton(&xdata, &askgiveup.no);
    else if(ev.xexpose.window == AskwinId(askdrop))		/* askdrop */
	XClearWindow(xdata.display, AskwinId(askdrop));
    else if(ev.xexpose.window == AskwinYesId(askdrop))
	ExposeButton(&xdata, &askdrop.yes);
    else if(ev.xexpose.window == AskwinNoId(askdrop))
	ExposeButton(&xdata, &askdrop.no);
    else if(ev.xexpose.window == AskwinId(askpromote))	/* askpromote */
	XClearWindow(xdata.display, AskwinId(askpromote));
    else if(ev.xexpose.window == AskwinYesId(askpromote))
	ExposeButton(&xdata, &askpromote.yes);
    else if(ev.xexpose.window == AskwinNoId(askpromote))
	ExposeButton(&xdata, &askpromote.no);
    return;
}

static void SendMove(int promotion)
{
    char buff[BUFFLEN];
    if(pastpos.x != FROM_DAI)
    {
	if(promotion)
	    sprintf(buff, "MOVE %d %d %d %d PROMOTE\n",
		    pastpos.x + 1, pastpos.y + 1, nowpos.x + 1, nowpos.y + 1);
	else
	    sprintf(buff, "MOVE %d %d %d %d NOTPROMOTE\n",
		    pastpos.x + 1, pastpos.y + 1, nowpos.x + 1, nowpos.y + 1);
    }
    else
    {
        sprintf(buff, "DROP %d %d %s\n",
                nowpos.x + 1, nowpos.y + 1, KomaName(nowpos.type));
    }
    SendData(buff, strlen(buff));
    return;
}

char *KomaName(int type)
{
    switch(type)
    {
      case FU:
        return "PAWN";
      case KYOSHA:
        return "LANCE";
      case KEIMA:
        return "KNIGHT";
      case GIN:
        return "SILVER";
      case KIN:
        return "GOLD";
      case KAKU:
        return "BISHOP";
      case HISHA:
        return "ROOK";
    }
    return "";
}

void PromoteKoma(unsigned char *type)
{
    *type |= TuPromotionMask;
}

static int DropKoma(int px, int py)
{
    int ret;

    if(px < BAN_OFFSET || px >= BAN_WIDTH - BAN_OFFSET ||
       py < BAN_OFFSET || py >= BAN_HEIGHT - BAN_OFFSET)
	return 0;

    nowpos.x = (px - BAN_OFFSET) / MASU_WIDTH;
    nowpos.y = (py - BAN_OFFSET) / MASU_HEIGHT;
    if(side == SENTE)
	nowpos.x = 8 - nowpos.x;
    else
	nowpos.y = 8 - nowpos.y;
    nowpos.type = pastpos.type;

    if(pastpos.x != FROM_DAI)
	ret = canmove(side, ban.data, dai.data[0],
		      pastpos.y * 9 + pastpos.x,
		      nowpos.y * 9 + nowpos.x, nowpos.type);
    else
	ret = canmove(side, ban.data, dai.data[0],
		      pastpos.y + 81,
		      nowpos.y * 9 + nowpos.x, nowpos.type);

    if(ret != 0)
    {
	if(pastpos.x != FROM_DAI)
	    ban.data[pastpos.y][pastpos.x] = EMPTY;
	else
	    dai.data[0][pastpos.y]--;
	ban.data[nowpos.y][nowpos.x] = nowpos.type;
    }
	
    return ret;
}

static int GripBanKoma(int px, int py)
{
    if(px < BAN_OFFSET || px >= BAN_WIDTH - BAN_OFFSET ||
       py < BAN_OFFSET || py >= BAN_HEIGHT - BAN_OFFSET)
	return 0;

    pastpos.x = (px - BAN_OFFSET) / MASU_WIDTH;
    pastpos.y = (py - BAN_OFFSET) / MASU_HEIGHT;
    if(side == SENTE)
	pastpos.x = 8 - pastpos.x;
    else
	pastpos.y = 8 - pastpos.y;
    pastpos.type = ban.data[pastpos.y][pastpos.x];
    if(pastpos.type == EMPTY)
	return 0;

    return 1;
}

static int GripDaiKoma(int px, int py)
{
    static convert_koma[] = {FU, KYOSHA, KEIMA, GIN, KIN, KAKU, HISHA};
    if(py < DAI_OFFSET || py >= DAI_HEIGHT - DAI_OFFSET)
	return 0;

    pastpos.x = FROM_DAI;
    pastpos.y = (py - DAI_OFFSET) / MASU_HEIGHT;
    pastpos.type = convert_koma[pastpos.y];
    if(dai.data[0][pastpos.y] <= 0)
	return 0;

    return 1;
}

static void KeyPressProc(XKeyEvent *keyev)
{
    KeySym sym;

    union {
	Window w;
	int i;
	unsigned u;
    } dummy;

    sym = XKeycodeToKeysym(xdata.display, keyev->keycode, 0);

    if( keyev->window == ban.win.id && (phase == NORMAL || phase == GRIP) ) {
	int dx, dy;
	Window old_wpt, new_wpt;

	dx = dy = 0;

	switch( sym ) {
	case XK_Up:
	case XK_KP_Up:
	    dy = - MASU_HEIGHT;
	    break;
	case XK_Down:
	case XK_KP_Down:
	    dy = + MASU_HEIGHT;
	    break;
	case XK_Left:
	case XK_KP_Left:
	    dx = - MASU_WIDTH;
	    break;
	case XK_Right:
	case XK_KP_Right:
	    dx = + MASU_WIDTH;
	    break;
	case XK_Return:
	    SelectOnBan(keyev->x, keyev->y);
	    return;
	default:
	    return;
	}

	if( keyev->x + dx < BAN_OFFSET			||
	    keyev->x + dx > BAN_OFFSET + MASU_WIDTH * 9 ||
	    keyev->y + dy < BAN_OFFSET			||
	    keyev->y + dy > BAN_OFFSET + MASU_HEIGHT * 9 )
	    return;	/* ư׾夫ФƤޤ */

	XTranslateCoordinates(xdata.display, keyev->root, keyev->root,
			      keyev->x_root, keyev->y_root,
			      &dummy.i, &dummy.i, &old_wpt);
	XTranslateCoordinates(xdata.display, keyev->root, keyev->root,
			      keyev->x_root + dx, keyev->y_root + dy,
			      &dummy.i, &dummy.i, &new_wpt);

	/* ưȰưκɸΥȥåץ٥롦ɥƱ
	 * (ʤƱץꥱ󥦥ɥ)Τ
	 * ݥ󥿤ư롣ɥޥ͡β뤿
	 * ȥåץ٥롦ɥ win.id ȤϰۤʤäƤ롣
	 */
	if( old_wpt == new_wpt )
	    XWarpPointer(xdata.display, ban.win.id, ban.win.id,
			 keyev->x,  keyev->y, 1, 1,
			 keyev->x + dx, keyev->y + dy);
    }
    else {
	XEvent newevent;
	Window targetwin;
	Askwin *target_askwin;

	if( phase == ASKDROP )
	    target_askwin = &askdrop;
	else if( phase == ASKPROMOTE )
	    target_askwin = &askpromote;
	else
	    return;

	if( sym == XK_Return )
	    targetwin = AskwinYesId((*target_askwin));
	else if( sym == XK_Escape )
	    targetwin = AskwinNoId((*target_askwin));
	else
	    return;

	if( offsetof(XButtonEvent, button) == offsetof(XKeyEvent, keycode) ) {
	    newevent = *keyev;
	    newevent.type = ButtonPress;
	    newevent.xbutton.window = targetwin;
	    newevent.xbutton.button = 1;
	    XSendEvent(keyev->display, targetwin, False,
		       ButtonPressMask, &newevent);
	    ifDebug('X')
		fprintf(stderr,
			"Resending XEvent (KeyPress to ButtonPress) Ok.\n");
	}
	else {
	    fprintf(stderr, "FATAL: Event offsets are NOT equal.\n");
	    exit(100);
	}
    }
}
