/*
 *	kifu.c
 *
 *	This program is a part of XTuitate.
 *	Programmed by TAKAI, Kousuke (tak@kmc.kyoto-u.ac.jp), 1997.
 *
 *	[ͻ] http://www/nsn.co.jp/s-kikaku/kifu.html
 */

#include <stdio.h>
#include <string.h>
#ifdef	TEST
#include <unistd.h>	/* for isatty(3)	*/
#include <ctype.h>	/* for toupper(3)	*/
#endif

#include "all.h"
#include "tuitate.h"
#include "strings.h"
#include "canmove.h"

#ifndef	TEST
#include "drawstring.h"
#else
#define	CodeSente	'S'
#define	CodeGote	'G'
#endif

static const char *words[ ] = {
#define	NIL	0
	"",
#define	MIGI	1
	"",
#define	NAKA	2
	"",
#define	HIDARI	3
	"",
#define	AGARU	4
	"",
#define	YORU	5
	"",
#define	HIKU	6
	"",
#define	SUGU	7
	"ľ",
};

/* <ɽʸ>
 *	buf	ʸǼ
 *	move	
 *	ban	ư
 *
 * ֤ͤbufΤޤ
 */
char * composeKifu(buf, move, ban)
    char *buf;
    TuNotationMove *move;
    unsigned char ban[][9];
{
    unsigned char x, y;
    unsigned char other_x = 0;
    short int dousa;
    short int houkou;
    unsigned char koma;
    unsigned char dousa_count  = 0;
    unsigned char houkou_count = 0;
    char *ptr;

#undef	GOTE_P
#define	GOTE_P	(move->flags & TuSideMask)

    koma = ban[move->dy][move->dx] ^ (move->flags & TuPromotionMask);

    ptr = buf;
    *ptr++ = GOTE_P ? CodeGote : CodeSente;
    *ptr++ = ' ';
    if( move->flags & TuDoMask )
	STRCPY(ptr, "Ʊ");
    else {
	STRCPY(ptr, jissuuji[move->dx + 1]);
	STRCPY(ptr, kansuuji[move->dy + 1]);
    }
    *ptr++ = ' ';
    STRCPY(ptr, koma_abbrev[koma & TuMyKomaMask]);

    for( y = 0 ; y < 9 ; y++ )
	for( x = 0 ; x < 9 ; x++ ) {
	    if( x == move->dx && y == move->dy )
		continue;
	    if( x == move->dx && y == move->dy )
		continue;
	    if( ban[y][x] == koma
		&& TuIsAbleToMove(ban,
				  x, y, move->dx, move->dy, ban[y][x]) ) {
		if( move->type == TDROP ) {
		    STRCPY(ptr, "");
		    goto final;
		}
		other_x = x;
		if( (move->sx < move->dx && x < move->dx) ||
		    (move->sx == move->dx && x == move->dx) ||
		    (move->sx > move->dx && x > move->dx) )
		    houkou_count++;
		if( (move->sy < move->dy && y < move->dy) ||
		    (move->sy == move->dy && y == move->dy) ||
		    (move->sy > move->dy && y > move->dy) )
		    dousa_count++;
	    }
	}

    houkou = dousa_count ?
	(move->sx > move->dx ? (GOTE_P ? MIGI : HIDARI) :
	 move->sx == move->dx ? NAKA : (GOTE_P ? HIDARI : MIGI) )
	: NIL;
    dousa = houkou_count ?
	(move->sy > move->dy ? (GOTE_P ? HIKU : AGARU) :
	 move->sy == move->dy ? YORU : (GOTE_P ? AGARU : HIKU) )
	: NIL;

    if( houkou == NAKA ) {
	koma &= ~TuSideMask;
	if( koma == RYU || koma == UMA )
	    houkou = other_x > move->sx ? MIGI : HIDARI;
	else if( dousa != HIKU ) {
	    houkou = SUGU;
	    dousa  = NIL;
	}
    }

    STRCPY(ptr, words[houkou]);
    STRCPY(ptr, words[dousa]);
    if( TuIsAbleToPromote(move->dx, move->dy, koma) )
	STRCPY(ptr, "" + (move->flags & TuPromotionMask ? 2 : 0));
final:
    return buf;
}

#ifdef	TEST

char *myname;
TuNotation notation;
unsigned char initial[9][9];
unsigned char ban[9][9];

void RecvData(int dummy, char *buf, size_t bufsiz)
{
    char *ptr;

    fgets(buf, bufsiz, stdin);
    for( ptr = buf ; *ptr ; ptr++ ) {
	if( *ptr == '\n' )
	    *ptr = '\0';
	else
	    *ptr = toupper(*ptr);
    }
    if( !isatty(0) )
	puts(buf);
}

int main(int argc, char *argv[ ])
{
    int i;
    int x, y, ym;
    char *ptr;
    char buf[256];
    TuNotationMove *mptr;

    myname = argv[0];

    TuInitializeNotation(&notation);

    for( ym = 0 ; ym < 9 ; ym++ ) {
	printf(">");
	RecvData(0, buf, sizeof(buf));
	if( !buf[0] ) break;
	ptr = buf;
	for( x = 0 ; x < 9 ; x++ ) {
	    int p = 0;

	    while( *ptr == ' ' ) ptr++;
	    if( ptr[0] == 'P' && ptr[1] == '-' ) {
		ptr += 2;
		p = TuPromotionMask;
	    }
	    if( ptr[0] == '.' ) {
		initial[ym][8 - x] = EMPTY;
		ptr++;
	    }
	    else {
		if( (i = TuParseToken(&ptr)) == 0 ) {
		    while( x < 9 )
			initial[ym][8 - x++] = 0;
		    break;
		}
		if( !(initial[ym][8 - x] = TuTokenToKoma(i)) ) {
		    fprintf(stderr, "Read error.\n");
		    return 1;
		}
		initial[ym][8 - x] |= p;
	    }
	}
    }

    while( 1 ) {
	for( x = 0 ; x < 9 ; x++ )
	    printf("%-5s", jissuuji[9 - x]);
	printf("\n");
	for( y = 0 ; y < ym ; y++ ) {
	    for( x = 0 ; x < 9 ; x++ ) {
		if( (ban[y][8 - x] = initial[y][8 - x]) )
		    printf("%-5s",
			   koma_abbrev[ban[y][8 - x] & TuMyKomaMask]);
		else
		    printf("     ");
	    }
	    printf("%s\n", kansuuji[y + 1]);
	}
	printf(":");
	notation.nmove = 0;
	if( TuGetOneMove(&notation) )
	    return 0;
	mptr = &notation.move[0];
	switch( mptr->type ) {
	case TMOVE:
	    ban[mptr->dy][mptr->dx] =
		ban[mptr->sy][mptr->sx] ^ (mptr->flags & TuPromotionMask);
	    ban[mptr->sy][mptr->sx] = EMPTY;
	    break;
	case TDROP:
	    ban[mptr->dy][mptr->dx] = mptr->piece;
	    break;
	}
	composeKifu(buf, notation.move, ban);
	printf("%s\n", buf);
    }
}
#endif	/* TEST */
