/*
 *	misc.c			X-Tuitate miscellous routines
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>

#include "all.h"
#include "misc.h"

/*
 * malloc(3)ΥåѴؿݤǤʤexit롣
 */
void *xmalloc(size)
    size_t size;
{
    void *ret;

    ret = malloc(size);
    if( ret == NULL ) {
	fprintf(stderr, "%s: No enough memory to allocate.\n", myname);
	exit(1);
    }
    return ret;
}

/*
 * realloc(3)ΥåѴؿԤexit롣
 */
void *xrealloc(void *ptr, size_t size)
{
    void *ret;

    ret = realloc(ptr, size);
    if( ret == NULL ) {
	fprintf(stderr, "%s: No enough memory to re-allocate.\n", myname);
	exit(1);
    }
    return ret;
}

/*
 * malloc Ѥ realloc(3)ؿԤexit롣
 */
void *cpyalloc(void *ptr, size_t oldsize, size_t newsize)
{
    void *ret;

    ret = malloc(newsize);
    if( ret == NULL ) {
	fprintf(stderr, "%s: No enough memory to re-allocate.\n", myname);
	exit(1);
    }
    memcpy(ret, ptr, oldsize);
    free( ptr );
    return ret;
}

#ifndef	HAVE_STRDUP

char *strdup(const char *str)
{
    char *allocated;

    if( (allocated = malloc(strlen(str) + 1)) != NULL )
	strcpy(allocated, str);
    return allocated;
}

#endif

/*
 * format ؿ
 *
 * char * format ( char * fmt, ... ) ;
 *
 *	fmt Ǽ񼰻ʸˤäưեޥåȤ
 *	Ǥʸ֤ʸĹ¤Ϥʤ
 *
 *	Emacs LISP  format ؿ Perl  sprintf ˻Ƥ롣
 */
#if 1
#ifdef	HAVE_VASPRINTF

char * format(const char *fmt, ...)
{
    va_list ap;
    char * str;
    int ret;

    va_start(ap, fmt);
    ret = vasprintf(&str, fmt, ap);
    if( ret < 0 && str ) {
	free(str);
	fprintf(stderr, "%s: format: error occured\n", myname);
	exit(1);
    }
    va_end(ap);

    return str;
}

#else	/* HAVE_VASPRINTF */

#ifndef	TEST
#define	FORMAT_BLKSIZ	256
#else
#define	FORMAT_BLKSIZ	8
#endif

#ifdef	__NetBSD__

static int format_write(void * cookie, char const *buf, int n)
{
    register struct __sbuf *context = (struct __sbuf *)cookie;

    context->_base = xrealloc(context->_base, context->_size + n);
    memcpy(context->_base + context->_size, buf, n);
    context->_size += n;
    return n;
}

char * format(const char *fmt, ...)
{
    va_list ap;
    FILE f;
    struct __sbuf context = { NULL, 0 };
    unsigned char buf[FORMAT_BLKSIZ];

    f._flags = __SWR;
    f._bf._base = f._p = buf;
    f._bf._size = f._w = FORMAT_BLKSIZ;
    f._cookie = &context;
    f._write = format_write;
    va_start(ap, fmt);
    if( vfprintf(&f, fmt, ap) < 0 || fputc('\0', &f) == EOF || fflush(&f) ) {
	free(context._base);
	return NULL;
    }
    va_end(ap);
    return context._base;
}
#else	/* __NetBSD__ */

#if 1
char * format(const char *fmt, ...)
{
    va_list ap;
    FILE * fp;
    pid_t child;
    int fd[2];
    char * string;
    int len;
    int this_len;
    int status;

    if( pipe(fd) ) {
	fprintf(stderr, "%s: format: cannot create pipe\n",	myname);
	exit(1);
    }
    switch( child = fork() ) {
    case -1:
	fprintf(stderr, "%s: format: cannot create child process\n", myname);
	exit(1);
    case 0:
	close(fd[0]);
	if( (fp = fdopen(fd[1], "w")) == NULL ) {
	    close(fd[1]);
	    fprintf(stderr, "%s: format: cannot open pipe(child)\n", myname);
	    exit(1);
	}
	va_start(ap, fmt);
	vfprintf(fp, fmt, ap);
	va_end(ap);
	fputc('\0', fp);
	fclose(fp);
	exit(0);
    }
    close(fd[1]);
    if( (string = malloc(FORMAT_BLKSIZ)) == NULL )
	goto failed_alloc;
    len = 0;
    while( 1 ) {
	if( (this_len = read(fd[0], string + len, FORMAT_BLKSIZ)) < 0 ) {
	    fprintf(stderr, "%s: format: cannot read from child process\n",
		    myname);
	error:
	    close(fd[0]);
	    kill(child, SIGTERM);
	    waitpid(child, &status, 0);
	    exit(1);
	}
	len += this_len;
	if( this_len < FORMAT_BLKSIZ )
	    break;
	if( (string = realloc(string, len + FORMAT_BLKSIZ)) == NULL ) {
	failed_alloc:
	    fprintf(stderr, "%s: format: cannot allocate memory\n", myname);
	    goto error;
	}
    }
    close(fd[0]);
    waitpid(child, &status, 0);
    if( !WIFEXITED(status) || WEXITSTATUS(status) ) {
	fprintf(stderr, "%s: format: child process error\n", myname);
	exit(1);
    }
    return xrealloc(string, len);
}	
#else

static char * format_itoa(char * buf,
			  unsigned long int val, int radix, int upper)
{
    unsigned long int r, q;

    if( (r = val / radix) )
	buf = format_itoa(buf, r, radix, upper);
    q = val % radix;
    *buf++ = q < 10 ? '0' + q : upper ? 'A' - 10 + q : 'a' - 10 + q;
    return buf;
}

#ifdef	__GNUC__
#define	NESTED_FUNCTION
#endif

#ifndef	NESTED_FUNCTION
static size_t format_put(char **base, size_t *alloc, size_t idx, char ch)
{
}
#endif
char * format(const char *fmt, ...)
{
    va_list ap;
    char * str;
    char buf[33];	/* for 00000000111111110000000011111111 */
    char padch;
    unsigned int flags;
    char c;
    unsigned int precision;
    int width;
    char * ptr;
    long int sval;
    unsigned long int uval;

#define	FMT_S_INT	(1U << 0)
#define	FMT_S_SHORT	(2U << 0)
#define	FMT_S_QUAD	(3U << 0)
#define	FMT_S_MASK	(3U << 0)
#define	FMT_LEFTADJ	(1U << 8)
#define	FMT_ALTFORM	(1U << 9)
#define	FMT_NEEDSPC	(1U <<10)
#define	FMT_NEEDPLS	(1U <<11)
#define	FMT_XUPPER	(1U <<12)

    va_start(ap, fmt);

    while( (c = *fmt++) ) {
	if( c == '%' ) {
	    padch = ' ';
	    flags = 0;
	    width = 0;
	    ptr = buf;
	    while( (c = *fmt++) ) {
		if( c == '#' )
		    flags |= FMT_ALTFORM;
		else if( c == '-' )
		    flags |= FMT_LEFTADJ;
		else if( c == '+' )
		    flags |= FMT_NEEDPLS;
		else
		    break;
	    }
	    if( c == '0' ) {
		if( !(flags & FMT_LEFTADJ) )
		    padch = '0';
		c = *fmt++;
	    }
	    if( c == '*' ) {
		width = va_arg(ap, int);
		c = *fmt++;
	    }
	    if( '0' <= c && c <= '9' ) {
		width = (int)strtol(--fmt, &fmt, 10);
		c = *fmt++;
	    }
	    if( c == '.' ) {
		if( *fmt == '*' ) {
		    precision = va_arg(ap, int);
		    fmt++;
		}
		if( '0' <= *fmt && *fmt <= '9' )
		    precision = (int)strtol(fmt, &fmt, 10);
		c = *fmt++;
	    }
	    switch( c ) {
	    case 'h':
		flags = (flags & ~FMT_S_MASK) | FMT_S_SHORT;
		c = *fmt++;
		break;
	    case 'l':
		flags = (flags & ~FMT_S_MASK) | FMT_S_LONG;
		c = *fmt++;
		break;
	    }
	    switch( c ) {
	    case 'd':
		switch( flags & FMT_S_MASK ) {
		case FMT_S_SHORT:
		    sval = va_arg(ap, short int);
		    break;
		case FMT_S_LONG:
		    sval = va_arg(ap, long int);
		    break;
		default:
		    sval = va_arg(ap, int);
		    break;
		}
		if( sval < 0 ) {
		    *ptr++ = '-';
		    uval = - sval;
		}
		else {
		    if( flags & FMT_NEEDPLS ) {
			*ptr++ = '+';
		    }
		    else if( flags & FMT_NEEDSPC ) {
			*ptr++ = ' ';
			width--;
		    }
		    uval = sval;
		}
		radix = 10;
		goto vals;
	    case 'u':
		radix = 10;
		goto getval;
	    case 'X':
		flags |= FMT_XUPPER;
	    case 'x':
		radix = 16;
		goto getval;
	    case 'o':
		radix = 8;
		goto getval;
	    case 'b':
		radix = 2;
	    getval:
		switch( flags & FMT_S_MASK ) {
		case FMT_S_SHORT:
		    uval = va_arg(ap, unsigned short int);
		    break;
		case FMT_S_LONG:
		    uval = va_arg(ap, unsigned long int);
		    break;
		default:
		    uval = va_arg(ap, unsigned int);
		    break;
		}
	    vals:
		ptr = fotmat_itoa(ptr, uval, radix, flags & FMT_XUPPER);
		goto pad;
	    case 'c':
		*ptr ++ = va_arg(ap, char);
	    pad:
		*ptr = '\0';
		ptr = buf;
		goto spad;
	    spad:
		width -= strlen(ptr);
		if( !(flags & FMT_LEFTADJ) )
		    while( width > 0 ) {
			width--;
			PUT(padch);
		    }
		while( *ptr ) {
		    PUT(*ptr);
		    ptr++;
		}
		if( flags & FMT_LEFTADJ )
		    while( width > 0 ) {
			width--;
			PUT(padch);
		    }
		break;

	    case '\0':
		goto ret;
	    default:
		break;
	    }
	}
	else
	    PUT(c);
    }

    va_end(ap);

    return str;
}
#endif

#endif	/* __NetBSD__ */
#endif	/* HAVE_VASPRINTF */
#endif

int shexec(const char *command)
{
    char *ap;
    const char *cp;
    int flag, argc;
    char argbuf[strlen(command) + 1];
    char * argv[strlen(command) /2 + 1];

    argc = 0;
    flag = 0;
    for( ap = argbuf, cp = command ; *cp ; ) {
	if( *cp == '"' || *cp == '$' || *cp == '&' || *cp == '*' ||
	    *cp == '<' || *cp == '>' || *cp == '\''|| *cp == '|' ||
	    *cp == ';' ) {
	    if( Debugging('S') )
		fprintf(stderr, "Execute via sh: %s\n", command);
	    return execl("/bin/sh", "sh", "-c", command, NULL);
	}
	if( *cp == ' ' ) {
	    *ap++ = '\0';
	    do { cp++; } while( *cp == ' ' );
	    flag = 0;
	}
	else {
	    if( !flag ) {
		argv[argc++] = ap;
		flag = 1;
	    }
	    if( *cp == '\\' )
		if( !*++cp ) break;
	    *ap++ = *cp++;
	}
    }
    *ap = '\0';
    argv[argc] = NULL;

    if( Debugging('S') ) {
	char **p;
	fprintf(stderr, "Execute directly: %s", argv[0]);
	for( p = argv ; *p ; p++ )
	    fprintf(stderr, " '%s'", *p);
	fprintf(stderr, "\n");
    }

    return execvp(argv[0], argv);
}

#ifdef	TEST
char * myname;

int main()
{
    myname = "format() test";

    puts(format("This is a test"));
    puts(format("This is a %s test", "more"));
    puts(format("Number is %d", 12345678));

    return 0;
}
#endif
