/*
 *	resource.c	XTuitate resource loader
 *
 * Note:
 *	These functions will be obsoleted when XTuitate adopts X Toolkit.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <pwd.h>

#include "all.h"
#include "config.h"
#include "resource.h"
#include "misc.h"
#include "version.h"

#ifndef	X11LIBDIR
#define	X11LIBDIR	"/usr/lib/X11"
#endif
#define	APPLCLASS	"XTuitate"
#ifndef	XAPPLOADDIR
#define	XAPPLOADDIR	"/usr/lib/X11/app-defaults"
#endif

static char class[ ] = APPLCLASS;

static char *fallback_resources[ ] = {
#include "XTuitate.h"
    NULL
};

static char *progname;
static XrmDatabase database;

XrmDatabase* resource_database(void)
{
    return &database;
}

void load_resources(Display *disp)
{
    char *env;
    char *home;
    XrmDatabase temp;
    size_t home_len;

    home = getenv("HOME");
    home_len = home ? strlen(home) : 0;

    if( (env = getenv("XENVIRONMENT")) != NULL )
	XrmCombineFileDatabase(env, &database, False);
    else if( home ) {
	char filename[home_len + sizeof("/.Xdefaults-") + MAXHOSTNAMELEN];
	char *ptr = filename;

	STRCPY(ptr, home);
	STRCPY(ptr, "/.Xdefaults-");
	gethostname(ptr, MAXHOSTNAMELEN);
	XrmCombineFileDatabase(filename, &database, False);
    }
    {
	char *disp_res;

	if( (disp_res = XResourceManagerString(disp)) != NULL ) {
	    temp = XrmGetStringDatabase(disp_res);
	    XrmCombineDatabase(temp, &database, False);
	}
	else if( home ) {
	    char filename[home_len + sizeof("/.Xdefaults")];

	    strcpy(filename, home);
	    strcat(filename, "/.Xdefaults");
#ifdef	TEST
	    printf("Loading \"%s\"\n", filename);
#endif
	    XrmCombineFileDatabase(filename, &database, False);
	}
    }

    if( (env = getenv("XAPPLRESDIR")) != NULL ) {
	char filename[strlen(env) + sizeof("/" APPLCLASS)];

	strcpy(filename, env);
	strcat(filename, "/" APPLCLASS);
	XrmCombineFileDatabase(filename, &database, False);
    }

    {
	char filename[strlen(XAPPLOADDIR) + sizeof(class) + 1];
	char *ptr = filename;

	STRCPY(ptr, XAPPLOADDIR);
	*ptr++ = '/';
	STRCPY(ptr, class);
	XrmCombineFileDatabase(ptr, &database, False);
    }
    {	/* Fallback resources */
	XrmDatabase temporaryDB = NULL;
	char **ptr;

	for( ptr = fallback_resources ; *ptr != NULL ; ptr++ )
	    XrmPutLineResource(&temporaryDB, *ptr);
	XrmCombineDatabase(temporaryDB, &database, False);
    }
}

static XrmOptionDescRec options[] = {
    { "+bell",		"*bell",	XrmoptionNoArg,		"False" },
/*    { "-D",		".debugFlag",	XrmoptionStickyArg,	NULL }, */
    { "-autoraise",	"*autoRaise",	XrmoptionNoArg,		"True" },
    { "-background",	".background",	XrmoptionSepArg,	NULL },
    { "-bgimage",	".backgroundImage", XrmoptionSepArg,	NULL },
    { "-d",		".display",	XrmoptionSepArg,	NULL },
    { "-display",	".display",	XrmoptionSepArg,	NULL },
    { "-help",		".help",	XrmoptionNoArg,		"True" },
    { "-geometry",	".geometry",	XrmoptionSepArg,	NULL },
    { "-nobell",	"*bell",	XrmoptionNoArg,		"False" },
    { "-nb",		"*bell",	XrmoptionNoArg,		"False" },
    { "-newfile",	"*newFile",	XrmoptionNoArg,		"True" },
    { "-nf",		"*newFile",	XrmoptionNoArg,		"True" },
    { "-s",		"*server",	XrmoptionSepArg,	NULL },
    { "-server",	"*server",	XrmoptionSepArg,	NULL },
    { "-synchronous",	"*synchronous",	XrmoptionNoArg,		"True" },
    { "-xrm",		NULL,		XrmoptionResArg,	NULL },
};

void parse_command(int argc, char* argv[ ])
{
    char *ptr;
    Display* disp;

    XrmInitialize( );

    if( (ptr = strrchr(argv[0], '/')) == NULL )
	ptr = argv[0];
    else
	ptr++;
    progname = ptr;

    XrmParseCommand(&database, options, sizeof(options) / sizeof(options[0]),
		    ptr, &argc, argv);

#ifdef	DEBUG
    {
	char * ptr;
	char ** arg_d, ** arg_s;

	arg_d = arg_s = argv + 1;

	while( (ptr = *arg_d++ = *arg_s++) != NULL )
	    if( *ptr++ == '-' && *ptr == 'D' ) {
		for( ++ptr ; *ptr ; ptr++ ) {
		    if( *ptr == '-' )
			debug = ~debug;
		    else if( '@' <= *ptr && *ptr <= 'Z' )
			debug ^= 1 << (*ptr - '@');
		    else if( 'a' <= *ptr && *ptr <= 'z' )
			debug ^= 1 << (*ptr - 'a' + 1);
		    else {
			fprintf(stderr,
				"%s: Invalid debug flag `%c' in `%s'\n",
				argv[0], *ptr, *(arg_s - 1));
			exit(1);
		    }
		}
		arg_d--;
		argc--;
	    }
    }
#endif

    if( argc != 1 ) {
	fprintf(stderr, "%s: Undefined option `%s'\n", argv[0], argv[1]);
	exit(1);
    }

    disp = XOpenDisplay(get_resource_string(".display", NULL));
    if( disp == NULL ) {
	fprintf(stderr, "%s: Can't open display.\n", argv[0]);
	exit(1);
    }
    load_resources(disp);
    XCloseDisplay(disp);
}

void putHelp(void)
{
    printf("xtuitate (Tuitate Shogi client for X Window System)\n");
    printf("Version %s (Tuitate protocol version %s)\n",
	   version, protocol_version);
    printf("Compiled at %s\n\n", compilation_time);

    printf("usage: %s [-options ...]\n\nwhere options include:\n", progname);
    printf("\
	-display dpy	X server on which to display\n\
	-server host	Tuitate server\n\
	-autoraise	raise the window when your turn come\n\n");
}
	   
char *get_resource_string(char *name, char *def)
{
    char *buf;
    char *type;
    XrmValue value;
    char res_name[strlen(progname) + strlen(name) + 1];
    char *ptr = res_name;

    STRCPY(ptr, progname);
    STRCPY(ptr, name);

    if( XrmGetResource(database, res_name, class, &type, &value) ) {
	if( (buf = malloc(value.size + 1)) == NULL ) {
	    fprintf(stderr, "Memory exhausted\n");
	    exit(1);
	}
	memcpy(buf, value.addr, value.size);
	buf[value.size] = '\0';
	return buf;
    }
    return def;
}

char *getResourceString(const char *name, const char *rclass)
{
    char *buf;
    char resource_name[strlen(progname) + strlen(name) + 1];
    char class_name[strlen(class) + strlen(rclass) + 1];
    char *type;
    XrmValue value;
    char *ptr;

    ptr = resource_name;	STRCPY(ptr, progname);	STRCPY(ptr, name);
    ptr = class_name;		STRCPY(ptr, class);	STRCPY(ptr, rclass);

    if( !XrmGetResource(database, resource_name, class_name, &type, &value) ) {
	fprintf(stderr, "%s: Resource \"%s\" (class \"%s\") not found\n",
		progname, resource_name, class_name);
	exit(1);
    }
    buf = xmalloc(value.size + 1);
    memcpy(buf, value.addr, value.size);
    buf[value.size] = '\0';
    return buf;
}

char *getResourceStringNoError(const char *name, const char *rclass)
{
    char *buf;
    char resource_name[strlen(progname) + strlen(name) + 1];
    char class_name[strlen(class) + strlen(rclass) + 1];
    char *type;
    XrmValue value;
    char *ptr;

    ptr = resource_name;	STRCPY(ptr, progname);	STRCPY(ptr, name);
    ptr = class_name;		STRCPY(ptr, class);	STRCPY(ptr, rclass);

    if( !XrmGetResource(database, resource_name, class_name, &type, &value) )
	return NULL;

    buf = xmalloc(value.size + 1);
    memcpy(buf, value.addr, value.size);
    buf[value.size] = '\0';
    return buf;
}

int getResourceInteger(const char *name, const char *rclass)
{
    char resource_name[strlen(progname) + strlen(name) + 1];
    char class_name[strlen(class) + strlen(rclass) + 1];
    char *type;
    XrmValue value;
    char *ptr;

    ptr = resource_name;	STRCPY(ptr, progname);	STRCPY(ptr, name);
    ptr = class_name;		STRCPY(ptr, class);	STRCPY(ptr, rclass);

    if( XrmGetResource(database, resource_name, class_name, &type, &value) ) {
	char buf[value.size + 1];
	memcpy(buf, value.addr, value.size);
	buf[value.size] = '\0';
	return atoi(buf);
    }
    else {
	fprintf(stderr, "%s: Resource \"%s\" (class \"%s\") not found\n",
		progname, resource_name, class_name);
	exit(1);
    }
}

int get_resource_boolean(char *name, int def)
{
    char *res;

    res = get_resource_string(name, NULL);
    if( res != NULL ) {
	if( strcasecmp(res, "true") == 0 ||
	    strcasecmp(res, "on"  ) == 0 ||
	    strcasecmp(res, "yes" ) == 0 ) {
	    free(res);
	    return 1;
	}
	else if( strcasecmp(res, "false") == 0 ||
	    strcasecmp(res, "off"  ) == 0 ||
	    strcasecmp(res, "no"   ) == 0 ) {
	    free(res);
	    return 0;
	}
	else {
	    fprintf(stderr, "warning: Incompatible resource type for `%s'\n",
		    name);
	}
    }
    return def;
}

int getResourceBoolean(const char *name, const char *rclass)
{
    char resource_name[strlen(progname) + strlen(name) + 1];
    char class_name[strlen(class) + strlen(rclass) + 1];
    char *type;
    XrmValue value;
    char *ptr;

    ptr = resource_name;	STRCPY(ptr, progname);	STRCPY(ptr, name);
    ptr = class_name;		STRCPY(ptr, class);	STRCPY(ptr, rclass);

    if( XrmGetResource(database, resource_name, class_name, &type, &value) ) {
	char buf[value.size + 1];
	memcpy(buf, value.addr, value.size);
	buf[value.size] = '\0';
	if( !strcasecmp(buf, "true") || !strcasecmp(buf, "enable") ||
	    !strcasecmp(buf, "on")   || !strcasecmp(buf, "yes") )
	    return 1;
	return 0;
    }
    else {
	fprintf(stderr, "%s: Resource \"%s\" (class \"%s\") not found\n",
		progname, resource_name, class_name);
	exit(1);
    }
}

struct passwd *retrievePasswd(void)
{
    static struct passwd *passwd = NULL;

    if( passwd == NULL )
	if( (passwd = getpwuid(getuid())) == NULL ) {
	    fprintf(stderr, "Can't retrieve your passwd entry.\n");
	    exit(1);
	}
    return passwd;
}

char *retrieveDomain(void)
{
    static char domain[256] = "";

    if( !domain[0] ) {
	FILE *fp;
	char *s;

	if( (fp = fopen("/usr/local/lib/default-domain", "r")) != NULL ) {
	    s = fgets(domain, sizeof(domain), fp);
	    fclose(fp);
	    if( s == NULL ) goto Error;
	    if( (s = strchr(domain, '\n')) != NULL )
		*s = '\0';
	}
	else {
	Error:
	    fprintf(stderr, "Can't read your domain\n");
	    perror("/usr/local/lib/default-domain");
	    exit(1);
	}
    }
    return domain;
}

/*
 * userName, userAddress ΥեޥåȻŸ
 *
 * ѤǤեޥåȻ:
 *
 *	%D	/usr/local/lib/default-domain ˵ҤƤɥᥤ̾
 *	%F	/etc/passwd (ޤNISǡ١)Υ桼Υե͡
 *	%u	桼Υ̾
 *		($USER  $LOGNAME  passwd ȥν˸)
 *	%U	桼Υ̾ (passwd ȥΤ߸)
 *		 su ƤȤˤ %u  %U η̤ۤʤ뤳Ȥ롩
 *	%%	1Ĥ`%'
 */
void parseUserName(char *buf, size_t buflen, const char *fmt)
{
    char c;
    char * source;

    if( !buflen || !--buflen )
	goto EndOfString;

    while( (c = *fmt++) ) {
	if( c == '%' ) {
	    switch( (c = *fmt++) ) {
	    case '\0':
		goto EndOfString;
	    case 'D':	/* domain name from /usr/local/lib/default-domain */
		source = retrieveDomain();
		goto CopyString;
	    case 'F':	/* User's real name from /etc/passwd gecos field */
		source = retrievePasswd()->pw_gecos;
		while( (c = *source++) && c != ',' ) {
		    if( !--buflen ) goto EndOfString;
		    *buf++ = (c == ' ') ? '+' : c;
		}
		break;
	    case 'u':	/* User's name from $USER, $LOGNAME or passwd */
		if( (source = getenv("USER")) != NULL
		    || (source = getenv("LOGNAME")) != NULL )
		    goto CopyString;
		/*  */
	    case 'U':	/* User's name from /etc/passwd */
		source = retrievePasswd()->pw_name;
	    CopyString:
		while( (c = *source++) ) {
		    if( !--buflen ) goto EndOfString;
		    *buf++ = c == ' ' ? '+' : c;
		}
		break;
	    default:
		goto OneCharacter;
	    }
	}
	else if( c == ' ' ) {
	    c = '+';
	    goto OneCharacter;
	}
	else {
	OneCharacter:
	    if( !--buflen ) goto EndOfString;
	    *buf++ = c;
	}
    }
EndOfString:
    *buf = '\0';
}

#ifdef	TEST
int main(int argc, char *argv[])
{
    Display *disp;

    if( (disp = XOpenDisplay(NULL)) == NULL ) {
	fprintf(stderr, "Can't open display.\n");
	exit(1);
    }

    init_resources(disp, argv[0]);

    printf("server: %s\n",
	   get_resource_string(".server", NULL));
    printf("user name: %s\n",
	   get_resource_string(".userName", NULL));
    printf("user address: %s\n",
	   get_resource_string(".userAddress", NULL));
    
    return 0;
}
#endif
