#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "commdata.h"
#include "all.h"
#include "selectenemy.h"
#include "pixmaps.h"

#define ENEMY_WIDTH			512
#define ENEMY_HEIGHT			40
#define ENEMY_DISP			8
#define ENEMY_MAX			64

#define MESG_X				24
#define MESG_Y				(ENEMYLIST_Y + ENEMYLIST_HEIGHT + 24)

#define ENEMYLIST_WIDTH			ENEMY_WIDTH
#define ENEMYLIST_HEIGHT		(ENEMY_HEIGHT * (ENEMY_DISP + 1))
#define ENEMYLIST_X			24
#define ENEMYLIST_Y			24

#define ENEMYPAGE_WIDTH			64
#define ENEMYPAGE_HEIGHT		20
#define ENEMYPAGE_X			(ENEMYLIST_WIDTH - ENEMYPAGE_WIDTH)
#define ENEMYPAGE_Y			0

#define ENEMYFLUSH_WIDTH		64
#define ENEMYFLUSH_HEIGHT		40
#define ENEMYFLUSH_X			(ENEMYPAGE_X - ENEMYFLUSH_WIDTH)
#define ENEMYFLUSH_Y			0

#define QUIT_WIDTH			128
#define QUIT_HEIGHT			64
#define QUIT_X				(WIN_WIDTH - QUIT_WIDTH - 24)
#define QUIT_Y				(WIN_HEIGHT - QUIT_HEIGHT - 24)

#define ASKACCEPT_X			((WIN_WIDTH - ASKWIN_WIDTH) / 2)
#define ASKACCEPT_Y			((WIN_HEIGHT - ASKWIN_HEIGHT) / 2)

#define NORMAL			0
#define CHALLENGE		1
#define ASKACCEPT		2

static void DisplayProc(void);
static void ServerProc(void);
static void ExposeEnemylist(void);
static void DrawAllEnemy(void);
static void DrawEnemy(int n);
static void ReadEnemy(void);
char **MallocWordbuff(void);
void FreeWordbuff(char **wordbuff);

static Windata enemylist;
static Button enemypage, enemyflush, enemy[ENEMY_DISP], quit;
static Askwin askaccept;
static Mesg mesg;
static int enemyid[ENEMY_MAX];
static char enemyname[ENEMY_MAX][BUFFLEN];
static char enemyaddress[ENEMY_MAX][BUFFLEN];
static int enemy_num;
static int page_num, page;
static int phase;
static int done;

int SelectEnemy(void)
{
    int i;

    /* enemybox */
    ConstructWindata(&xdata, win.id,
                     ENEMYLIST_X, ENEMYLIST_Y,
                     ENEMYLIST_WIDTH, ENEMYLIST_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &enemylist);
    XSetWindowBackgroundPixmap(xdata.display, enemylist.id,
			       ENEMYLIST_XPM.pixmap);
    XSelectInput(xdata.display, enemylist.id, ExposureMask);

    /* button */
    CreateButton(&xdata, &enemylist,
                 ENEMYPAGE_XPM.pixmap, ENEMYPAGE_F_XPM.pixmap,
                 ENEMYPAGE_X, ENEMYPAGE_Y,
                 ENEMYPAGE_WIDTH, ENEMYPAGE_HEIGHT, 0, &enemypage);
    CreateButton(&xdata, &enemylist,
                 ENEMYFLUSH_XPM.pixmap, ENEMYFLUSH_F_XPM.pixmap,
                 ENEMYFLUSH_X, ENEMYFLUSH_Y,
                 ENEMYFLUSH_WIDTH, ENEMYFLUSH_HEIGHT, 0, &enemyflush);
    for(i = 0; i < ENEMY_DISP; i++)
         CreateButton(&xdata, &enemylist,
                      ENEMY_XPM.pixmap, ENEMY_F_XPM.pixmap,
                      0, ENEMY_HEIGHT * (i + 1),
                      ENEMY_WIDTH, ENEMY_HEIGHT, 0, &enemy[i]);
    CreateButton(&xdata, &win,
                 QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, QUIT_Y,
                 QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);

    /* askaccept */
    CreateAskwin(&xdata, &win, "Accept?",
                 ASKACCEPT_X, ASKACCEPT_Y, &askaccept);

    /* mesg */
    CreateMesg(&xdata, &win, MESG_X, MESG_Y, &mesg);

    /* map window */
    MapButton(&xdata, &enemypage);
    MapButton(&xdata, &enemyflush);
    MapButton(&xdata, &quit);
    MapMesg(&xdata, &mesg);
    XMapRaised(xdata.display, enemylist.id);

    /* init data */
    done = SE_DONE_NORMAL;
    ReadEnemy();
    page = 0;
    DrawAllEnemy();
    phase = NORMAL;

    /* main loop */
    while(done == SE_DONE_NORMAL)
    {
        XFlush(xdata.display);
        switch( WaitComm() )
        {
          case FROM_DISPLAY:
            DisplayProc();
            break;
          case FROM_SERVER:
            ServerProc();
            break;
          default:
            fprintf(stderr, "SelectEnemy() : "
                    "WaitComm() returns FROM_OTHER(%d)\n", FROM_OTHER);
            exit(0);
        }
    }

    XDestroyWindow(xdata.display, enemylist.id);
    XDestroyWindow(xdata.display, quit.win.id);
    XDestroyWindow(xdata.display, askaccept.win.id);
    XDestroyWindow(xdata.display, mesg.win.id);
	
    return done;
}

/* call from display */
static void DisplayProc(void)
{
    XEvent ev;
    char buff[BUFFLEN];
    int i;

    while(XPending(xdata.display))
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            switch(phase)
            {
              case NORMAL:
                for(i = 0; i < ENEMY_DISP; i++)
                {
                    if(ev.xbutton.window == ButtonId(enemy[i]))
                    {
                        phase = CHALLENGE;
                        strcpy(name[1], enemyname[i]);
                        strcpy(address[1], enemyaddress[i]);
                        sprintf(buff, "CHALLENGE %d\n",
                                enemyid[i + ENEMY_DISP * page]);
                        SendData(buff, strlen(buff));
#if 0
                        sprintf(buff, "%s's answer.", enemyname[i]);
                        SetMesgText(&mesg, 0, "Please wait for", 7);
                        SetMesgText(&mesg, 1, buff, 7);
                        SetMesgText(&mesg, 2, "", 7);
                        SetMesgText(&mesg, 3, "", 7);
#else
			SetMesgTextF(&mesg, 0, 7, "%s ", enemyname[i]);
			SetMesgText(&mesg, 1, "ֻԤäƤޤ...", 7);
			SetMesgText(&mesg, 2, "", 7);
			SetMesgText(&mesg, 3, "", 7);
#endif
                        ExposeMesg(&xdata, &mesg);
                    }
                }
                if(ev.xbutton.window == ButtonId(enemypage))
                {
                    page++;
                    page %= page_num;
                    DrawAllEnemy();
                    ExposeEnemylist();
                }
                else if(ev.xbutton.window == ButtonId(enemyflush))
                {
                    ReadEnemy();
                    page = 0;
                    DrawAllEnemy();
                    ExposeEnemylist();
                }
                else if(ev.xbutton.window == ButtonId(quit))
                {
                    done = SE_DONE_QUIT;
                    SendData("QUIT\n", 7);
                }
                break;
              case ASKACCEPT:
                if(ev.xbutton.window == AskwinYesId(askaccept))
                {
                    phase = NORMAL;
                    SendData("ACCEPT\n", 7);
                    UnmapAskwin(&xdata, &askaccept);
                }
                else if(ev.xbutton.window == AskwinNoId(askaccept))
                {
                    phase = NORMAL;
                    SendData("REFUSE\n", 7);
                    UnmapAskwin(&xdata, &askaccept);
#if 0
                    sprintf(buff, "%s's challenge.", name[1]);
                    SetMesgText(&mesg, 0, "You refused", 7);
                    SetMesgText(&mesg, 1, buff, 7);
#else
		    SetMesgTextF(&mesg, 0, 7, "%s ", name[1]);
		    SetMesgText(&mesg, 1, "ĩݤޤ", 7);
#endif
                    SetMesgText(&mesg, 2, "", 7);
                    SetMesgText(&mesg, 3, "", 7);
                    ExposeMesg(&xdata, &mesg);
                }
                else if(ev.xbutton.window == ButtonId(quit))
                {
                    done = SE_DONE_QUIT;
                    SendData("QUIT\n", 7);
                }
                break;
            } /* switch(phase) */
            break;
          case MotionNotify:
            if(ev.xmotion.window == AskwinId(askaccept))
                MoveAskwin(&xdata, &askaccept, &win);
            break;
          case EnterNotify:
            switch(phase)
            {
              case NORMAL:
                for(i = 0; i < ENEMY_DISP; i++)
                {
                    if(ev.xcrossing.window == ButtonId(enemy[i]))
                    {
                        FocusButton(&xdata, &enemy[i], BUTTON_ENTER);
                        DrawEnemy(i);
                    }
                }
                if(ev.xcrossing.window == ButtonId(enemypage))
                    FocusButton(&xdata, &enemypage, BUTTON_ENTER);
                else if(ev.xcrossing.window == ButtonId(enemyflush))
                    FocusButton(&xdata, &enemyflush, BUTTON_ENTER);
                else if(ev.xcrossing.window == ButtonId(quit))
                    FocusButton(&xdata, &quit, BUTTON_ENTER);
                break;
              case ASKACCEPT:
                if(ev.xcrossing.window == AskwinYesId(askaccept))
                    FocusButton(&xdata, &askaccept.yes, BUTTON_ENTER);
                else if(ev.xcrossing.window == AskwinNoId(askaccept))
                    FocusButton(&xdata, &askaccept.no, BUTTON_ENTER);
                else if(ev.xcrossing.window == ButtonId(quit))
                    FocusButton(&xdata, &quit, BUTTON_ENTER);
                break;
            } /* switch(phase) */
            break;
          case LeaveNotify:
            switch(phase)
            {
              case NORMAL:
                for(i = 0; i < ENEMY_DISP; i++)
                {
                    if(ev.xcrossing.window == ButtonId(enemy[i]))
                    {
                        FocusButton(&xdata, &enemy[i], BUTTON_LEAVE);
                        DrawEnemy(i);
                    }
                }
                if(ev.xcrossing.window == ButtonId(enemypage))
                    FocusButton(&xdata, &enemypage, BUTTON_LEAVE);
                else if(ev.xcrossing.window == ButtonId(enemyflush))
                    FocusButton(&xdata, &enemyflush, BUTTON_LEAVE);
                else if(ev.xcrossing.window == ButtonId(quit))
                    FocusButton(&xdata, &quit, BUTTON_LEAVE);
                break;
              case ASKACCEPT:
                if(ev.xcrossing.window == AskwinYesId(askaccept))
                    FocusButton(&xdata, &askaccept.yes, BUTTON_LEAVE);
                else if(ev.xcrossing.window == AskwinNoId(askaccept))
                    FocusButton(&xdata, &askaccept.no, BUTTON_LEAVE);
                else if(ev.xcrossing.window == ButtonId(quit))
                    FocusButton(&xdata, &quit, BUTTON_LEAVE);
            } /* switch(phase) */
            break;
          case Expose:
            for(i = 0; i < ENEMY_DISP; i++)
            {
                if(ev.xexpose.window == ButtonId(enemy[i]))
                {
                    ExposeButton(&xdata, &enemy[i]);
                    DrawEnemy(i);
                }
            }
            if(ev.xexpose.window == win.id)
                XClearWindow(xdata.display, win.id);
            else if(ev.xexpose.window == enemylist.id)
                ExposeEnemylist();
            else if(ev.xexpose.window == ButtonId(enemypage))
                ExposeButton(&xdata, &enemypage);
            else if(ev.xexpose.window == ButtonId(enemyflush))
                ExposeButton(&xdata, &enemyflush);
            else if(ev.xexpose.window == ButtonId(quit))
                ExposeButton(&xdata, &quit);
            else if(ev.xexpose.window == AskwinId(askaccept))
                XClearWindow(xdata.display, AskwinId(askaccept));
            else if(ev.xexpose.window == AskwinYesId(askaccept))
                ExposeButton(&xdata, &askaccept.yes);
            else if(ev.xexpose.window == AskwinNoId(askaccept))
                ExposeButton(&xdata, &askaccept.no);
            else if(ev.xexpose.window == mesg.win.id)
                ExposeMesg(&xdata, &mesg);
            break;
        }
    }

    return;
}

/* call from server */
static void ServerProc(void)
{
    char buff[BUFFLEN];
    char **wordbuff;

    wordbuff = MallocWordbuff();

    RecvData(buff, BUFFLEN);
    CutStr(buff, wordbuff, ' ');


    if(strcmp(wordbuff[0], "CHALLENGED") == 0)		/* CHALLENGED */
    {
        phase = ASKACCEPT;
        strcpy(name[1], wordbuff[2]);
        strcpy(address[1], wordbuff[3]);
        MapAskwin(&xdata, &askaccept);
#if 0
        SetMesgText(&mesg, 0, wordbuff[2], 7);
        SetMesgText(&mesg, 1, "challenged you.", 7);
        SetMesgText(&mesg, 2, "Accept?", 4);
        SetMesgText(&mesg, 3, "", 7);
#else
	SetMesgTextF(&mesg, 0, 7, "%s ", wordbuff[2]);
	SetMesgText(&mesg, 1, "ĩ路Ƥޤ", 7);
	SetMesgText(&mesg, 2, "ĩޤ?", 7);
	SetMesgText(&mesg, 3, "", 7);
#endif
    }
    else if(strcmp(wordbuff[0], "NOTEXIST") == 0)		/* NOTEXIST */
    {
        phase = NORMAL;
        SetMesgText(&mesg, 0, "Your opponent", 7);
        SetMesgText(&mesg, 1, "does not exist now.", 7);
        SetMesgText(&mesg, 2, "", 7);
        SetMesgText(&mesg, 3, "", 7);
    }
    else if(strcmp(wordbuff[0], "BUSY") == 0)			/* BUSY */
    {
        phase = NORMAL;
        SetMesgText(&mesg, 0, "Your opponent or", 7);
        SetMesgText(&mesg, 1, "you are busy now.", 7);
        SetMesgText(&mesg, 2, "", 7);
        SetMesgText(&mesg, 3, "", 7);
    }
    else if(strcmp(wordbuff[0], "REFUSED") == 0)		/* REFUSED */
    {
        phase = NORMAL;
#if 0
        SetMesgText(&mesg, 0, "Your opponent", 7);
        SetMesgText(&mesg, 1, "refused your", 7);
        SetMesgText(&mesg, 2, "challenge.", 7);
#else
	SetMesgText(&mesg, 0, "", 7);
	SetMesgText(&mesg, 1, "ĩݤޤ", 7);
	SetMesgText(&mesg, 2, "", 7);
#endif
        SetMesgText(&mesg, 3, "", 7);
    }
    else if(strcmp(wordbuff[0], "START") == 0)			/* START */
    {
        if(strcmp(wordbuff[1], "BLACK") == 0)
            side = SENTE;
        else
            side = GOTE;
        done = SE_DONE_START;
    }
    ExposeMesg(&xdata, &mesg);

    return;
}

static void ExposeEnemylist(void)
{
    char buff[BUFFLEN];

    XClearWindow(xdata.display, enemylist.id);
    sprintf(buff, "%2d/%2d", page + 1, page_num);
    XDrawString(xdata.display, enemylist.id, gcset.type[7],
                465, 36, buff, strlen(buff));

    return;
}

static void DrawAllEnemy(void)
{
    int i;

    for(i = 0; i < ENEMY_DISP; i++)
    {
        if(i + ENEMY_DISP * page < enemy_num)
            MapButton(&xdata, &enemy[i]);
        else
            UnmapButton(&xdata, &enemy[i]);
        ExposeButton(&xdata, &enemy[i]);
        DrawEnemy(i);
    }

    return;
}

static void DrawEnemy(int n)
{
    if(n + ENEMY_DISP * page < enemy_num)
    {
        XDrawString(xdata.display, ButtonId(enemy[n]), gcset.type[7],
                    4, 16, enemyname[n + ENEMY_DISP * page],
                    strlen(enemyname[n + ENEMY_DISP * page]));
        XDrawString(xdata.display, ButtonId(enemy[n]), gcset.type[7],
                    4, 36, enemyaddress[n + ENEMY_DISP * page],
                    strlen(enemyaddress[n + ENEMY_DISP * page]));
    }
    return;
}

static void ReadEnemy(void)
{
    char buff[BUFFLEN];
    char *ptr;
    int i;

    SendData("CHALLENGERLIST\n", 15);

    SetMesgTextF(&mesg, 0, 7, "жɼԥꥹȤ");
    SetMesgTextF(&mesg, 1, 7, "׵ᤷޤ");
    SetMesgTextF(&mesg, 2, 7, "ԤäƤޤ...");
    SetMesgTextF(&mesg, 3, 7, "");
    ExposeMesg(&xdata, &mesg);
    if( WaitComm() == FROM_DISPLAY ) {
	DisplayProc();
	if( done != SE_DONE_NORMAL )
	    return;
    }

    RecvData(buff, BUFFLEN);

    SetMesgTextF(&mesg, 0, 7, "жɼԥꥹȤ");
    SetMesgTextF(&mesg, 1, 7, "Ƥޤ...");
    SetMesgTextF(&mesg, 2, 7, "");
    SetMesgTextF(&mesg, 3, 7, "");
    ExposeMesg(&xdata, &mesg);
    XSync(xdata.display, False);

    ptr = buff;
    if( !token_equal(&ptr, "CHALLENGERLIST") ) {
	fprintf(stderr,
		"SERVER ERROR: Server replied `%s' to CHALLENGERLIST.\n",
		buff);
	exit(1);
    }
    if(	(enemy_num = fetch_numerical_token(&ptr)) < 0 || *ptr ) {
	fprintf(stderr,
		"SERVER ERROR: Incompatible parameter: %s\n", buff);
	exit(1);
    }

    for( i = 0; i < enemy_num; i++ ) {
        RecvData(buff, BUFFLEN);
        sscanf(buff, "%d %s %s", &enemyid[i], enemyname[i], enemyaddress[i]);
    }
    RecvData(buff, BUFFLEN);
    if( strcmp(buff, ".") ) {
	fprintf(stderr, "SERVER ERROR: Conflict number of challengers\n");
	exit(1);
    }
	
    page_num = ((enemy_num - 1) / ENEMY_DISP);
    if(page_num < 0)
        page_num = 0;
    page_num++;

    if( enemy_num ) {
	SetMesgTextF(&mesg, 0, 7, "%d̾Υ桼", enemy_num);
	SetMesgTextF(&mesg, 1, 7, "ĩԤäƤޤ");
    }
    else {
	SetMesgTextF(&mesg, 0, 7, "ĩԤäƤ桼");
	SetMesgTextF(&mesg, 1, 7, "Ĥޤ");
    }
    SetMesgTextF(&mesg, 2, 7, "");
    SetMesgTextF(&mesg, 3, 7, "ʸ : λ");
    ExposeMesg(&xdata, &mesg);
    XFlush(xdata.display);

    return;
}

char **MallocWordbuff(void)
{
    char **wordbuff;
    int i;

    wordbuff = (char **)malloc(sizeof(char *) * ENEMY_MAX);
    for(i = 0; i < ENEMY_MAX; i++)
        wordbuff[i] = (char *)malloc(sizeof(char) * BUFFLEN);
	
    return wordbuff;
}

void FreeWordbuff(char **wordbuff)
{
    int i;

    for(i = 0; i < ENEMY_MAX; i++)
        free(wordbuff[i]);
    free(wordbuff);

    return;
}
