#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "all.h"
#include "selectgame.h"
#include "commdata.h"
#include "pixmaps.h"

#define GAME_WIDTH                      512
#define GAME_HEIGHT                     40
#define GAME_DISP                       8
#define GAME_MAX                        64

#define MESG_X                          24
#define MESG_Y                          (GAMELIST_Y + GAMELIST_HEIGHT + 24)

#define GAMELIST_WIDTH                  GAME_WIDTH
#define GAMELIST_HEIGHT                 (GAME_HEIGHT * (GAME_DISP + 1))
#define GAMELIST_X                      24
#define GAMELIST_Y                      24

#define GAMEPAGE_WIDTH                  64
#define GAMEPAGE_HEIGHT                 20
#define GAMEPAGE_X                      (GAMELIST_WIDTH - GAMEPAGE_WIDTH)
#define GAMEPAGE_Y                      0

#define GAMEFLUSH_WIDTH                 64
#define GAMEFLUSH_HEIGHT                40
#define GAMEFLUSH_X                     (GAMEPAGE_X - GAMEFLUSH_WIDTH)
#define GAMEFLUSH_Y                     0

#define QUIT_WIDTH                      128
#define QUIT_HEIGHT                     64
#define QUIT_X                          (WIN_WIDTH - QUIT_WIDTH - 24)
#define QUIT_Y                          (WIN_HEIGHT - QUIT_HEIGHT - 24)

static void ExposeGamelist(void);
static void DrawAllGame(void);
static void DrawGame(int n);
static void ReadGame(void);

char **MallocWordbuff(void);            /* in selectenemy.c */
void FreeWordbuff(char **wordbuff);

static Windata gamelist;
static Button gamepage, gameflush, game[GAME_DISP], quit;
static int gameid[GAME_MAX];
static char gamename[GAME_MAX][2][BUFFLEN];
static char gameaddress[GAME_MAX][2][BUFFLEN];
static int game_num;
static int page_num, page;
static int done;

int SelectGame(void)
{
    XEvent ev;
    char buff[BUFFLEN];
    int i;

    /* gamebox */
    ConstructWindata(&xdata, win.id,
                     GAMELIST_X, GAMELIST_Y,
                     GAMELIST_WIDTH, GAMELIST_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &gamelist);
    XSetWindowBackgroundPixmap(xdata.display, gamelist.id,
			       ENEMYLIST_XPM.pixmap);
    XSelectInput(xdata.display, gamelist.id, ExposureMask);

    /* button */
    CreateButton(&xdata, &gamelist,
                 ENEMYPAGE_XPM.pixmap, ENEMYPAGE_F_XPM.pixmap,
                 GAMEPAGE_X, GAMEPAGE_Y,
                 GAMEPAGE_WIDTH, GAMEPAGE_HEIGHT, 0, &gamepage);
    CreateButton(&xdata, &gamelist,
                 ENEMYFLUSH_XPM.pixmap, ENEMYFLUSH_F_XPM.pixmap,
                 GAMEFLUSH_X, GAMEFLUSH_Y,
                 GAMEFLUSH_WIDTH, GAMEFLUSH_HEIGHT, 0, &gameflush);
    for(i = 0; i < GAME_DISP; i++)
         CreateButton(&xdata, &gamelist,
                      ENEMY_XPM.pixmap, ENEMY_F_XPM.pixmap,
                      0, GAME_HEIGHT * (i + 1),
                      GAME_WIDTH, GAME_HEIGHT, 0, &game[i]);
    CreateButton(&xdata, &win,
                 QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, QUIT_Y,
                 QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);

    /* map window */
    MapButton(&xdata, &gamepage);
    MapButton(&xdata, &gameflush);
    MapButton(&xdata, &quit);
    XMapRaised(xdata.display, gamelist.id);

    /* init data */
    ReadGame();
    page = 0;
    DrawAllGame();

    /* main loop */
    done = SG_DONE_NORMAL;
    while(done == SG_DONE_NORMAL)
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            for(i = 0; i < GAME_DISP; i++)
            {
                if(ev.xbutton.window == ButtonId(game[i]))
                {
                    sprintf(buff, "VIEWGAME %d\n",
                            gameid[i + GAME_DISP * page]);
                    SendData(buff, strlen(buff));
                    RecvData(buff, BUFFLEN);
                    done = SG_DONE_START;
                    strcpy(name[0], gamename[i + GAME_DISP * page][0]);
                    strcpy(name[1], gamename[i + GAME_DISP * page][1]);
                    strcpy(address[0], gameaddress[i + GAME_DISP * page][0]);
                    strcpy(address[1], gameaddress[i + GAME_DISP * page][1]);
                }
            }
            if(ev.xbutton.window == ButtonId(gamepage))
            {
                page++;
                page %= page_num;
                DrawAllGame();
                ExposeGamelist();
            }
            else if(ev.xbutton.window == ButtonId(gameflush))
            {
                ReadGame();
                page = 0;
                DrawAllGame();
                ExposeGamelist();
            }
            else if(ev.xbutton.window == ButtonId(quit))
            {
                done = SG_DONE_QUIT;
                SendData("QUIT\n", 7);
            }
            break;
          case EnterNotify:
            for(i = 0; i < GAME_DISP; i++)
            {
                if(ev.xcrossing.window == ButtonId(game[i]))
                {
                    FocusButton(&xdata, &game[i], BUTTON_ENTER);
                    DrawGame(i);
                }
            }
            if(ev.xcrossing.window == ButtonId(gamepage))
                FocusButton(&xdata, &gamepage, BUTTON_ENTER);
            else if(ev.xcrossing.window == ButtonId(gameflush))
                FocusButton(&xdata, &gameflush, BUTTON_ENTER);
            else if(ev.xcrossing.window == ButtonId(quit))
                FocusButton(&xdata, &quit, BUTTON_ENTER);
            break;
          case LeaveNotify:
            for(i = 0; i < GAME_DISP; i++)
            {
                if(ev.xcrossing.window == ButtonId(game[i]))
                {
                    FocusButton(&xdata, &game[i], BUTTON_LEAVE);
                    DrawGame(i);
                }
            }
            if(ev.xcrossing.window == ButtonId(gamepage))
                FocusButton(&xdata, &gamepage, BUTTON_LEAVE);
            else if(ev.xcrossing.window == ButtonId(gameflush))
                FocusButton(&xdata, &gameflush, BUTTON_LEAVE);
            else if(ev.xcrossing.window == ButtonId(quit))
                FocusButton(&xdata, &quit, BUTTON_LEAVE);
            break;
          case Expose:
            for(i = 0; i < GAME_DISP; i++)
            {
                if(ev.xexpose.window == ButtonId(game[i]))
                {
                    ExposeButton(&xdata, &game[i]);
                    DrawGame(i);
                }
            }
            if(ev.xexpose.window == win.id)
                XClearWindow(xdata.display, win.id);
            else if(ev.xexpose.window == gamelist.id)
                ExposeGamelist();
            else if(ev.xexpose.window == ButtonId(gamepage))
                ExposeButton(&xdata, &gamepage);
            else if(ev.xexpose.window == ButtonId(gameflush))
                ExposeButton(&xdata, &gameflush);
            else if(ev.xexpose.window == ButtonId(quit))
                ExposeButton(&xdata, &quit);
            break;
        }
    }

    XDestroyWindow(xdata.display, gamelist.id);
    XDestroyWindow(xdata.display, quit.win.id);
	
    return done;
}

static void ExposeGamelist(void)
{
    char buff[BUFFLEN];

    XClearWindow(xdata.display, gamelist.id);
    sprintf(buff, "%2d/%2d", page + 1, page_num);
    XDrawString(xdata.display, gamelist.id, gcset.type[7],
                465, 36, buff, strlen(buff));

    return;
}

static void DrawAllGame(void)
{
    int i;

    for(i = 0; i < GAME_DISP; i++)
    {
        if(i + GAME_DISP * page < game_num)
            MapButton(&xdata, &game[i]);
        else
            UnmapButton(&xdata, &game[i]);
        DrawGame(i);
    }

    return;
}

static void DrawGame(int n)
{
    char buff[BUFFLEN];
    
    if(n + GAME_DISP * page < game_num)
    {
        sprintf(buff, "%s vs %s",
                gamename[n + GAME_DISP * page][0],
                gamename[n + GAME_DISP * page][1]);
        XDrawString(xdata.display, ButtonId(game[n]), gcset.type[7],
                    4, 16, buff, strlen(buff));
        sprintf(buff, "");
        XDrawString(xdata.display, ButtonId(game[n]), gcset.type[7],
                    4, 36, buff, strlen(buff));
    }
    return;
}

static void ReadGame(void)
{
    char buff[BUFFLEN];
    char **wordbuff;
    int i;

    wordbuff = MallocWordbuff();

    SendData("GAMELIST\n", 9);
    RecvData(buff, BUFFLEN);
    sscanf(buff, "GAMELIST %d", &game_num);

    for(i = 0; i < game_num; i++)
    {
        RecvData(buff, BUFFLEN);
        sscanf(buff, "%d %s %s %s %s", &gameid[i], gamename[i][0], gameaddress[i][0], gamename[i][1], gameaddress[i][1]);
    }
    RecvData(buff, BUFFLEN);

    page_num = ((game_num - 1) / GAME_DISP);
    if(page_num < 0)
        page_num = 0;
    page_num++;

    FreeWordbuff(wordbuff);

    return;
}
