#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <time.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "all.h"
#include "selectnotation.h"
#include "commdata.h"
#include "drawstring.h"
#include "notation.h"
#include "tuitate.h"
#include "cursor.h"
#include "pixmaps.h"

#define NOTE_WIDTH                      512
#define NOTE_HEIGHT                     40
#define NOTE_DISP                       8
#define NOTE_MAX                        1024

#define MESG_X                          24
#define MESG_Y                          (NOTELIST_Y + NOTELIST_HEIGHT + 24)

#define NOTELIST_WIDTH                  NOTE_WIDTH
#define NOTELIST_HEIGHT                 (NOTE_HEIGHT * (NOTE_DISP + 1))
#define NOTELIST_X                      24
#define NOTELIST_Y                      24

#define NOTEPAGE_WIDTH                  64
#define NOTEPAGE_HEIGHT                 20
#define NOTEPAGE_X                      (NOTELIST_WIDTH - NOTEPAGE_WIDTH)
#define NOTEPAGE_Y                      0

#define NOTEFLUSH_WIDTH                 64
#define NOTEFLUSH_HEIGHT                40
#define NOTEFLUSH_X                     (NOTEPAGE_X - NOTEFLUSH_WIDTH)
#define NOTEFLUSH_Y                     0

#define QUIT_WIDTH                      128
#define QUIT_HEIGHT                     64
#define QUIT_X                          (WIN_WIDTH - QUIT_WIDTH - 24)
#define QUIT_Y                          (WIN_HEIGHT - QUIT_HEIGHT - 24)

static void ExposeNotelist(void);
static void DrawAllNote(void);
static void DrawNote(int n);
static void ReadNote(void);

char **MallocWordbuff(void);            /* in selectenemy.c */
void FreeWordbuff(char **wordbuff);

static Windata notelist;
static Button notepage, noteflush, note[NOTE_DISP], quit;
static int noteid[NOTE_MAX];
static int notebegin[NOTE_MAX];
static int noteend[NOTE_MAX];
static char notename[NOTE_MAX][2][BUFFLEN];
static char noteaddress[NOTE_MAX][2][BUFFLEN];
static int note_num;
static int page_num, page;
static int done;

int SelectNotation(void)
{
    XEvent ev;
    char buff[BUFFLEN];
    int i;

    /* notebox */
    ConstructWindata(&xdata, win.id,
                     NOTELIST_X, NOTELIST_Y,
                     NOTELIST_WIDTH, NOTELIST_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &notelist);
    XSetWindowBackgroundPixmap(xdata.display, notelist.id,
			       ENEMYLIST_XPM.pixmap);
    XSelectInput(xdata.display, notelist.id, ExposureMask);

    /* button */
    CreateButton(&xdata, &notelist,
                 ENEMYPAGE_XPM.pixmap, ENEMYPAGE_F_XPM.pixmap,
                 NOTEPAGE_X, NOTEPAGE_Y, NOTEPAGE_WIDTH, NOTEPAGE_HEIGHT, 0,
                 &notepage);
    CreateButton(&xdata, &notelist,
                 ENEMYFLUSH_XPM.pixmap, ENEMYFLUSH_F_XPM.pixmap,
                 NOTEFLUSH_X, NOTEFLUSH_Y,
                 NOTEFLUSH_WIDTH, NOTEFLUSH_HEIGHT, 0, &noteflush);
    for(i = 0; i < NOTE_DISP; i++)
        CreateButton(&xdata, &notelist, ENEMY_XPM.pixmap, ENEMY_F_XPM.pixmap,
                     0, NOTE_HEIGHT * (i + 1), NOTE_WIDTH, NOTE_HEIGHT, 0,
                     &note[i]);
    CreateButton(&xdata, &win, QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                        QUIT_X, QUIT_Y, QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);

    /* map window */
    MapButton(&xdata, &notepage);
    MapButton(&xdata, &noteflush);
    MapButton(&xdata, &quit);
    XMapRaised(xdata.display, notelist.id);

    /* init data */
    ReadNote();
    page = 0;
    DrawAllNote();

    /* main loop */
    done = SN_DONE_NORMAL;
    while(done == SN_DONE_NORMAL)
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            for(i = 0; i < NOTE_DISP; i++)
            {
                if(ev.xbutton.window == ButtonId(note[i]))
                {
		    sprintf(buff, "VIEWNOTATION %d\n",
			    noteid[i + NOTE_DISP * page]);
		    SendData(buff, strlen(buff));
		    RecvData(buff, sizeof(buff));
		    {
			char *ptr = buff;
			TuParseToken(&ptr);
			TuGetNotation(get_decimal_token(&ptr), &notation);
		    }
                    done = SN_DONE_START;
                    strcpy(name[0], notename[i + NOTE_DISP * page][0]);
                    strcpy(name[1], notename[i + NOTE_DISP * page][1]);
                    strcpy(address[0], noteaddress[i + NOTE_DISP * page][0]);
                    strcpy(address[1], noteaddress[i + NOTE_DISP * page][1]);
		    break;
                }
            }
            if(ev.xbutton.window == ButtonId(notepage))
            {
                page++;
                page %= page_num;
                DrawAllNote();
                ExposeNotelist();
            }
            else if(ev.xbutton.window == ButtonId(noteflush))
            {
                ReadNote();
                page = 0;
                DrawAllNote();
                ExposeNotelist();
            }
            else if(ev.xbutton.window == ButtonId(quit))
            {
                done = SN_DONE_QUIT;
                SendData("QUIT\n", 7);
            }
            break;
          case EnterNotify:
            for(i = 0; i < NOTE_DISP; i++)
            {
                if(ev.xcrossing.window == ButtonId(note[i]))
                {
                    FocusButton(&xdata, &note[i], BUTTON_ENTER);
                    DrawNote(i);
                }
            }
            if(ev.xcrossing.window == ButtonId(notepage))
                FocusButton(&xdata, &notepage, BUTTON_ENTER);
            else if(ev.xcrossing.window == ButtonId(noteflush))
                FocusButton(&xdata, &noteflush, BUTTON_ENTER);
            else if(ev.xcrossing.window == ButtonId(quit))
                FocusButton(&xdata, &quit, BUTTON_ENTER);
            break;
          case LeaveNotify:
            for(i = 0; i < NOTE_DISP; i++)
            {
                if(ev.xcrossing.window == ButtonId(note[i]))
                {
                    FocusButton(&xdata, &note[i], BUTTON_LEAVE);
                    DrawNote(i);
                }
            }
            if(ev.xcrossing.window == ButtonId(notepage))
                FocusButton(&xdata, &notepage, BUTTON_LEAVE);
            else if(ev.xcrossing.window == ButtonId(noteflush))
                FocusButton(&xdata, &noteflush, BUTTON_LEAVE);
            else if(ev.xcrossing.window == ButtonId(quit))
                FocusButton(&xdata, &quit, BUTTON_LEAVE);
            break;
          case Expose:
            for(i = 0; i < NOTE_DISP; i++)
            {
                if(ev.xexpose.window == ButtonId(note[i]))
                {
                    ExposeButton(&xdata, &note[i]);
                    DrawNote(i);
                }
            }
            if(ev.xexpose.window == win.id)
                XClearWindow(xdata.display, win.id);
            else if(ev.xexpose.window == notelist.id)
                ExposeNotelist();
            else if(ev.xexpose.window == ButtonId(notepage))
                ExposeButton(&xdata, &notepage);
            else if(ev.xexpose.window == ButtonId(noteflush))
                ExposeButton(&xdata, &noteflush);
            else if(ev.xexpose.window == ButtonId(quit))
                ExposeButton(&xdata, &quit);
            break;
        }
    }

    XDestroyWindow(xdata.display, notelist.id);
    XDestroyWindow(xdata.display, quit.win.id);
	
    return done;
}

static void ExposeNotelist(void)
{
    char buff[BUFFLEN];

    XClearWindow(xdata.display, notelist.id);
    sprintf(buff, "%2d/%2d", page + 1, page_num);
    XDrawString(xdata.display, notelist.id, gcset.type[7],
                465, 36, buff, strlen(buff));

    return;
}

static void DrawAllNote(void)
{
    int i;

    for(i = 0; i < NOTE_DISP; i++)
    {
        if(i + NOTE_DISP * page < note_num)
        {
            MapButton(&xdata, &note[i]);
            XClearWindow(xdata.display, ButtonId(note[i]));
        }
        else
            UnmapButton(&xdata, &note[i]);
        DrawNote(i);
    }

    return;
}

static void DrawNumberFlushRight(Drawable win, GC gc, int x, int y, int n)
{
    char buf[16];

    sprintf(buf, "%d", n);
    XtuitateDrawEUCStringFlushRight(xdata.display, win, gc, x, y, buf);
}

static void DrawDate(Drawable win, GC gc, int x, int y,
		    time_t start, time_t end)
{
    struct tm *tmp;

    static char *wday[ ] = {
	"", "", "", "", "", "", ""
    };

    tmp = localtime(&start);
    x += 48;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_year + 1900);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "ǯ");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_mon + 1);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_mday);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "(");
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y,
			      wday[tmp->tm_wday]);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, ")");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_hour);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_min);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "ʬ");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, tmp->tm_sec);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, " (");
    end -= start;
    x += 48;
    DrawNumberFlushRight(win, gc, x, y, end / 3600);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, (end / 60) % 60);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, "ʬ");
    x += 24;
    DrawNumberFlushRight(win, gc, x, y, end % 60);
    x = XtuitateDrawEUCString(xdata.display, win, gc, x, y, ")");
}
    
static void DrawNote(int n)
{
    char buff[BUFFLEN];
    
    if(n + NOTE_DISP * page < note_num)
    {
        DrawNumberFlushRight(ButtonId(note[n]), gcset.type[7], 40, 16,
			     noteid[n + NOTE_DISP * page]);
	DrawDate(ButtonId(note[n]), gcset.type[7], 48, 16, 
		 notebegin[n + NOTE_DISP * page],
		 noteend[n + NOTE_DISP * page]);
        sprintf(buff, "%s vs %s",
                notename[n + NOTE_DISP * page][0],
                notename[n + NOTE_DISP * page][1]);
        XtuitateDrawEUCString(xdata.display, ButtonId(note[n]), gcset.type[7],
			      4, 36, buff);
    }
    return;
}

static void ReadNote(void)
{
    char buff[BUFFLEN];
    char **wordbuff;
    int i;

    wordbuff = MallocWordbuff();

    SetCursor(win.id, CSR_WAITING);
    SendData("NOTATIONLIST\n", 13);
    RecvData(buff, BUFFLEN);
    sscanf(buff, "NOTATIONLIST %d", &note_num);
    if( DebuggingProtocol )
	fprintf(stderr, "(Notation: %d entries)\n", note_num);

    for(i = 0; i < note_num; i++)
    {
        RecvData(buff, BUFFLEN);
	if( buff[0] == '.' ) {
	    fprintf(stderr, "Notation-list abnormally terminated\n");
	    exit(1);
	}
        sscanf(buff, "%d %d %d", &noteid[i], &notebegin[i], &noteend[i]);
        RecvData(buff, BUFFLEN);
        sscanf(buff, "%s %s %s %s", notename[i][0], noteaddress[i][0],
               notename[i][1], noteaddress[i][1]);
    }
    RecvData(buff, BUFFLEN);

    page_num = (note_num - 1) / NOTE_DISP;
    if(page_num < 0)
        page_num = 0;
    page_num++;

    SetCursor(win.id, CSR_NORMAL);
    FreeWordbuff(wordbuff);

    return;
}
