#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>

#include "all.h"

int setclient(char *hostname, u_short port)
{
    struct hostent *servhost;
    struct sockaddr_in server = { };
    int s;

    if( hostname == NULL )
	hostname = "localhost";

    while( (servhost = gethostbyname(hostname)) == NULL ) {
	switch( h_errno ) {
	case HOST_NOT_FOUND:
	    fprintf(stderr, "Error: `%s' is unknown.\n", hostname);
	    return -1;
	case NO_DATA:
	    fprintf(stderr,
		    "Error: `%s' is valid but does not have an IP address.\n",
		    hostname);
	    return -1;
	case NO_RECOVERY:
	    fprintf(stderr,
		    "Error: A non-recoverable name server error occurred.\n");
	    return -1;
	case TRY_AGAIN:
	    break;
	default:
	    fprintf(stderr,
		    "Fatal Error: Undefined error occured, h_errno = %d\n",
		    h_errno);
	    exit(1);
	}
	fprintf(stderr,
		"A temporary error occurred on an authoritative name "
		"server. Retrying...\n");
	sleep(1);
    }

    /* bzero((char *)&server, sizeof(server)); */
    server.sin_family = AF_INET;
    server.sin_port = htons(port);
    bcopy(servhost->h_addr, (char *)&server.sin_addr, servhost->h_length);

    if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
	perror("Socket allocation");
	return -1;
    }

    ifDebug('N') {
	int i;
	fprintf(stderr, "Trying %s(%d",
		servhost->h_name, (unsigned char)servhost->h_addr_list[0][0]);
	for( i = 1 ; i < servhost->h_length ; i++ )
	    fprintf(stderr, ".%d", (unsigned char)servhost->h_addr_list[0][i]);
	fprintf(stderr, "):%u ...\n", port);
    }

    if( connect(s, &server, sizeof(server)) ) {
	perror("CONNECT");
	return -1;
    }
    ifDebug('N')
	fprintf(stderr, "Connection established.\n");

    return s;
}
