/*
 *	tuitate.c		X-Tuitate TUP 0.4.0 interface
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "tuitate.h"
#include "commdata.h"
#include "misc.h"
#include "all.h"

#define	REALLOC_UNIT	128

/* prototype declaretions */
static int TuIsViolation(TuToken);

static int TuIsViolation(token)
    TuToken token;
{
    switch( token ) {
    case TJUMPOVER:
    case TCHECK:
    case TDROPON:
    case TCHECKMATEBYDROPPINGPAWN:
	return token;
    }
    return 0;
}

/*
 *	ν (ڽΤẸդǤ뤳Ȥ)
 */
static unsigned char initial_ban[3][9] = {
    {	FU,	FU,	FU,	FU,	FU,	FU,	FU,	FU,	FU   },
    {	EMPTY,	HISHA,	EMPTY,	EMPTY,	EMPTY,	EMPTY,	EMPTY,	KAKU,	EMPTY},
    {	KYOSHA,	KEIMA,	GIN,	KIN,	OU,	KIN,	GIN,	KEIMA,	KYOSHA}
};

static void initialize_ban(unsigned char ban[][9], int mode)
{
    int x, y;

    if( mode )
	for( y = 0 ; y < 3 ; y++ )
	    for( x = 0 ; x < 9 ; x++ )
		ban[y][x] = initial_ban[2 - y][8 - x] | TuSideMask;
    else
	for( y = 0 ; y < 3 ; y++ )
	    for( x = 0 ; x < 9 ; x++ )
		ban[y][x] = 0;
    for(/* y = 3 */; y < 6 ; y++ )
	for( x = 0 ; x < 9 ; x++ )
	    ban[y][x] = 0;
    for(/* y = 6 */; y < 9 ; y++ )
	for( x = 0 ; x < 9 ; x++ )
	    ban[y][x] = initial_ban[y - 6][x];
}

void TuInitializeNotation(TuNotation *notation)
{
    notation->move	= NULL;
    notation->nmove	= 0;
    notation->nalloc	= 0;
    notation->current	= 0;
    notation->count	= 0;
    notation->flags	= 0;
}

static int TuParseCoord(char **ptr);

static int TuParseCoord(ptr)
    char **ptr;
{
    char *p = *ptr;
    int value;

    while( *p == ' ' || *p == '\t' ) p++;

    if( *p < '1' || '9' < *p ) {
	fprintf(stderr, "%s: Can't parse coord: \"%s\"\n", myname, p);
	abort( );
    }
    value = *p++ - '1';
    if( *p && *p != ' ' && *p != '\t' ) {
	fprintf(stderr, "%s: Coordinate overflow:\"%s\"\n", myname, p - 1);
	exit( 1 );
    }
    *ptr = p;
    return value;
}

int TuGetOneMove(TuNotation *notation)
{
    TuToken token;
    TuNotationMove *mptr;
    char *ptr;
    char *errtype;
    char buf[BUFFLEN];

    RecvData(buf, sizeof(buf));
    ptr = buf;
    token = TuParseToken(&ptr);
    if( token == TEND ) {
	token = TuParseToken(&ptr);
	if( token != TBLACK && token != TWHITE && token != TNOSIDE ) {
	    errtype = "winner";
	    goto syntaxerror;
	}
	notation->winner = token;
	token = TuParseToken(&ptr);
	if( token != TCHECKMATE && token != TLOSELIFE && token != TGIVEUP &&
	    token != TEMERGENCY ) {
	    errtype = "reason_of_end";
	    goto syntaxerror;
	}
	notation->reason = token;
	notation->flags |= TuEndMask;
	return 1;
    }

    if( notation->nmove >= notation->nalloc ) {
	notation->nalloc += REALLOC_UNIT;
        notation->move = xrealloc(notation->move,
				  sizeof(TuNotationMove[notation->nalloc]));
    }

    mptr = &notation->move[notation->nmove];
    mptr->flags = 0;

    switch( (mptr->type = token) ) {
    case TMOVE:
	mptr->sx = TuParseCoord(&ptr);
	mptr->sy = TuParseCoord(&ptr);
	mptr->dx = TuParseCoord(&ptr);
	mptr->dy = TuParseCoord(&ptr);
	token = TuParseToken(&ptr);
	switch( token ) {
	case TPROMOTE:
	    mptr->flags |= TuPromotionMask;
	    /*  */
	case TNOTPROMOTE:
	    token = TuParseToken(&ptr);
	}
	switch( token ) {
	case TCHECKED:
	    mptr->flags |= TuCheckedMask;
	    /*  */
	case TNOTCHECKED:
	    token = TuParseToken(&ptr);
	}
	if( (mptr->piece = TuTokenToKoma(token)) ) {
	    token = TuParseToken(&ptr);
	    switch( token ) {
	    case TPROMOTE:
		mptr->piece |= TuPromotionMask;
		/*  */
	    case TNOTPROMOTE:
		token = TuParseToken(&ptr);
	    }
	}
	goto checkviolation;
    case TDROP:
	mptr->dx = TuParseCoord(&ptr);
	mptr->dy = TuParseCoord(&ptr);
	token = TuParseToken(&ptr);
	if( (mptr->piece = TuTokenToKoma(token)) == 0 ) {
	    errtype = "PIECE";
	    goto syntaxerror;
	}
	token = TuParseToken(&ptr);
	switch( token ) {
	case TCHECKED:
	    mptr->flags |= TuCheckedMask;
	    /*  */
	case TNOTCHECKED:
	    token = TuParseToken(&ptr);
	}
    checkviolation:
	if( (mptr->violation = TuIsViolation(token)) )
	    token = TuParseToken(&ptr);
	else
	    notation->flags ^= TuSideMask;
	if( token ) {
	    errtype = "end of line";
	    goto syntaxerror;
	}
	break;
    default:
	fprintf(stderr, "%s: Invalid first token %d\n", myname, (int)token);
	exit( 1 );
    syntaxerror:
	fprintf(stderr, "%s: Message parce error: \"%s\"\n", myname, buf);
	fprintf(stderr, "%s: (parcing %s, \"%s\")\n", myname, errtype, ptr);
	abort( );
    }
    {	/* ȽꡢƱפΥå */
	unsigned int	cnt	= notation->nmove;
	TuNotationMove	*temp	= mptr;

	while( cnt ) {
	    temp--;
	    if( temp->violation == 0 ) {
		if( !(temp->flags & TuSideMask) )	/*  */
		    mptr->flags |= TuSideMask;		/* ϸ */
		if( temp->dx == mptr->dx && temp->dy == mptr->dy )
		    mptr->flags |= TuDoMask;		/* Ʊ */
		break;
	    }
	    cnt--;
	}
	/* while롼פæ  Ƭ  ɬ */
    }

    notation->nmove++;
    return 0;
}

int TuGetNotation(nmove, notation)
    size_t	nmove;
    TuNotation	*notation;
{
    if( nmove == 0 ) {
	fprintf(stderr, "%s: FATAL: no notation data\n",
		myname);
	abort( );
    }
    nmove--;
    TuInitializeNotation(notation);
    notation->move = xmalloc(sizeof(TuNotationMove) * nmove);
    notation->nalloc = nmove;

    while( !TuGetOneMove(notation) );

    return 0;
}

void TuFreeNotation(notation)
    TuNotation *notation;
{
    if( notation != NULL ) {
	free(notation->move);
    }
}

int TuTokenToKoma(token)
    TuToken token;
{
    switch( token ) {
    case TPAWN:		return FU;
    case TLANCE:	return KYOSHA;
    case TKNIGHT:	return KEIMA;
    case TSILVER:	return GIN;
    case TGOLD:		return KIN;
    case TBISHOP:	return KAKU;
    case TROOK:		return HISHA;
    case TKING:		return OU;
    }
    return 0;
}

void TuNotationPatch(TuNotationMove *mptr,
		     unsigned char ban[][9],
		     unsigned char dai[][8],
		     int life[])
{
    unsigned char *my_dai;

    if( mptr->violation ) {
	life[mptr->flags & TuSideMask ? 1 : 0]--;
	return;		/* `ban' is unchanged */
    }

    switch( mptr->type ) {
    case TMOVE:
	ban[mptr->dy][mptr->dx]
	    = ban[mptr->sy][mptr->sx] ^ (mptr->flags & TuPromotionMask);
	ban[mptr->sy][mptr->sx] = 0;
	if( mptr->piece ) {
	    my_dai = dai[mptr->flags & TuSideMask ? 1 : 0];
	    my_dai[mptr->piece & TuRawKomaMask]++;
	}
	break;
    case TDROP:
	ban[mptr->dy][mptr->dx] = mptr->piece | (mptr->flags & TuSideMask);
	my_dai = dai[mptr->flags & TuSideMask ? 1 : 0];
	my_dai[mptr->piece & TuRawKomaMask]--;
	break;
    }
}

void TuNotationRevPatch(TuNotationMove *mptr,
			unsigned char ban[][9],
			unsigned char dai[][8],
			int life[])
{
    unsigned char *my_dai;

    if( mptr->violation ) {
	life[mptr->flags & TuSideMask ? 1 : 0]++;
	return;
    }

    switch( mptr->type ) {
    case TMOVE:
	ban[mptr->sy][mptr->sx]
	    = ban[mptr->dy][mptr->dx] ^ (mptr->flags & TuPromotionMask);
	if( mptr->piece ) {
	    ban[mptr->dy][mptr->dx]
		= mptr->piece | ((mptr->flags & TuSideMask) ^ TuSideMask);
	    my_dai = dai[mptr->flags & TuSideMask ? 1 : 0];
	    my_dai[mptr->piece & TuRawKomaMask]--;
	}
	else
	    ban[mptr->dy][mptr->dx] = 0;
	break;
    case TDROP:
	ban[mptr->dy][mptr->dx] = 0;
	my_dai = dai[mptr->flags & TuSideMask ? 1 : 0];
	my_dai[mptr->piece & TuRawKomaMask]++;
	break;
    }
}
#if 0
static int TuAdjustNotation(TuNotation *notation, unsigned char ban[][9]);


static int TuRetractNotation(TuNotation *notation);
#endif

unsigned int TuSeekNotation(TuNotation *notation,
			    unsigned char ban[][9], unsigned char dai[][8],
			    int life[], int offset, int whence)
{
    unsigned int target;

    assert(whence == SEEK_SET || whence == SEEK_CUR || whence == SEEK_END);

    switch( whence ) {
    case SEEK_END:
	assert( offset <= 0 && notation->nmove >= - offset );
	target = notation->nmove;
	break;
    case SEEK_CUR:
	target = notation->current;
	break;
    case SEEK_SET:
	assert( offset >= 0 && notation->nmove <= offset );
	initialize_ban(ban, 1);
	notation->current = 0;
	target = 0;
	break;
    default:
	fprintf(stderr, "TuSeekNotation(): invalid WHENCE=%d\n", whence);
	exit(1);
    }
    target += offset;

    while( notation->current != target ) {
	if( notation->current < target )
	    TuNotationPatch(&notation->move[notation->current++],
			    ban, dai, life);
	else
	    TuNotationRevPatch(&notation->move[--notation->current],
			       ban, dai, life);
    }
    return notation->current;
}
