#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "commdata.h"
#include "all.h"
#include "ban.h"
#include "tuitate.h"
#include "pixmaps.h"

#define DAI1_X			16
#define DAI1_Y			16

#define BAN_X			(DAI1_X + DAI_WIDTH + 16)
#define BAN_Y			DAI1_Y

#define DAI0_X			(BAN_X + BAN_WIDTH + 16)
#define DAI0_Y			(BAN_Y + BAN_HEIGHT - DAI_HEIGHT)

#define LIFE_X			DAI1_X
#define LIFE_Y			(BAN_Y + BAN_HEIGHT + 16)

#define MESG_X			(LIFE_X + LIFE_WIDTH + 16)
#define MESG_Y			LIFE_Y

#define QUIT_WIDTH		128
#define QUIT_HEIGHT		64
#define QUIT_X			(MESG_X + MESG_WIDTH + 16)
#define QUIT_Y			(MESG_Y + MESG_HEIGHT - QUIT_HEIGHT)

#define CTRL_X			(MESG_X + MESG_WIDTH + 16)
#define CTRL_Y			(QUIT_Y - CTRL_HEIGHT - 16)

#define CTRL_TOP		0
#define CTRL_PREV		1
#define CTRL_NEXT		2
#define CTRL_BOTTOM		3

#define VG_DONE_NORMAL          0
#define VG_DONE_END             1
#define VG_DONE_QUIT            2

#define FlipSide(n)		(1 - n)

static int DisplayProc(void);
static int ServerProc(void);

static void ButtonPressProc(XEvent ev);
static void CrossNotifyProc(XEvent ev, int mode);
static void ExposeProc(XEvent ev);

static void ControlNotation(int ctrl_mode);
static void ForwordControl(void);
static void BackwordControl(void);

void UnpromoteKoma(int *type);                  /* in viewnotation.c */

void PromoteKoma(int *type);			/* in game.c */

char **MallocWordbuff(void);			/* in selectenemy.c */
void FreeWordbuff(char **wordbuff);

static Ban ban;
static Dai dai[2];
static Ctrl ctrl;
static Life life;
static Mesg mesg;
static Button quit;
static Pos pastpos, nowpos;
static int nowtime;            /* now viewing */
static int nowside;
static int done;

void ViewGame(void)
{
    /* create window */
    CreateBan(BAN_X, BAN_Y, BOTH_SIDE, &ban);
    CreateDai(DAI0_X, DAI0_Y, &dai[0]);
    CreateDai(DAI1_X, DAI1_Y, &dai[1]);
    CreateCtrl(CTRL_X, CTRL_Y, &ctrl);
    CreateLife(LIFE_X, LIFE_Y, &life);
    CreateMesg(&xdata, &win, MESG_X, MESG_Y, &mesg);
    CreateButton(&xdata, &win, QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, QUIT_Y, QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);

    /* map window */
    MapBan(&ban);
    MapDai(&dai[0]);
    MapDai(&dai[1]);
    MapCtrl(&ctrl, CTRL_MODE_NONE);
    MapLife(&life);
    MapMesg(&xdata, &mesg);
    MapButton(&xdata, &quit);

    /* init data */
    nowtime = 0;
    nowside = 0;
    nowpos.x = 4;
    nowpos.y = 8;
    notation.nmove = 0;
    notation.nalloc = 0;
    notation.move = NULL;

    done = VG_DONE_NORMAL;
    while(done == VG_DONE_NORMAL)
    {
        XFlush(xdata.display);
        switch(WaitComm(xdata, sock))
        {
          case FROM_DISPLAY:
            DisplayProc();
            break;
          case FROM_SERVER:
            ServerProc();
            break;
          default:
            fprintf(stderr, "FROM_OTHER?\n");
            exit(0);
            break;
        } /* switch(WaitComm()) */
    } /* while() */

    if(done == VG_DONE_END)
    {
        done = VG_DONE_NORMAL;
        while(done == VG_DONE_NORMAL)
                DisplayProc();
    }

    XDestroyWindow(xdata.display, ban.win.id);
    XDestroyWindow(xdata.display, dai[0].win.id);
    XDestroyWindow(xdata.display, dai[1].win.id);
    XDestroyWindow(xdata.display, ctrl.win.id);
    XDestroyWindow(xdata.display, life.win.id);
    XDestroyWindow(xdata.display, mesg.win.id);
    XDestroyWindow(xdata.display, ButtonId(quit));

    return;
}

static int ServerProc(void)
{
    if( TuGetOneMove(&notation) )
        done = VG_DONE_END;

    if( nowtime == notation.nmove - 1 )
    {
        ControlNotation(CTRL_NEXT);
        MapCtrl(&ctrl, CTRL_MODE_BOTTOM);
    }

    return 0;
}

static int DisplayProc(void)
{
    XEvent ev;
	
    while(XPending(xdata.display))
    {
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            ButtonPressProc(ev);
            if(notation.nmove == 0)
                MapCtrl(&ctrl, CTRL_MODE_NONE);
            else if(nowtime == 0)
                MapCtrl(&ctrl, CTRL_MODE_TOP);
            else if(nowtime == notation.nmove)
                MapCtrl(&ctrl, CTRL_MODE_BOTTOM);
            else
                MapCtrl(&ctrl, CTRL_MODE_MID);
            break;
          case EnterNotify:
            CrossNotifyProc(ev, BUTTON_ENTER);
            break;
          case LeaveNotify:
            CrossNotifyProc(ev, BUTTON_LEAVE);
            break;
          case Expose:
            ExposeProc(ev);
            break;
        }
    }
    
    return 0;
}

static void ButtonPressProc(XEvent ev)
{
    if(ev.xbutton.window == ButtonId(quit))
        done = VG_DONE_QUIT;
    else if(ev.xbutton.window == ButtonId(ctrl.top))
        ControlNotation(CTRL_TOP);
    else if(ev.xbutton.window == ButtonId(ctrl.prev))
        ControlNotation(CTRL_PREV);
    else if(ev.xbutton.window == ButtonId(ctrl.next))
        ControlNotation(CTRL_NEXT);
    else if(ev.xbutton.window == ButtonId(ctrl.bottom))
        ControlNotation(CTRL_BOTTOM);
    return;
}

static void CrossNotifyProc(XEvent ev, int mode)
{
    if(ev.xcrossing.window == ButtonId(quit))
        FocusButton(&xdata, &quit, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.top))
        FocusButton(&xdata, &ctrl.top, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.prev))
        FocusButton(&xdata, &ctrl.prev, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.next))
        FocusButton(&xdata, &ctrl.next, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.bottom))
        FocusButton(&xdata, &ctrl.bottom, mode);
    return;
}

static void ExposeProc(XEvent ev)
{
    if(ev.xexpose.window == win.id)
        XClearWindow(xdata.display, win.id);
    else if(ev.xexpose.window == ban.win.id)
    {
        DrawBan(&ban, nowpos, 1);
        ExposeBan(&ban);
    }
    else if(ev.xexpose.window == dai[0].win.id)
    {
        DrawDai(&dai[0], nowpos, 1);
        ExposeDai(&dai[0]);
    }
    else if(ev.xexpose.window == dai[1].win.id)
    {
        DrawDai(&dai[1], nowpos, 1);
        ExposeDai(&dai[1]);
    }
    else if(ev.xexpose.window == life.win.id)
        ExposeLife(&life);
    else if(ev.xexpose.window == mesg.win.id)
        ExposeMesg(&xdata, &mesg);
    else if(ev.xexpose.window == ButtonId(quit))
        ExposeButton(&xdata, &quit);
    else if(ev.xexpose.window == ButtonId(ctrl.top))
        ExposeButton(&xdata, &ctrl.top);
    else if(ev.xexpose.window == ButtonId(ctrl.prev))
        ExposeButton(&xdata, &ctrl.prev);
    else if(ev.xexpose.window == ButtonId(ctrl.next))
        ExposeButton(&xdata, &ctrl.next);
    else if(ev.xexpose.window == ButtonId(ctrl.bottom))
        ExposeButton(&xdata, &ctrl.bottom);
    return;
}

static void ControlNotation(int ctrl_mode)
{
    switch(ctrl_mode)
    {
      case CTRL_TOP:
        while(nowtime > 0)
            BackwordControl();
        break;
      case CTRL_PREV:
        if(nowtime > 0)
            BackwordControl();
        break;
      case CTRL_NEXT:
        if(nowtime < notation.nmove)
            ForwordControl();
        break;
      case CTRL_BOTTOM:
        while(nowtime < notation.nmove)
            ForwordControl();
        break;
    }
    DrawBan(&ban, nowpos, 1);
    ExposeBan(&ban);
    DrawDai(&dai[0], nowpos, 1);
    ExposeDai(&dai[0]);
    DrawDai(&dai[1], nowpos, 1);
    ExposeDai(&dai[1]);
    ExposeLife(&life);
    ExposeMesg(&xdata, &mesg);
    return;
}

static void ForwordControl(void)
{
    char text[BUFFLEN];
    int getkomatype;
    int cy;
    int i;
    TuNotationMove *current;

    getkomatype = EMPTY;
    cy = 0;
    InitMesgText(&mesg);

    if( notation.nmove < nowtime ) {
	current = &notation.move[nowtime];
	switch( current->type ) {
	case TMOVE:
	    pastpos.x = current->sx - 1;
	    pastpos.y = current->sy - 1;
	    nowpos.x  = current->dx - 1;
	    nowpos.y  = current->dy - 1;
	    nowpos.type = pastpos.type =
		ban.data[nowside][pastpos.y][pastpos.x];

	    if( current->promotion == TPROMOTE ) {
		PromoteKoma(&nowpos.type);
		sprintf(text, "%s (%d,%d) %s Promote.", name[nowside],
			nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
	    }
	    else
		sprintf(text, "%s (%d,%d) %s.", name[nowside],
			nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
	    SetMesgText(&mesg, cy++, text, 7);

	    if( current->check == TCHECKED )
		SetMesgText(&mesg, cy++, "Check.", 4);

	    if( current->piece ) {
		getkomatype = i = TuTokenToKoma(current->piece);
		if( current->p_promotion == TPROMOTE )
		    sprintf(text, "%s got P_%s.", name[nowside], komaname[i]);
		else
		    sprintf(text, "%s got %s.", name[nowside], komaname[i]);
		SetMesgText(&mesg, cy++, text, 4);
	    }

	    switch( current->violation ) {
	    case TJUMPOVER:
		life.life[nowside]--;
		SetMesgText(&mesg, cy++, "Violate : JUMPOVER", 2);
		break;
	    case TCHECK:
		life.life[nowside]--;
		SetMesgText(&mesg, cy++, "Violate : CHECK", 2);
		break;
	    default:
		ban.data[nowside][pastpos.y][pastpos.x] = EMPTY;
		ban.data[nowside][nowpos.y][nowpos.x] = nowpos.type;
		ban.data[FlipSide(nowside)][nowpos.y][nowpos.x] = EMPTY;
		if(getkomatype != EMPTY)
		    dai[nowside].data[nowside][getkomatype]++;
		nowside = FlipSide(nowside);
		break;
	    }
	    break;	/* case TMOVE */

	case TDROP:
	    nowpos.x  = current->dx - 1;
	    nowpos.y  = current->dy - 1;

	    nowpos.type = i = TuTokenToKoma(current->piece);
	    sprintf(text, "%s (%d,%d) %s Drop.", name[nowside],
		    nowpos.x + 1, nowpos.y + 1, komaname[i]);
	    SetMesgText(&mesg, cy++, text, 7);

	    if( current->check == TCHECKED )
		SetMesgText(&mesg, cy++, "Check.", 4);

	    switch( current->violation ) {
	    case TDROPON:
		life.life[nowside]--;
		SetMesgText(&mesg, cy++, "Violate : DROPON", 2);
		break;
	    case TCHECKMATEBYDROPPINGPAWN:
		life.life[nowside]--;
		SetMesgText(&mesg, cy++, "Violate : CMbyDP", 2);
		break;
	    default:
		dai[nowside].data[nowside][nowpos.type]--;
		ban.data[nowside][nowpos.y][nowpos.x] = nowpos.type;
		nowside = FlipSide(nowside);
	    }
	    break;	/* case TDROP */
	}
    }
    else {
	switch( notation.winner ) {
	case TBLACK:
	    if(side == SENTE)
		sprintf(text, "End : Winner is %s.", name[0]);
	    else
		sprintf(text, "End : Winner is %s.", name[1]);
	    break;
	case TWHITE:
	    if(side == SENTE)
		sprintf(text, "End : Winner is %s.", name[1]);
	    else
		sprintf(text, "End : Winner is %s.", name[0]);
	    break;
	default:	/* TNOSIDE */
	    sprintf(text, "End : Drawgame.");
	    SetMesgText(&mesg, cy++, text, 4);
	    sprintf(text, "Reason : %s.", "???");
	    SetMesgText(&mesg, cy++, text, 7);
	}
	nowtime++;
    }
}

static void BackwordControl(void)
{
    char text[BUFFLEN];
    int getkomatype;
    int cy;
    int i;
    TuNotationMove *current;

    nowtime--;

    getkomatype = EMPTY;
    cy = 0;
    InitMesgText(&mesg);

    current = &notation.move[nowtime];
    switch( current->type ) {
    case TMOVE:
	pastpos.x = current->sx - 1;
	pastpos.y = current->sy - 1;
	nowpos.x  = current->dx - 1;
	nowpos.y  = current->dy - 1;
	if( !current->violation ) {
	    nowside = FlipSide(nowside);
	    nowpos.type = pastpos.type = ban.data[nowside][nowpos.y][nowpos.x];

	    if( current->promotion == TPROMOTE ) {
		UnpromoteKoma(&pastpos.type);
		sprintf(text, "%s (%d,%d) %s Promote.", name[nowside],
			nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
	    }
	    else
		sprintf(text, "%s (%d,%d) %s.", name[nowside],
			nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
	    SetMesgText(&mesg, cy++, text, 7);
        }
        else {
            nowpos.type = pastpos.type
		= ban.data[nowside][pastpos.y][pastpos.x];
            if( current->promotion == TPROMOTE ) {
                sprintf(text, "%s (%d,%d) %s Promote.", name[nowside],
                        nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
            }
            else
                sprintf(text, "%s (%d,%d) %s.", name[nowside],
                        nowpos.x + 1, nowpos.y + 1, komaname[pastpos.type]);
            SetMesgText(&mesg, cy++, text, 7);
        }

        if( current->check == TCHECKED )
            SetMesgText(&mesg, cy++, "Check.", 4);

	if( current->piece ) {
	    getkomatype = i = TuTokenToKoma(current->piece);
	    if( current->p_promotion == TPROMOTE ) {
		PromoteKoma(&getkomatype);
		sprintf(text, "%s got P_%s.", name[nowside], komaname[i]);
	    }
	    else
		sprintf(text, "%s got %s.", name[nowside], komaname[i]);
	    SetMesgText(&mesg, cy++, text, 4);
	}

        switch( current->violation ) {
	case TJUMPOVER:
            life.life[nowside]++;
            SetMesgText(&mesg, cy++, "Violate : JUMPOVER", 2);
	    break;
	case TCHECK:
            life.life[nowside]++;
            SetMesgText(&mesg, cy++, "Violate : CHECK", 2);
	    break;
	default:
            ban.data[nowside][pastpos.y][pastpos.x] = pastpos.type;
            ban.data[nowside][nowpos.y][nowpos.x] = EMPTY;
            ban.data[FlipSide(nowside)][nowpos.y][nowpos.x] = getkomatype;
            if(getkomatype >= TO && getkomatype != EMPTY)
                UnpromoteKoma(&getkomatype);
            if(getkomatype != EMPTY)
                dai[nowside].data[nowside][getkomatype]--;
        }
	break;	/* case MOVE */
    case TDROP:
        if( !current->violation )
            nowside = FlipSide(nowside);

        nowpos.x = current->dx - 1;
        nowpos.y = current->dy - 1;

	nowpos.type = i = TuTokenToKoma(current->piece);
	sprintf(text, "%s (%d,%d) %s Drop.", name[nowside],
		nowpos.x + 1, nowpos.y + 1, komaname[i]);
	SetMesgText(&mesg, cy++, text, 7);

        if( current->check == TCHECKED )
            SetMesgText(&mesg, cy++, "Check.", 4);

	switch( current->violation ) {
	case TDROPON:
            life.life[nowside]++;
            SetMesgText(&mesg, cy++, "Violate : DROPON", 2);
	    break;
        case TCHECKMATEBYDROPPINGPAWN:
            life.life[nowside]++;
            SetMesgText(&mesg, cy++, "Violate : CMbyDP", 2);
	    break;
	default:
            dai[nowside].data[nowside][nowpos.type]++;
            ban.data[nowside][nowpos.y][nowpos.x] = EMPTY;
	    break;
        }
    } /* case TDROP */
#if 0
    else if(strcmp(wordbuff[0], "END") == 0)
    {
        if(strcmp(wordbuff[1], "BLACK") == 0)
        {
            if(side == SENTE)
                sprintf(text, "End : Winner is %s.", name[0]);
            else
                sprintf(text, "End : Winner is %s.", name[1]);
        }
        else if(strcmp(wordbuff[1], "WHITE") == 0)
        {
            if(side == SENTE)
                sprintf(text, "End : Winner is %s.", name[1]);
            else
                sprintf(text, "End : Winner is %s.", name[0]);
        }
        else if(strcmp(wordbuff[1], "NOSIDE") == 0)
            sprintf(text, "End : Drawgame.");
        SetMesgText(&mesg, cy++, text, 4);
        sprintf(text, "Reason : %s.", wordbuff[2]);
        SetMesgText(&mesg, cy++, text, 7);
    }
#endif
}
