#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "config.h"

#include "xdata.h"
#include "xtool.h"
#include "commdata.h"
#include "all.h"
#include "ban.h"
#include "viewnotation.h"
#include "drawstring.h"
#include "tuitate.h"
#include "strings.h"
#include "kifu.h"
#include "misc.h"
#include "pixmaps.h"

#define DAI1_X			16
#define DAI1_Y			16

#define BAN_X			(DAI1_X + DAI_WIDTH + 16)
#define BAN_Y			DAI1_Y

#define DAI0_X			(BAN_X + BAN_WIDTH + 16)
#define DAI0_Y			(BAN_Y + BAN_HEIGHT - DAI_HEIGHT)

#define LIFE_X			DAI1_X
#define LIFE_Y			(BAN_Y + BAN_HEIGHT + 16)

#define MESG_X			(LIFE_X + LIFE_WIDTH + 16)
#define MESG_Y			LIFE_Y

#define QUIT_WIDTH		128
#define QUIT_HEIGHT		64
#define QUIT_X			(MESG_X + MESG_WIDTH + 16)
#define QUIT_Y			(MESG_Y + MESG_HEIGHT - QUIT_HEIGHT)

#define CTRL_X			(MESG_X + MESG_WIDTH + 16)
#define CTRL_Y			(QUIT_Y - CTRL_HEIGHT - 16)

#define CTRL_TOP		0
#define CTRL_PREV		1
#define CTRL_NEXT		2
#define CTRL_BOTTOM		3

static void ButtonPressProc(XEvent ev);
static void CrossNotifyProc(XEvent ev, int mode);
static void ExposeProc(XEvent *ev);
static int ServerProc(void);

static void ControlNotation(int ctrl_mode);
static void ForwardControl(void);
static void BackwardControl(void);
static void display_current_move(TuNotation *notation, unsigned char ban[][9]);

static Ban ban;
static Dai dai[2];
static Ctrl ctrl;
static Life life;
static Mesg mesg;
static Button quit;
static Pos pastpos, nowpos;
static int nowside;
static int done;

void ViewNotation(void)
{
    XEvent ev;
	
    /* create window */
    CreateBan(BAN_X, BAN_Y, BOTH_SIDE, &ban);
    CreateDai(DAI0_X, DAI0_Y, &dai[0]);
    CreateDai(DAI1_X, DAI1_Y, &dai[1]);
    CreateCtrl(CTRL_X, CTRL_Y, &ctrl);
    CreateLife(LIFE_X, LIFE_Y, &life);
    CreateMesg(&xdata, &win, MESG_X, MESG_Y, &mesg);
    CreateButton(&xdata, &win, QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, QUIT_Y, QUIT_WIDTH, QUIT_HEIGHT, 1, &quit);

    /* map window */
    MapBan(&ban);
    MapDai(&dai[0]);
    MapDai(&dai[1]);
    MapCtrl(&ctrl, CTRL_TOP);
    MapLife(&life);
    MapMesg(&xdata, &mesg);
    MapButton(&xdata, &quit);

    /* init data */
    nowside = (side == SENTE) ? 0 : 1;
    nowpos.x = 4;
    nowpos.y = (side == SENTE) ? 8 : 0;
    notation.current = -1;
    display_current_move(&notation, ban.data);
    notation.current = 0;

    done = 0;
    while(done == 0)
    {
	if( !(notation.flags & TuEndMask) && WaitComm() == FROM_SERVER ) {
	    ServerProc();
	}
	else {
	    XNextEvent(xdata.display, &ev);
	    switch(ev.type) {
	    case ButtonPress:
		ButtonPressProc(ev);
		if( notation.current == 0 )
		    MapCtrl(&ctrl, CTRL_MODE_TOP);
		else if( notation.current > notation.nmove)
		    MapCtrl(&ctrl, CTRL_MODE_BOTTOM);
		else
		    MapCtrl(&ctrl, CTRL_MODE_MID);
		break;
	    case EnterNotify:
		CrossNotifyProc(ev, BUTTON_ENTER);
		break;
	    case LeaveNotify:
		CrossNotifyProc(ev, BUTTON_LEAVE);
		break;
	    case Expose:
		ExposeProc(&ev);
		break;
	    }
	}
    }

    TuFreeNotation(&notation);

    XDestroyWindow(xdata.display, ban.win.id);
    XDestroyWindow(xdata.display, dai[0].win.id);
    XDestroyWindow(xdata.display, dai[1].win.id);
    XDestroyWindow(xdata.display, ctrl.win.id);
    XDestroyWindow(xdata.display, life.win.id);
    XDestroyWindow(xdata.display, mesg.win.id);
    XDestroyWindow(xdata.display, ButtonId(quit));

    return;
}

static int ServerProc(void)
{
    TuGetOneMove(&notation);

    if( notation.current >= notation.nmove - 1 ) {
        ControlNotation(CTRL_NEXT);
        MapCtrl(&ctrl, CTRL_MODE_BOTTOM);
    }

    return 0;
}

static void ButtonPressProc(XEvent ev)
{
    if(ev.xbutton.window == ButtonId(quit))
        done = 1;
    else if(ev.xbutton.window == ButtonId(ctrl.top))
	ControlNotation(CTRL_TOP);
    else if(ev.xbutton.window == ButtonId(ctrl.prev))
	ControlNotation(CTRL_PREV);
    else if(ev.xbutton.window == ButtonId(ctrl.next))
	ControlNotation(CTRL_NEXT);
    else if(ev.xbutton.window == ButtonId(ctrl.bottom))
	ControlNotation(CTRL_BOTTOM);
    return;
}

static void CrossNotifyProc(XEvent ev, int mode)
{
    if(ev.xcrossing.window == ButtonId(quit))
	FocusButton(&xdata, &quit, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.top))
	FocusButton(&xdata, &ctrl.top, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.prev))
	FocusButton(&xdata, &ctrl.prev, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.next))
	FocusButton(&xdata, &ctrl.next, mode);
    else if(ev.xcrossing.window == ButtonId(ctrl.bottom))
	FocusButton(&xdata, &ctrl.bottom, mode);
    return;
}

static void ExposeProc(XEvent *ev)
{
    if(ev->xexpose.window == win.id)
        XClearWindow(xdata.display, win.id);
    else if(ev->xexpose.window == ban.win.id)
    {
        DrawBan(&ban, nowpos, 1);
        ExposeBan(&ban);
    }
    else if(ev->xexpose.window == dai[0].win.id)
    {
        DrawDai(&dai[0], nowpos, 1);
        ExposeDai(&dai[0]);
    }
    else if(ev->xexpose.window == dai[1].win.id)
    {
        DrawDai(&dai[1], nowpos, 1);
        ExposeDai(&dai[1]);
    }
    else if(ev->xexpose.window == life.win.id)
        ExposeLife(&life);
    else if(ev->xexpose.window == mesg.win.id)
        ExposeMesg(&xdata, &mesg);
    else if(ev->xexpose.window == ButtonId(quit))
        ExposeButton(&xdata, &quit);
    else if(ev->xexpose.window == ButtonId(ctrl.top))
        ExposeButton(&xdata, &ctrl.top);
    else if(ev->xexpose.window == ButtonId(ctrl.prev))
        ExposeButton(&xdata, &ctrl.prev);
    else if(ev->xexpose.window == ButtonId(ctrl.next))
        ExposeButton(&xdata, &ctrl.next);
    else if(ev->xexpose.window == ButtonId(ctrl.bottom))
        ExposeButton(&xdata, &ctrl.bottom);
    return;
}

static void ControlNotation(int ctrl_mode)
{
    switch(ctrl_mode) {
    case CTRL_TOP:
	if( notation.current == 0 )
	    ding();
	else
	    while(notation.current > 0)
		BackwardControl();
	break;
    case CTRL_PREV:
	if(notation.current > 0)
	    BackwardControl();
	else
	    ding();
	break;
    case CTRL_NEXT:
	if(notation.current <= notation.nmove)
	    ForwardControl();
	else
	    ding();
	break;
    case CTRL_BOTTOM:
	if( notation.current > notation.nmove )
	    ding();
	else
	    while( notation.current <= notation.nmove )
		ForwardControl();
	break;
    }
    DrawBan(&ban, nowpos, 1);
    ExposeBan(&ban);
    DrawDai(&dai[0], nowpos, 1);
    ExposeDai(&dai[0]);
    DrawDai(&dai[1], nowpos, 1);
    ExposeDai(&dai[1]);
    ExposeLife(&life);
    ExposeMesg(&xdata, &mesg);
}

static void display_current_move(TuNotation *notation, unsigned char ban[][9])
{
    TuNotationMove *mptr;
    char text[64];
    char sente_gote;
    int cy = 0;

    InitMesgText(&mesg);

    if( notation->current < notation->nmove ) {
	mptr = &notation->move[notation->current];
	sente_gote = mptr->flags & TuSideMask ? CodeGote : CodeSente;
	SetMesgText(&mesg, cy++, composeKifu(text, mptr, ban), 7);
	switch( mptr->violation ) {
	case TJUMPOVER:
	    SetMesgText(&mesg, cy++, "ζӱۤ (ȿ§)", 2);
	    break;
	case TCHECK:
	    SetMesgText(&mesg, cy++, "ե (ȿ§)", 2);
	    break;
	case TDROPON:
	    SetMesgText(&mesg, cy++, "ζξǤȤȤ (ȿ§)", 2);
	    break;
	case TCHECKMATEBYDROPPINGPAWN:
	    SetMesgText(&mesg, cy++, "Ǥͤ (ȿ§)", 2);
	    break;
	default:
	    if( mptr->type == TMOVE && mptr->piece )
		SetMesgTextF(&mesg, cy++, 4, "%c %s%s%s ޤ",
			     sente_gote,
			     jissuuji[mptr->dx + 1], kansuuji[mptr->dy + 1],
			     koma_abbrev[mptr->piece & TuMyKomaMask]);
	    if( mptr->flags & TuCheckedMask )
		SetMesgTextF(&mesg, cy++, 4, "%c ؤޤ",
			     sente_gote);
	}
	if( !(notation->flags & TuEndMask)
	    && notation->current == notation->nmove - 1 )
	    SetMesgTextF(&mesg, cy++, 7, "%s(%s)ͤƤޤ...",
			 notation->flags & TuSideMask ? "" : "",
			 name[notation->flags & TuSideMask ? 0 : 1]);
    }
    else if( notation->current == notation->nmove ) {
	if( notation->flags & TuEndMask ) {
	    if( notation->reason == TEMERGENCY ) {
		SetMesgText(&mesg, cy++, "жΥ饤Ȥ", 7);
		SetMesgText(&mesg, cy++, "Ф³", 7);
		SetMesgText(&mesg, cy++, "ڤޤ", 7);
	    }
	    else {
		char *winner, *loser, *reason;

		switch( notation->winner ) {
		case TBLACK:
		    if(side == SENTE) {
			winner = name[0];
			loser  = name[1];
		    }
		    else {
			winner = name[1];
			loser  = name[0];
		    }
		    goto setmessage;
		case TWHITE:
		    if(side == SENTE) {
			winner = name[1];
			loser  = name[0];
		    }
		    else {
			winner = name[0];
			loser  = name[1];
		    }
		setmessage:
		    SetMesgTextF(&mesg, cy++, 4, "λ: %s ޤ",
				 winner);
		    switch( notation->reason ) {
		    case TCHECKMATE:
			reason = "()";
			break;
		    case TLOSELIFE:
			reason = "(%s ȿ§9ãޤ)";
			break;
		    case TGIVEUP:
			reason = "(%s λޤ)";
			break;
		    default:
			reason = "???";
		    }
		    SetMesgTextF(&mesg, cy++, 7, reason, loser);
		    break;
		default:
		    SetMesgText(&mesg, cy++, "λ (ʬ)", 4);
		    break;
		}
	    }
	}
	else {
	    SetMesgTextF(&mesg, cy++, 7, "%s(%s)ͤƤޤ...",
			 notation->flags & TuSideMask ? "" : "",
			 name[notation->flags & TuSideMask ? 0 : 1]);
	}
    }
    else {
	SetMesgTextF(&mesg, cy++, 7, "%c %s (%s)",
		     CodeSente, name[0], address[0]);
	SetMesgTextF(&mesg, cy++, 7, "%c %s (%s)",
		     CodeGote, name[1], address[1]);
	SetMesgTextF(&mesg, cy++, 7, " (%u)", notation->nmove);
    }
}

static void ForwardControl(void)
{
    nowpos.x = notation.move[notation.current].dx;
    nowpos.y = notation.move[notation.current].dy;
    TuNotationPatch(&notation.move[notation.current],
		    ban.data,
		    dai[notation.move[notation.current].flags & TuSideMask
		       ? 1 : 0].data, life.life);
    if( notation.move[notation.current].violation == 0 )
	notation.count++;
    display_current_move(&notation, ban.data);
    notation.current++;
}

static void BackwardControl(void)
{
    notation.current--;
    nowpos.x = notation.move[notation.current].dx;
    nowpos.y = notation.move[notation.current].dy;
    TuNotationRevPatch(&notation.move[notation.current],
		       ban.data,
		       dai[notation.move[notation.current].flags & TuSideMask
			  ? 1 : 0].data, life.life);
    if( notation.move[notation.current].violation == 0 )
	notation.count--;
    notation.current--;
    display_current_move(&notation, ban.data);
    notation.current++;
}
