#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "xdata.h"

void RegistGCSetNamed(Xdata *xdata, GCSet *gcset, char *colname, int num)
{
    XColor dummy, screen;

    if(!XAllocNamedColor(xdata->display, xdata->cmap, colname, &screen, &dummy))
    {
        fprintf(stderr, "in RegistGCSet: cannot allocate the color %s.\n", colname);
    }
    XSetForeground(xdata->display, gcset->type[num], screen.pixel);

    return;
}

void RegistGCSet(Xdata *xdata, GCSet *gcset,
                 int red, int green, int blue, int num)
{
    XColor exact;

    exact.red = red;
    exact.green = green;
    exact.blue = blue;
    if(!XAllocColor(xdata->display, xdata->cmap, &exact))
    {
        fprintf(stderr, "in RegistGCSet: cannot allocate the color %d %d %d.\n", red, green, blue);
    }
    XSetForeground(xdata->display, gcset->type[num], exact.pixel);

    return;
}

void InitGCSet(Xdata *xdata, GCSet *gcset)
{
    int i;

    for(i = 0; i < MAX_GCSET; i++)
        gcset->type[i] = XCreateGC(xdata->display, xdata->root.id, 0, 0);
	
    RegistGCSetNamed(xdata, gcset, "black", 0);
    RegistGCSetNamed(xdata, gcset, "blue", 1);
    RegistGCSetNamed(xdata, gcset, "red", 2);
    RegistGCSetNamed(xdata, gcset, "purple", 3);
    RegistGCSetNamed(xdata, gcset, "green", 4);
    RegistGCSetNamed(xdata, gcset, "cyan", 5);
    RegistGCSetNamed(xdata, gcset, "yellow", 6);
    RegistGCSetNamed(xdata, gcset, "white", 7);
}

Xdata ConstructXdata(char *displayname)
{
    Xdata xdata;
    int dummy;
    Window rr;

    if( (xdata.display = XOpenDisplay(displayname)) == NULL ) {
	fprintf(stderr, "Can't open display.\n");
	exit(1);
    }
    xdata.screen = DefaultScreen(xdata.display);
    xdata.gc = DefaultGC(xdata.display, xdata.screen);
    xdata.visual = DefaultVisual(xdata.display, xdata.screen);
    xdata.cmap = DefaultColormap(xdata.display, xdata.screen);
    xdata.depth = DefaultDepth(xdata.display, xdata.screen);
    xdata.root.id = DefaultRootWindow(xdata.display);
    xdata.root.parent = xdata.root.id;
    xdata.root.x = 0;
    xdata.root.y = 0;
    XGetGeometry(xdata.display, xdata.root.id, &rr,
                 &dummy, &dummy,
                 &(xdata.root.width), &(xdata.root.height),
                 &dummy, &dummy);
    xdata.root.bw = 0;

    /* This is not sure. */
    xdata.root.bg = BlackPixel(xdata.display, xdata.screen);
    xdata.root.fg = WhitePixel(xdata.display, xdata.screen);

    return xdata;
}

void ConstructWindata(Xdata *xdata, Window parent,
                         int x, int y,
                         unsigned int width, unsigned int height,
                         unsigned int bw,
                         unsigned long fg, unsigned long bg,
                         int over, Windata *win)
{
    XSetWindowAttributes attr;
    unsigned long valuemask;

    win->parent = parent;
    win->x = x;
    win->y = y;
    win->width = width;
    win->height = height;
    win->bw = bw;
    win->fg = fg;
    win->bg = bg;

    win->id = XCreateSimpleWindow(xdata->display, parent,
                                 x, y, width, height, bw,
                                 fg, bg);

    if(over)
    {
        valuemask = 0;
        attr.override_redirect = True;
        valuemask |= CWOverrideRedirect;
        XChangeWindowAttributes(xdata->display, win->id, valuemask, &attr);
    }
}

void CopyWindata(Xdata *xdata, Windata *dist, Windata *src)
{
    ConstructWindata(xdata, src->parent,
                     src->x, src->y, src->width, src->height, src->bw,
                     src->fg, src->bg, 0, dist);
    return;
}

void MakeSizeHints(Xdata *xdata, Windata *win, char *winname, char *iconname,
		     Pixmap iconpixmap, int argc, char **argv)
{
    XSizeHints hint;

    hint.x = 0;
    hint.y = 0;
    hint.width = win->width;
    hint.height = win->height;
    hint.flags = PPosition | PSize;
    XSetStandardProperties(xdata->display, win->id, winname, iconname,
                           None, argv, argc, &hint);

    return;
}

void MakeClassHint(xdata, win, name, class)
    Xdata  *xdata;
    Windata *win;
    char *name;
    char *class;
{
    XClassHint classhint;
    classhint.res_name  = name;
    classhint.res_class = class;
    XSetClassHint(xdata->display, win->id, &classhint);
}
