#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/keysym.h>

#include "config.h"

#include "all.h"
#include "xdata.h"
#include "xtool.h"
#include "drawstring.h"
#include "pixmaps.h"

#define BUFFLEN			256

/* for xinput */
#define TITLE_Y			40
#define DISP_WIDTH		300
#define DISP_HEIGHT		24
#define DISP_OFFSET_X	4
#define DISP_OFFSET_Y	20
#define DISP_X			((INPUTWIN_WIDTH - DISP_WIDTH) / 2)
#define DISP_Y			60
#define CURSOR_WIDTH	8
#define CURSOR_HEIGHT	8
/* for xbutton */
#define BUTTON_WIDTH	128
#define BUTTON_HEIGHT	64
/* for xaskwin */
#define ITEM_WIDTH		56
#define ITEM_HEIGHT		40
#define YES_X			6
#define YES_Y			48
#define NO_X			66
#define NO_Y			YES_Y
#define ASKWIN_MESG_Y	32
#define ASKWIN_POINTER_X	(ASKWIN_WIDTH / 2)
#define ASKWIN_POINTER_Y	24

/* for xinput */
void ExposeInputwin(Xdata xdata, Windata inputwin, GC gc,
                    XFontStruct *fontstruct, char *title);
void ExposeDisp(Xdata xdata, Windata disp, GC gc, XFontStruct *fontstruct,
                char *ret);

/* funcs */
void CreateMesg(Xdata *xdata, Windata *win, int x, int y, Mesg *mesg)
{
    static Font font = None;
    int i;

    ConstructWindata(xdata, win->id,
                     x, y, MESG_WIDTH, MESG_HEIGHT, 1,
                     xdata->root.fg, xdata->root.bg, 0, &mesg->win);
    XSetWindowBackgroundPixmap(xdata->display, mesg->win.id, MESG_XPM.pixmap);
    XSelectInput(xdata->display, mesg->win.id, ExposureMask);

    InitGCSet(xdata, &mesg->gcset);
    if( font == None ) {
        font = XLoadFont(xdata->display,
                     "-adobe-times-bold-r-normal-*-*-240-*-*-p-*-iso8859-*");
    }
    for(i = 0; i < 8; i++)
        XSetFont(xdata->display, mesg->gcset.type[i], font);

    InitMesgText(mesg);
}

void InitMesgText(Mesg *mesg)
{
    int i;
    for(i = 0; i < MESG_LINE; i++)
    {
        strcpy(mesg->text[i], "");
        mesg->color[i] = 7;
    }
    return;
}

void MapMesg(Xdata *xdata, Mesg *mesg)
{
    XMapRaised(xdata->display, mesg->win.id);
    return;
}

void UnmapMesg(Xdata *xdata, Mesg *mesg)
{
    XUnmapWindow(xdata->display, mesg->win.id);
    return;
}

void ExposeMesg(Xdata *xdata, Mesg *mesg)
{
    int i;

    XClearWindow(xdata->display, mesg->win.id);
    for(i = 0; i < MESG_LINE; i++)
	XtuitateDrawEUCString(xdata->display, mesg->win.id,
			      mesg->gcset.type[mesg->color[i]],
			      MESG_OFFSET_X,
			      MESG_OFFSET_Y + i * MESG_LINE_HEIGHT,
			      mesg->text[i]);
    return;
}

void SetMesgText(Mesg *mesg, int line, char *text, int color)
{
    strcpy(mesg->text[line], text);
    mesg->color[line] = color;
    return;
}

/* This function was added by K.Takai */
void SetMesgTextF(Mesg *mesg, int line, int color, const char *fmt, ...)
{
    va_list ap;

    mesg->color[line] = color;
    va_start(ap, fmt);
#ifdef	__sun__
    vsprintf(mesg->text[line], fmt, ap);
#else
    vsnprintf(mesg->text[line], sizeof(mesg->text), fmt, ap);
#endif
    va_end(ap);
}

void MoveAskwin(Xdata *xdata, Askwin *askwin, Windata *parent)
{
    Window dummy_root, dummy_child;
    int root_x, root_y, win_x, win_y, keybutton;

    XQueryPointer(xdata->display, askwin->win.id,
                  &dummy_root, &dummy_child,
                  &root_x, &root_y, &win_x, &win_y, &keybutton);
    askwin->win.x += win_x - ASKWIN_POINTER_X;
    askwin->win.y += win_y - ASKWIN_POINTER_Y;
    if(askwin->win.x < 0)
        askwin->win.x = 0;
    else if(askwin->win.x >= parent->width - ASKWIN_WIDTH)
        askwin->win.x = parent->width - ASKWIN_WIDTH - 1;
    if(askwin->win.y < 0)
        askwin->win.y = 0;
    else if(askwin->win.y >= parent->height - ASKWIN_HEIGHT)
        askwin->win.y = parent->height - ASKWIN_HEIGHT - 1;
    XMoveWindow(xdata->display, askwin->win.id, askwin->win.x, askwin->win.y);

    return;
}

void MapAskwin(Xdata *xdata, Askwin *askwin)
{
    XMapSubwindows(xdata->display, askwin->win.id);
    XMapRaised(xdata->display, askwin->win.id);
    return;
}

void UnmapAskwin(Xdata *xdata, Askwin *askwin)
{
    XUnmapWindow(xdata->display, askwin->win.id);
    return;
}

void CreateAskwin(Xdata *xdata, Windata *win, char *mesg,
                    int x, int y, Askwin *askwin)
{
    GC gc;
    int strwidth;
    static XFontStruct *fontstruct = NULL;

    gc = XCreateGC(xdata->display, xdata->root.id, 0, 0);
    XSetForeground(xdata->display, gc, xdata->root.fg);
    XSetBackground(xdata->display, gc, xdata->root.bg);
    if( fontstruct == NULL ) {
        fontstruct = XLoadQueryFont(xdata->display,
                      "-adobe-times-bold-r-normal-*-*-240-*-*-p-*-iso8859-*");
        if( fontstruct == NULL ) {
	    fprintf(stderr, "%s: cannot query font!\n", myname);
	    exit(1);
        }
    }
    XSetFont(xdata->display, gc, fontstruct->fid);

    /* flame */
    ConstructWindata(xdata, win->id, x, y,
                     ASKWIN_WIDTH, ASKWIN_HEIGHT, 1,
                     xdata->root.fg, xdata->root.bg, 0, &askwin->win);
    askwin->pm = XCreatePixmap(xdata->display, askwin->win.id,
                           ASKWIN_XPM.width, ASKWIN_XPM.height, xdata->depth);
    if( askwin->pm == None ) {
        fprintf(stderr, "%s: cannot create pixmap!\n", myname);
        exit(1);
    }
    strwidth = XTextWidth(fontstruct, mesg, strlen(mesg));
    XCopyArea(xdata->display, ASKWIN_XPM.pixmap, askwin->pm, gc, 0, 0,
	      ASKWIN_XPM.width, ASKWIN_XPM.height, 0, 0);
    XDrawString(xdata->display, askwin->pm, gc,
		(ASKWIN_WIDTH - strwidth) / 2, ASKWIN_MESG_Y,
		mesg, strlen(mesg));
    XSetWindowBackgroundPixmap(xdata->display, askwin->win.id, askwin->pm);
    XSelectInput(xdata->display, askwin->win.id,
		 KeyPressMask | ButtonPressMask | PointerMotionHintMask |
		 ButtonMotionMask | ExposureMask);
    /* yes */
    CreateButton(xdata, &askwin->win, YES_XPM.pixmap, YES_F_XPM.pixmap,
                 YES_X, YES_Y, ITEM_WIDTH, ITEM_HEIGHT, 1, &askwin->yes);
    /* no */
    CreateButton(xdata, &askwin->win, NO_XPM.pixmap, NO_F_XPM.pixmap,
                 NO_X, NO_Y, ITEM_WIDTH, ITEM_HEIGHT, 1, &askwin->no);

    XFreeGC(xdata->display, gc);
}

void DestroyAskwin(Xdata *xdata, Askwin *askwin)
{
    XFreePixmap(xdata->display, askwin->pm);
    XDestroyWindow(xdata->display, AskwinId(*askwin));
}

void ExposeButton(Xdata *xdata, Button *button)
{
    XClearWindow(xdata->display, button->win.id);
}

void FocusButton(Xdata *xdata, Button *button, int mode)
{
    if(mode == BUTTON_ENTER)
        XCopyArea(xdata->display, button->pm, button->win.id, xdata->gc,
                  0, 0, button->win.width, button->win.height, 0, 0);
    else
        XClearWindow(xdata->display, button->win.id);
    return;
}

void MapButton(Xdata *xdata, Button *button)
{
    XMapSubwindows(xdata->display, button->win.id);
    XMapRaised(xdata->display, button->win.id);
    return;
}

void UnmapButton(Xdata *xdata, Button *button)
{
    XUnmapWindow(xdata->display, button->win.id);
    return;
}

void CreateButton(Xdata *xdata, Windata *win,
                  Pixmap button_pixmap, Pixmap button_f_pixmap,
		  int x, int y, int width, int height, int bw, Button *button)
{
    ConstructWindata(xdata, win->id,
                     x, y, width, height, bw,
                     xdata->root.fg, xdata->root.bg, 0, &button->win);
    button->pm = button_f_pixmap;
    XSetWindowBackgroundPixmap(xdata->display, button->win.id, button_pixmap);
    XSelectInput(xdata->display, button->win.id,
                 ButtonPressMask | EnterWindowMask | LeaveWindowMask |
                 ExposureMask);
}


void InputWin(Xdata *xdata, Windata *parent,
	      int x, int y, char *title, char *ret)
{
    Windata inputwin, disp;

    XEvent ev;
    GC gc;
    XFontStruct *fontstruct;
    KeySym key;
    XComposeStatus comstatus;

    char buff[BUFFLEN];
    int len;
    int retlen;
    int done;
	
    ConstructWindata(xdata, parent->id,
                     x, y, INPUTWIN_WIDTH, INPUTWIN_HEIGHT, 1,
                     xdata->root.fg, xdata->root.bg, 0, &inputwin);
    XSetWindowBackgroundPixmap(xdata->display, inputwin.id, XINPUT_XPM.pixmap);
    XSelectInput(xdata->display, inputwin.id, KeyPressMask | ExposureMask);

    ConstructWindata(xdata, inputwin.id,
                     DISP_X, DISP_Y, DISP_WIDTH, DISP_HEIGHT, 1,
                     xdata->root.fg, xdata->root.bg, 0, &disp);
    XSelectInput(xdata->display, disp.id,
		 KeyPressMask | ExposureMask | ButtonPress);

    gc = XCreateGC(xdata->display, xdata->root.id, 0, 0);
    XSetForeground(xdata->display, gc, xdata->root.fg);
    XSetBackground(xdata->display, gc, xdata->root.bg);
    fontstruct = XLoadQueryFont(xdata->display,
		"-adobe-times-bold-r-normal-*-*-180-*-*-p-*-iso8859-*");
    XSetFont(xdata->display, gc, fontstruct->fid);
	
    XMapSubwindows(xdata->display, inputwin.id);
    XMapRaised(xdata->display, inputwin.id);

    ret[0] = '\0';

    done = 0;
    while(done == 0)
    {
	XNextEvent(xdata->display, &ev);
	switch(ev.type)
	{
          case KeyPress:
	    len = XLookupString(&ev.xkey, buff, BUFFLEN, &key, &comstatus);
	    buff[len] = '\0';
	    switch(key)
	    {
              case XK_Delete:
              case XK_BackSpace:
		retlen = strlen(ret);
		if(retlen > 0)
		    ret[retlen - 1] = '\0';
		break;
              case XK_Return:
		done = 1;
		break;
              default:
		if(isprint(buff[0]))
		    strcat(ret, buff);
		break;
	    }
	    ExposeDisp(*xdata, disp, gc, fontstruct, ret);
	    break;
          case MappingNotify:
	    XRefreshKeyboardMapping(&ev.xmapping);
	    break;
          case Expose:
	    if(ev.xexpose.window == inputwin.id)
		ExposeInputwin(*xdata, inputwin, gc, fontstruct, title);
	    else if(ev.xexpose.window == disp.id)
		ExposeDisp(*xdata, disp, gc, fontstruct, ret);
	    break;
          case ButtonPress:
	    if( ev.xbutton.button == 2 ) {
		char *cutbuf;
		int len;
		cutbuf = XFetchBytes(xdata->display, &len);
		if( cutbuf != NULL ) {
		    int i = 0;
		    while( 1 ) {
			if( i >= len ) {
                            char *dst = ret;
			    while( *dst ) dst++;
			    for( i = 0 ; i < len ; i++ )
				*dst++ = cutbuf[i];
			    *dst = '\0';
			    break;
			}
			else if( !isprint(cutbuf[i]) ) {
			    XBell(xdata->display, 0);
			    break;
			}
			i++;
		    }
		    XFree(cutbuf);
		}
		ExposeDisp(*xdata, disp, gc, fontstruct, ret);
	    }
	    break;
	}
    }

    XDestroyWindow(xdata->display, inputwin.id);
	
    return;
}

void ExposeDisp(Xdata xdata, Windata disp, GC gc, XFontStruct *fontstruct,
                char *ret)
{
    int strwidth;
	
    XClearWindow(xdata.display, disp.id);
    XDrawString(xdata.display, disp.id, gc, DISP_OFFSET_X, DISP_OFFSET_Y,
                ret, strlen(ret));
    strwidth = XTextWidth(fontstruct, ret, strlen(ret));
    XFillRectangle(xdata.display, disp.id, gc,
                   DISP_OFFSET_X + strwidth, DISP_OFFSET_Y - CURSOR_HEIGHT,
                   CURSOR_WIDTH, CURSOR_HEIGHT);

    return;
}

void ExposeInputwin(Xdata xdata, Windata inputwin, GC gc,
                    XFontStruct *fontstruct, char *title)
{
    int strwidth;

    XClearWindow(xdata.display, inputwin.id);
    strwidth = XTextWidth(fontstruct, title, strlen(title));
    XDrawString(xdata.display, inputwin.id, gc,
                (INPUTWIN_WIDTH - strwidth) / 2, TITLE_Y,
                title, strlen(title));

    return;
}
