#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>

#include "xdata.h"
#include "xtool.h"
#include "setsocket.h"
#include "commdata.h"
#include "tuitate.h"
#include "all.h"
#include "selectenemy.h"
#include "game.h"
#include "viewnotation.h"
#include "selectgame.h"
#include "viewgame.h"
#include "selectnotation.h"
#include "cursor.h"
#include "drawstring.h"
#include "version.h"
#include "misc.h"
#include "title.h"
#include "xmisc.h"
#include "resource.h"
#include "xpm/xtuitate_xpm.h"

#include "xpm/MASK_r.xbm"
#include "xpm/MASK_re.xbm"

#ifdef	USEICON
#include "tuitate.xbm"
#endif

#define INPUT_X		((WIN_WIDTH - INPUTWIN_WIDTH) / 2)
#define INPUT_Y		((WIN_HEIGHT - INPUTWIN_HEIGHT) / 2)

#if 0
#define TITLE_WIDTH	(WIN_WIDTH - 64)			/* 606 */
#define TITLE_HEIGHT	(WIN_HEIGHT - 96 - BUTTON_HEIGHT)	/* 440 */
#define TITLE_X		32
#define TITLE_Y		32
#else
#define	TITLE_WIDTH	640
#define	TITLE_HEIGHT	480
#define	TITLE_X		15
#define	TITLE_Y		8
#endif

#define BUTTON_WIDTH	128
#define BUTTON_HEIGHT	64
#define GAME_X		32
#define VIEWGAME_X	192
#define VIEWNOTE_X	352
#define QUIT_X		512
#define BUTTON_Y	(WIN_HEIGHT - BUTTON_HEIGHT - 32)

#define ASKQUIT_X	((WIN_WIDTH	- ASKWIN_WIDTH) / 2)
#define ASKQUIT_Y	((WIN_HEIGHT - ASKWIN_HEIGHT) / 2)

#define NORMAL		0		/* for phase */
#define ASKQUIT				1

static void ButtonPressProc(XEvent *);
static void MotionNotifyProc(XEvent ev);
static void CrossNotifyProc(XEvent ev, int crossmode);
static void ExposeProc(XEvent ev);
static void InputNameAddress(void);
static void CustomizeGCSet(void);
static void MapAllButton(void);
static void UnmapAllButton(void);
static void XtuitateProperties(void);
static void SetMode(int argc, char **argv);
static void InitKomaPixmap(void);

								/* global */
char *myname;
Xdata xdata;
Windata win;
GCSet gcset;
Pixmap komapm[32];
TuNotation notation;
const char *komaname[14] = {
    "PAWN", "LANCE", "KNIGHT", "SILVER", "GOLD",
    "BISHOP", "ROOK", "KING",
    "P_PAWN", "P_LANCE", "P_KNIGHT", "P_SILVER",
    "P_BISHOP", "P_ROOK"
};
majormode_t majormode;
char name[2][BUFFLEN];
char address[2][BUFFLEN];
int sock;
int side;
int bell;
int auto_raise = 0;
unsigned long debug = 0;
char *geometry = NULL;

/* local */
static Windata title;
static Button game, viewgame, viewnote, quit;
static Askwin askquit;
XColor back, poo;
static char *displayname;
static char *servername;
static char *background_pixmap_file;
static char username[BUFFLEN];
static char useraddress[BUFFLEN];
static newfile;
static int phase;
static int done;

#define	__PIX(x) x,
static char ** pixmap_data[ ] = {
#include "pixmaps.h"
};

MyPixmap mypixmap[NUMBER_OF_PIXMAPS];

int main(int argc, char **argv)
{
    XEvent ev;

    myname = argv[0];

    /* set option */
    SetMode(argc, argv);

    if( getResourceBoolean(".help", ".Help") ) {
	putHelp();
	exit(0);
    }

    /* set xdata */
    xdata = ConstructXdata(displayname);

    if( DebuggingXlib )
	XSynchronize(xdata.display, True);

    /* win */
    ConstructWindata(&xdata, xdata.root.id,
                     0, 0, WIN_WIDTH, WIN_HEIGHT, 1,
                     xdata.root.fg, backgroundPixel(), 0, &win);

    /* map win */
    XSelectInput(xdata.display, win.id, ExposureMask);
    XMapRaised(xdata.display, win.id);
    initial_title();

    /* check connection to server */
    sock = setup_connection(getResourceString(".game.server", ".Game.Server"),
			    getResourceStringNoError(".game.port",
						     ".Game.Port"));
    if(sock < 0)
    {
        fprintf(stderr, "Cannot connect to game-server `%s'.\n",
                servername);
        exit(-1);
    }
    SendData("QUIT\n", 5);
    closeConnection();
    sock = setup_connection(getResourceString(".viewGame.server",
					      ".ViewGame.Server"),
			    getResourceStringNoError(".viewGame.port",
						     ".ViewGame.Port"));
    if(sock < 0)
    {
        fprintf(stderr, "Cannot connect to viewgame-server `%s'.\n",
                servername);
        exit(-1);
    }
    SendData("QUIT\n", 5);
    closeConnection();
    sock = setup_connection(getResourceString(".viewNotation.server",
					  ".ViewNotation.Server"),
			getResourceStringNoError(".viewNotation.port",
						 ".ViewNotation.Port"));
    if(sock < 0)
    {
        fprintf(stderr, "Cannot connect to viewnotation-server `%s'.\n",
                servername);
        exit(-1);
    }
    SendData("QUIT\n", 5);
    closeConnection();

    ProvideCursors( );
    XtuitateInitializeFonts(xdata.display);

    /*
     * Initialize pixmaps ...
     */
    {
	int i;

	for( i = 0 ; i < NUMBER_OF_PIXMAPS ; i++ )
	    create_pixmap_from_data(&mypixmap[i], pixmap_data[i]);
    }

    SetCursor(win.id, CSR_NORMAL);

    XSelectInput(xdata.display, win.id, ExposureMask);
    {
	Pixmap iconpixmap;
#ifdef	USEICON
	iconpixmap = XCreateBitmapFromData(xdata.display, xdata.root.id,
					   tuitate_bits,
					   tuitate_width, tuitate_height);
	if( iconpixmap == None ) {
	    fprintf(stderr, "Can't create icon pixmap.\n");
	    exit( 1 );
	}
#else
	iconpixmap = None;
#endif
	{
	    XTextProperty winname, iconname;
	    XSizeHints normalhints;
	    XWMHints wmhints;
	    XClassHint classhints;

	    winname.value	= iconname.value	= "xtuitate";
	    winname.encoding	= iconname.encoding	= XA_STRING;
	    winname.format	= iconname.format	= 8;
	    winname.nitems	= iconname.nitems	= sizeof("xtuitate")-1;
	    normalhints.flags	= PSize;
	    normalhints.width	= WIN_WIDTH;
	    normalhints.height	= WIN_HEIGHT;
	    wmhints.flags	= WindowGroupHint;
	    wmhints.window_group= win.id;
	    classhints.res_name	= "xtuitate";
	    classhints.res_class= "Tuitate";

	    XSetWMProperties(xdata.display, win.id, &winname, &iconname,
			     argv, argc, &normalhints, &wmhints, &classhints);
	}
	/*
	  MakeSizeHints(&xdata, &win, "xtuitate", "xtuitate",
	  iconpixmap, argc, argv);
	  MakeClassHint(&xdata, &win, "xtuitate", "Tuitate");
	  */
	XtuitateProperties( );
    }

    {
	char *bg_pixmap_file;

	bg_pixmap_file = getResourceStringNoError(".backgroundImage",
						  ".BackgroundImage");
	if( bg_pixmap_file ) {
	    Pixmap pixmap;

	    pixmap = ReadFileToPixmap(bg_pixmap_file);
	    XSetWindowBackgroundPixmap(xdata.display, win.id, pixmap);
	    XClearWindow(xdata.display, win.id);
	}
    }
    /* title */
    ConstructWindata(&xdata, win.id,
                     TITLE_X, TITLE_Y, TITLE_WIDTH, TITLE_HEIGHT, 1,
                     xdata.root.fg, xdata.root.bg, 0, &title);
#if 0
    XpmCreatePixmapFromData(xdata.display, xdata.root.id,
#if 0
			    title_xpm, 
#else
			    xtuitate,
#endif
			    &titlepm, 0, 0);
#endif
    XSetWindowBackgroundPixmap(xdata.display, title.id, TITLE_XPM.pixmap);
    XSelectInput(xdata.display, title.id, ExposureMask);

    /* button */
    CreateButton(&xdata, &win, GAME_XPM.pixmap, GAME_F_XPM.pixmap,
                 GAME_X, BUTTON_Y, BUTTON_WIDTH, BUTTON_HEIGHT, 1, &game);
    CreateButton(&xdata, &win,
                 VIEWGAME_XPM.pixmap, VIEWGAME_F_XPM.pixmap,
                 VIEWGAME_X, BUTTON_Y,
                 BUTTON_WIDTH, BUTTON_HEIGHT, 1, &viewgame);
    CreateButton(&xdata, &win,
                 VIEWNOTE_XPM.pixmap, VIEWNOTE_F_XPM.pixmap,
                 VIEWNOTE_X, BUTTON_Y,
                 BUTTON_WIDTH, BUTTON_HEIGHT, 1, &viewnote);
    CreateButton(&xdata, &win, QUIT_XPM.pixmap, QUIT_F_XPM.pixmap,
                 QUIT_X, BUTTON_Y, BUTTON_WIDTH, BUTTON_HEIGHT, 1, &quit);

    /* askquit */
    CreateAskwin(&xdata, &win, "Quit?", ASKQUIT_X, ASKQUIT_Y, &askquit);

    /* init gcset */
    CustomizeGCSet();

    InitKomaPixmap();	    /* init komapm */

    /* input name & address*/
    InputNameAddress();

    /* map other window */
    MapAllButton();

    XStoreName(xdata.display, win.id, "xtuitate");

    /* init phase */
    phase = NORMAL;

    /* main loop */
    done = 0;
    while( done == 0 )
    {
	majormode = MAIN;
        XNextEvent(xdata.display, &ev);
        switch(ev.type)
        {
          case ButtonPress:
            ButtonPressProc(&ev);
            break;
          case MotionNotify:
            MotionNotifyProc(ev);
            break;
          case EnterNotify:
            CrossNotifyProc(ev, BUTTON_ENTER);
            break;
          case LeaveNotify:
            CrossNotifyProc(ev, BUTTON_LEAVE);
            break;
          case Expose:
            ExposeProc(ev);
            break;
        }
    }

    if( debug )
	fprintf(stderr, "%s: Successfully exit.\n", argv[0]);

    return 0;
}

static void ButtonPressProc(XEvent *ev)
{
    char buff[BUFFLEN];

    if(ev->xbutton.window == ButtonId(game))
    {
	UnmapAllButton();
	majormode = GAME;
	strcpy(name[0], username);
	strcpy(address[0], useraddress);
	sock = setup_connection(getResourceString(".game.server",
						  ".Game.Server"),
				getResourceStringNoError(".game.port",
							 ".Game.Port"));
        if(sock < 0)
        {
            fprintf(stderr, "Cannot connect to game-server `%s'.\n",
                    servername);
            exit(-1);
        }
        sprintf(buff, "GAME %s %s\n", name[0], address[0]);
        SendData(buff, strlen(buff));
        if(SelectEnemy() == SE_DONE_START)
            if(Game() == GAME_DONE_VIEWNOTE)
                ViewNotation();
        closeConnection();
        MapAllButton();
    }
    else if(ev->xbutton.window == ButtonId(viewgame))
    {
        UnmapAllButton();
	majormode = VIEWGAME;
	sock = setup_connection(getResourceString(".viewGame.server",
					      ".ViewGame.Server"),
			    getResourceStringNoError(".viewGame.port",
						     ".ViewGame.Port"));
        if(sock < 0)
        {
            fprintf(stderr, "Cannot connect to viewgame-server `%s'.\n",
                    servername);
            exit(-1);
        }
	sprintf(buff, "VIEW\n");
	SendData(buff, strlen(buff));
	if(SelectGame() == SG_DONE_START)
	    ViewNotation();
        closeConnection();
	MapAllButton();
    }
    else if(ev->xbutton.window == ButtonId(viewnote))
    {
	UnmapAllButton();
	majormode = VIEWNOTATION;
	sock = setup_connection(getResourceString(".viewNotation.server",
						  ".ViewNotation.Server"),
				getResourceStringNoError(".viewNotation.port",
							 ".ViewNotation.Port"));
        if(sock < 0)
        {
            fprintf(stderr, "Cannot connect to viewnotation-server `%s'.\n",
                    servername);
            exit(-1);
        }
	sprintf(buff, "NOTATION\n");
	SendData(buff, strlen(buff));
	if(SelectNotation() == SG_DONE_START)
	    ViewNotation();
        closeConnection();
	MapAllButton();
    }
    else if(ev->xbutton.window == ButtonId(quit))
    {
	MapAskwin(&xdata, &askquit);
	UnmapAllButton();
	phase = ASKQUIT;
    }
    else if(ev->xbutton.window == AskwinId(askquit))
	MoveAskwin(&xdata, &askquit, &win);
    else if(ev->xbutton.window == AskwinYesId(askquit))
	done = 1;
    else if(ev->xbutton.window == AskwinNoId(askquit))
    {
	UnmapAskwin(&xdata, &askquit);
	MapAllButton();
	phase = NORMAL;
    }
    XStoreName(xdata.display, win.id, "xtuitate");
    majormode = MAIN;
    return;
}

static void MotionNotifyProc(XEvent ev)
{
    MoveAskwin(&xdata, &askquit, &win);
    return;
}

static void CrossNotifyProc(XEvent ev, int crossmode)
{
    if(ev.xcrossing.window == ButtonId(game))
        FocusButton(&xdata, &game, crossmode);
    else if(ev.xcrossing.window == ButtonId(viewgame))
        FocusButton(&xdata, &viewgame, crossmode);
    else if(ev.xcrossing.window == ButtonId(viewnote))
        FocusButton(&xdata, &viewnote, crossmode);
    else if(ev.xcrossing.window == ButtonId(quit))
        FocusButton(&xdata, &quit, crossmode);
    else if(ev.xcrossing.window == AskwinYesId(askquit))
        FocusButton(&xdata, &askquit.yes, crossmode);
    else if(ev.xcrossing.window == AskwinNoId(askquit))
        FocusButton(&xdata, &askquit.no, crossmode);
    return;
}

static void ExposeProc(XEvent ev)
{
    if(ev.xexpose.window == win.id)
        XClearWindow(xdata.display, win.id);
    else if(ev.xexpose.window == title.id)
        XClearWindow(xdata.display, title.id);
    else if(ev.xexpose.window == ButtonId(game))
        ExposeButton(&xdata, &game);
    else if(ev.xexpose.window == ButtonId(viewgame))
        ExposeButton(&xdata, &viewgame);
    else if(ev.xexpose.window == ButtonId(viewnote))
        ExposeButton(&xdata, &viewnote);
    else if(ev.xexpose.window == ButtonId(quit))
        ExposeButton(&xdata, &quit);
    else if(ev.xexpose.window == AskwinId(askquit))
        XClearWindow(xdata.display, AskwinId(askquit));
    else if(ev.xexpose.window == AskwinYesId(askquit))
        ExposeButton(&xdata, &askquit.yes);
    else if(ev.xexpose.window == AskwinYesId(askquit))
        ExposeButton(&xdata, &askquit.no);
    return;
}

static void InputNameAddress(void)
{
#if 0
    FILE *fp;
    char buff[BUFFLEN];

    strcpy(buff, getenv("HOME"));
    strcat(buff, "/.xtuitate");

    fp = fopen(buff, "r");

    if(fp == NULL || newfile == 1) {
        InputWin(&xdata, &win, INPUT_X, INPUT_Y,
                 "Input your name ..", username);
        InputWin(&xdata, &win, INPUT_X, INPUT_Y,
                 "Input your address ..", useraddress);
        if(name[0][0] == '\0')
            strcpy(username, "Unknown");
        if(address[0][0] == '\0')
            strcpy(useraddress, "Unknown");
        fp = fopen(buff, "w+");
        fprintf(fp, "name: %s\naddress: %s\n", username, useraddress);
    }
    else {
	char *ptr;
	char buf[256];

	username[0] = useraddress[0] = '\0';

	while( fgets(buf, sizeof(buf), fp)!= NULL ) {

	    for( ptr = strchr(buf, '\0') - 1 ;
		 ptr >= buf && (*ptr == ' ' || *ptr == '\t' || *ptr == '\n') ;
		 *ptr = '\0' ) ;	/* ζ'\n' ä */

	    if( !buf[0] || buf[0] == '#' )
		continue;	/* ԡ# ϥå */

	    if( !strncasecmp(buf, "default-server: ", 16) ) {
		if( servername == NULL )
		    servername = strdup(&buf[16]);
	    }
	    else if( !strncasecmp(buf, "name: ", 6) ) {
		if( username[0] )
		    fprintf(stderr, "Warning: multiple `%s' in ~/.xtuitate\n",
			    "name:");
		strcpy(username, &buf[6]);
	    }
	    else if( !strncasecmp(buf, "address: ", 9) ) {
		if( useraddress[0] )
		    fprintf(stderr, "Warning: multiple `%s' in ~/.xtuitate\n",
			    "address:");
		strcpy(useraddress, &buf[9]);
	    }
	    else {
		fprintf(stderr,
			"Warning: can't understand `%s' in ~/.xtuitate (",
			buf);
		if( !username[0] ) {
		    fprintf(stderr, "assumed your name");
		    strcpy(username, buf);
		}
		else if( !useraddress[0] ) {
		    fprintf(stderr, "assumed your address");
		    strcpy(useraddress, buf);
		}
		else
		    fprintf(stderr, "discarded");
		fprintf(stderr, ")\n");
	    }
	}
    }

    fclose(fp);
#else
    parseUserName(username, sizeof(username),
		  getResourceString(".userName", ".UserName"));
    parseUserName(useraddress, sizeof(useraddress),
		  getResourceString(".userAddress", ".UserAddress"));

#ifdef	DEBUG
    if( Debugging('R') ) {
	printf("User   : %s\n", username);
	printf("Address: %s\n", useraddress);
    }
#endif

#endif
    return;
}

static void CustomizeGCSet(void)
{
    Font font;
    Pixmap MASK[2];
    int i;

    InitGCSet(&xdata, &gcset);

    font = XLoadFont(xdata.display,
                     "-adobe-times-bold-r-normal-*-*-180-*-*-p-*-iso8859-*");
    for(i = 0; i < 8; i++)
        XSetFont(xdata.display, gcset.type[7], font);

    RegistGCSet(&xdata, &gcset, 0, 0, 0, KOMAGC0);
    MASK[0] = XCreateBitmapFromData(xdata.display, xdata.root.id,
                                    MASK_r_bits, MASK_r_width,
                                    MASK_r_height);
    XSetClipMask(xdata.display, gcset.type[KOMAGC0], MASK[0]);
    RegistGCSet(&xdata, &gcset, 0, 0, 0, KOMAGC1);
    MASK[1] = XCreateBitmapFromData(xdata.display, xdata.root.id,
                                    MASK_re_bits, MASK_re_width,
                                    MASK_re_height);
    XSetClipMask(xdata.display, gcset.type[KOMAGC1], MASK[1]);

    return;
}

static void MapAllButton(void)
{
    XMapRaised(xdata.display, title.id);
    MapButton(&xdata, &game);
    MapButton(&xdata, &viewgame);
    MapButton(&xdata, &viewnote);
    MapButton(&xdata, &quit);
    return;
}

static void UnmapAllButton(void)
{
    XUnmapWindow(xdata.display, title.id);
    UnmapButton(&xdata, &game);
    UnmapButton(&xdata, &viewgame);
    UnmapButton(&xdata, &viewnote);
    UnmapButton(&xdata, &quit);
    return;
}

static void XtuitateProperties(void)
{
    Atom XA_TUITATE_VERSION;

    XA_TUITATE_VERSION = XInternAtom(xdata.display, "_TUITATE_VERSION", False);
    XChangeProperty(xdata.display, win.id, XA_TUITATE_VERSION, XA_STRING, 8,
		    PropModeReplace, (unsigned char *)version,
		    strlen(version));
}

static void SetMode(int argc, char **argv)
{
    int i;
    char *ptr;

    servername = getenv("TUITATESERVER");

    background_pixmap_file = NULL;

#ifdef	TUITATE_SERVER
    if( servername == NULL )
	servername = TUITATE_SERVER;
#endif

    parse_command(argc, argv);

    displayname = get_resource_string(".display", NULL);
    servername = get_resource_string(".server", servername);
    bell = get_resource_boolean(".bell", True);
    newfile = get_resource_boolean(".newFile", False);
    background_pixmap_file = get_resource_string(".background", NULL);
    auto_raise = get_resource_boolean(".autoRaise", False);

    {
	char *ptr;

	if( (ptr = getResourceString(".debugFlag", ".DebugFlag")) != NULL ) {
	    if( *ptr == '-' ) {
		debug = ~0;
		ptr++;
	    }
	    while( *ptr ) {
		if( '@' <= *ptr && *ptr <= 'Z' )
		    debug ^= 1 << (*ptr - '@');
		else if( 'a' <= *ptr && *ptr <= 'z' )
		    debug ^= 1 << (*ptr - 'a' + 1);
		else {
		    fprintf(stderr, "%s: Invalid debug flag `%c'\n",
			    argv[0], *ptr);
		    exit(1);
		}
		ptr++;
	    }
	}
    }

    if( (geometry = get_resource_string(".geometry", NULL)) != NULL ) {
	Junk	j;
	int	mask;

	mask = XParseGeometry(geometry, &j.i, &j.i, &j.u, &j.u);
	if( (mask & (WidthValue | HeightValue | XValue | YValue)) == 0 ) {
	    fprintf(stderr, "%s: Invalid position specification\n",
		    argv[0]);
	    exit(1);
	}
    }

#if 0
    else if(strcmp(argv[i], "-h") == 0 ||
                strcmp(argv[i], "-help") == 0)
        {
	    printf("Usage : %s <options>\n", argv[0]);
	    printf("<options> : [-server name][-s name]\n");
	    printf("            [-display name][-d name]\n");
	    printf("            [-newfile][-nf]\n");
	    printf("            [-nobell][-nb]\n");
	    printf("            [-nobell][-nb]\n");
	    printf("            [-background <filename>]\n");
	    printf("            [-autoraise]\n");
	    exit(0);
        }

#endif

    if( debug ) {
	int bit;
	fprintf(stderr, "Debug flag:");
	for( bit = 0 ; bit < 32 ; bit++ )
	    if( debug & (1 << bit) )
		fprintf(stderr, " %c", '@' + bit);
	fputc('\n', stderr);
    }

    return;
}

static void InitKomaPixmap()
{
    struct {
	char **data;
	int koma;
    } KomaPixmap[ ] = {
	{ fu_r_xpm,	FU },
	{ kyosha_r_xpm,	KYOSHA },
	{ keima_r_xpm,	KEIMA },
	{ gin_r_xpm,	GIN },
	{ kin_r_xpm,	KIN },
	{ kaku_r_xpm,	KAKU },
	{ hisha_r_xpm,	HISHA },
	{ ou_r_xpm,	OU },
	{ to_r_xpm,	TO },
	{ nkyo_r_xpm,	NKYO },
	{ nkei_r_xpm,	NKEI },
	{ ngin_r_xpm,	NGIN },
	{ ryu_r_xpm,	RYU },
	{ uma_r_xpm,	UMA },
	{ fu_re_xpm,	FU	| TuSideMask },
	{ kyosha_re_xpm,KYOSHA	| TuSideMask },
	{ keima_re_xpm,	KEIMA	| TuSideMask },
	{ gin_re_xpm,	GIN	| TuSideMask },
	{ kin_re_xpm,	KIN	| TuSideMask },
	{ kaku_re_xpm,	KAKU	| TuSideMask },
	{ hisha_re_xpm,	HISHA	| TuSideMask },
	{ ou_re_xpm,	OU	| TuSideMask },
	{ to_re_xpm,	TO	| TuSideMask },
	{ nkyo_re_xpm,	NKYO	| TuSideMask },
	{ nkei_re_xpm,	NKEI	| TuSideMask },
	{ ngin_re_xpm,	NGIN	| TuSideMask },
	{ ryu_re_xpm,	RYU	| TuSideMask },
	{ uma_re_xpm,	UMA	| TuSideMask },
	{ NULL,		0 }
    }, *ptr;

    for( ptr = KomaPixmap ; ptr->data != NULL ; ptr++ )
	CreatePixmapFromData(xdata.display, xdata.root.id,
			     ptr->data, &komapm[ptr->koma & TuKomaMask],
			     0, 0);
}
